/* $Workfile: AlvisPCIeH100Demo.c $       $Revision: 4 $      $Modtime: 12/18/08 10:36a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc OtxDriverTest

  @module

    AlvisPCIeH100Demo.c

    Application for Alvis-Pci-Express board.

  Note: hardware provides a termination switch for each board. Only the boards 
        at the end of the cable will need the switch set to ON. The Alvis-PCIe 
        card has the switch (S1) located on the back side of the board near the
        H.100 connector.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDmp.h"
#include "OtxEth.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"

#include "../../../Sys/DeviceCore/OdZlEthSwDef.h"

#define DEMO_DEFAULT_LOG_FILE "AlvisBasic.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_DEFAULT_DMP_NO   8
#define DEMO_MAX_DMP_NO       8
#define DEMO_MAX_HWY_NO       4

#define DEMO_GAMMA_INCREMENT  1 // Each TS will be increased by this value
#define DEMO_EVENT_DATA_BUF_SIZE 32 // burst highway receive and transmit

#define DEMO_MAX_BURST_TX   128  // 128 timeslots per frame

#define DEMO_BURST_HWY_0    8
#define DEMO_BURST_HWY_1    9
#define DEMO_BURST_HWY_2   10
#define DEMO_BURST_HWY_3   11
#define DEMO_H100_HWY_0    16
#define DEMO_H100_HWY_1    17
#define DEMO_H100_HWY_2    18
#define DEMO_H100_HWY_3    19

#define DEMO_H100_NOT_CONNECTED -1
#define DEMO_H100_RECEIVE        0
#define DEMO_H100_TRANSMIT       1

#define DEMO_BURST_BUF_SIZE  (DEMO_MAX_BURST_TX * 512) //64K

#define DEMO_HW_BURST_BUF_SIZE 0x4000

// GAMMA data is normally configured with each timeslot having a 
// different value. But for debug the frame can be set to the same 
// value in all timeslots. To enable extra debug code set define to 1.
#define DEMO_GAMMA_FRAME_CNT 1

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   logFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  OTX_UINT32             boardNo;
  OTX_UINT32             nDmp;
  OTX_BOOL               bDisableBurstTx;
  OTX_BOOL               bDisableBurstRx;
  OTX_UINT32             coBurstHighways;
  OTX_UINT32             nBurstHwyTimeslots;
  OTX_UINT32             nCurrentHwy;
  OTX_INT32              nHbusRate;
  OTX_INT32              nH100HwyRate;
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;


static void DemoLogf(char *formatStr, ...);
static void DemoExitQuestion(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoBurstSender(OTX_UINT32 nSourceId, OTX_UINT32 nBlockSize);
static void DemoBurstReceiver(OTX_UINT32 nSourceId, OTX_UINT32 nBlockSize);
static void DemoConnectBurstDevicesToH100(OTX_UINT32 nTransferType);
static void DemoDisconnectBurstDevices(void);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static OTX_UINT8 g_sqSendFixedBuf[DEMO_MAX_HWY_NO][DEMO_BURST_BUF_SIZE];

DemoEnvT    g_demoEnv;
FILE *g_pDemoLogFile = NULL;
OTX_EVENT   g_hNotificationEvent      = OTX_INVALID_EVENT_VALUE;
OTX_EVENT   g_hNotificationBurstEvent = OTX_INVALID_EVENT_VALUE;

OTX_HANDLE g_hBoard            = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hTss              = OTX_INVALID_HANDLE_VALUE;  
OTX_HANDLE g_hEthSw            = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hEventQueue       = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hBurstEventQueue  = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hBurst            = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hCtBus            = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hLiOut[DEMO_MAX_DMP_NO];
OTX_HANDLE g_hDmp[DEMO_MAX_DMP_NO];

// Used by the PrintStats function
OTX_UINT8 g_nFrame;
OTX_BOOL  g_nFrameFlag = OTX_FALSE;
OTX_UINT8 g_sqResponse[40];
OTX_INT32 g_nBurstExitFlag = 0;
OTX_INT32 g_FetchCount     = 0;
OTX_INT32 g_FetchCountCur  = 0;
// show that H.100 is not connected
OTX_INT32 g_nH100Direction = DEMO_H100_NOT_CONNECTED; 
OTX_UINT32 g_nHwyFrameCnt[DEMO_MAX_HWY_NO];

OTX_UINT32 g_nTstBufferCount[DEMO_MAX_HWY_NO];

OTX_BOOL g_bDmpLinkInitialized[DEMO_MAX_DMP_NO];

OTX_BOOL g_bBurstVerbose    = OTX_FALSE; // Enable data print message

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  // Default is -1 (All TS)
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;

  OTX_UINT32  nGammaZeroXferCnt;
  OTX_UINT8   sqGammaData[DEMO_MAX_BURST_TX];
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)

  OTX_INT32   nDataBlockCounter;
  OTX_BOOL    bGammaTestMode;
  OTX_BOOL    bFixedTestMode;
  OTX_BOOL    bGammaDisplayDataErr;
  OTX_BOOL    bTestReset;
  OTX_INT32   nGammaTestErrors[DEMO_MAX_BURST_TX];
  OTX_UINT8   sqGammaTestData[DEMO_MAX_BURST_TX];
  OTX_INT32   sqGammaTempData[8][4];
};

// one for each possible highway hosted by the Burst device
struct hLogicDevs {
  struct sBurstSender mBurstSender[DEMO_MAX_HWY_NO]; 
  struct sBurstReceiver mBurstReceiver[DEMO_MAX_HWY_NO]; 
} g_hLogicDevs;

struct hPipes {
  OTX_HANDLE hBurstIn[DEMO_MAX_DMP_NO];
  OTX_HANDLE hBurstOut[DEMO_MAX_DMP_NO];
  OTX_HANDLE hH100In[DEMO_MAX_HWY_NO][DEMO_MAX_BURST_TX];
  OTX_HANDLE hH100Out[DEMO_MAX_HWY_NO][DEMO_MAX_BURST_TX];
} g_hPipes;


/*******************************************************************************
*******************************************************************************/
void DemoCheckRc(IN OTX_RESULT result)
{
  if ( OTX_FAILURE(result) ) {
    char msgStr[256];
    OTX_RESULT otxRc;
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", 
      OTX_RESULT_CODE(result), 
      OTX_RESULT_SOURCE_ID(result), 
      OTX_RESULT_LINE_NUM(result), 
      msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  DemoLogf("Do you want to exit (Y/N) ");
  c = OsaConsoleGetCh();
  
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************
  @func
    Display a buffer of data that is conntrolled by the board device
*******************************************************************************/
void DemoDumpDataBuffer(OTX_UINT32 nAddress, OTX_UINT32 nCnt)
{
  OTX_RESULT otxRc;
  OTX_UINT32 readVal;
  OTX_UINT32 nIndex;
  OTX_UINT32 nOffset;

  DemoLogf("\n");

  nIndex = 0;
  DemoLogf("hBoard Address 0x%.4X, Count 0x%.8X\n", nAddress, nCnt);
  while(nIndex < nCnt) {
    // Calculate offset into buffer for this read
    nOffset = nIndex * sizeof(OTX_UINT32);
    // Group data values in sets of 4
    if(!(nIndex % 4)) {
      DemoLogf("\n");
      DemoLogf("Address 0x%.X : ", nAddress + nOffset);
    }
    otxRc = OtxDrvReadReg32(g_hBoard, nAddress + nOffset, &readVal); 
    DemoCheckRc(otxRc);
    DemoLogf("0x%.8X, ", (OTX_UINT32)readVal);

    nIndex++; 
  }

  DemoLogf("\n");
}


/*******************************************************************************
  @func
    Burst Device Test Stop all Testing
*******************************************************************************/
void DemoBurstEndTest(void)
{
  OTX_RESULT  otxRc;
  OtxDeviceStateE  eDevState;
  OTX_UINT32 nHwy;

  otxRc = OtxDrvGetState(g_hBurst, &eDevState); 
  DemoCheckRc(otxRc);

  if (eDevState == OTX_DEVICE_WORKING) {
    DemoLogf("Disable g_hBurst\n");
    otxRc = OtxDrvDisable(g_hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(otxRc);
  }
  else {
    DemoLogf("Burst device is already disabled\n");
  }

  // There can 1 burst highway per DMM
  for(nHwy = 0; nHwy < DEMO_MAX_HWY_NO; nHwy++) {
    g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode = OTX_FALSE;
    g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_FALSE;

    if (g_hLogicDevs.mBurstReceiver[nHwy].bActive) {
      if (g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter == 0) {
        DemoLogf("Error: No burst data was received, Highway[%d]\n", nHwy);
      }
    }

    g_hLogicDevs.mBurstReceiver[nHwy].bActive = OTX_FALSE;
    if (g_hLogicDevs.mBurstReceiver[nHwy].pRxFile) {
      DemoLogf("BurstReceiver[%d]: Closing file\n", nHwy);
      fclose(g_hLogicDevs.mBurstReceiver[nHwy].pRxFile);
      g_hLogicDevs.mBurstReceiver[nHwy].pRxFile = NULL;
    }

    g_hLogicDevs.mBurstSender[nHwy].bActive   = OTX_FALSE;
    if (g_hLogicDevs.mBurstSender[nHwy].pTxFile) {
      DemoLogf("BurstSender[%d]: Closing file\n", nHwy);
      fclose(g_hLogicDevs.mBurstSender[nHwy].pTxFile);
      g_hLogicDevs.mBurstSender[nHwy].pTxFile = NULL;
    }
  }
}


/*******************************************************************************
  @func
    Burst Device Test Function using GAMMA Pattern
*******************************************************************************/
void DemoBurstTestFixedData(OTX_UINT32 nHwy)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  i;
  OtxDeviceStateE  eDevState;

  otxRc = OtxDrvGetState(g_hBurst, &eDevState); 
  DemoCheckRc(otxRc);

  for(i = 0; i < DEMO_BURST_BUF_SIZE; i++) {
#if 1
    g_sqSendFixedBuf[nHwy][i] = (OTX_UINT8) nHwy + 1;
#else
    OTX_UINT8 nData;
    nData = (OTX_UINT8) nHwy | (OTX_UINT8) (i << 4);
    g_sqSendFixedBuf[nHwy][i] = nData;
#endif
  }


  if (eDevState == OTX_DEVICE_WORKING) {
    DemoLogf("Burst device is already enable\n");
  }
  else {
    // By enabling the Burst device, we start the process of transferring data
    // from the hardware to the driver via DMA.  The data is stored in the Burst
    // device's buffer for later processing to connected logical devices.
    DemoLogf("Enable g_hBurst\n");
    otxRc = OtxDrvEnable(g_hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(otxRc);
  }

  if(g_nH100Direction == DEMO_H100_TRANSMIT) {

    if (g_hLogicDevs.mBurstSender[nHwy].bActive == OTX_FALSE) {
      DemoLogf("Burst Sender Hwy[%d]. Sending a gamma pattern continuously\n", nHwy);

      g_hLogicDevs.mBurstSender[nHwy].pTxFile = NULL;

      // The starting point for the gamma pattern will be value=timeslot
      g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = 0;

      g_hLogicDevs.mBurstSender[nHwy].bActive = OTX_TRUE;

      g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = DEMO_BURST_BUF_SIZE;
    } 
    else {
      DemoLogf("Burst Sender Hwy[%d] already running.\n", nHwy);
    }
  }


  if(g_nH100Direction == DEMO_H100_RECEIVE) {
    if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_FALSE) {
      DemoLogf("Burst Receiver Hwy[%d]: Receiving data (no file logging).\n", nHwy);
      g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter = 0;
      g_hLogicDevs.mBurstReceiver[nHwy].bActive = OTX_TRUE;
    }
    else {
      DemoLogf("Burst Receiving Hwy[%d] already running.\n", nHwy);
    }
  }

  if (!g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode &&
      !g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode) {
    // Fixed mode is not enabled. So enable it.

    g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode = OTX_TRUE;
    DemoLogf("Burst Receiver Hwy[%d] Fixed Test Mode is now enabled\n", nHwy);

  } else {
    DemoLogf("Burst Receiver Hwy[%d] Test Mode already running\n", nHwy);
  }
}


/*******************************************************************************
  @func
    Burst Device Test Function using GAMMA Pattern
*******************************************************************************/
void DemoBurstTest(OTX_UINT32 nHwy)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  i;
  OtxDeviceStateE  eDevState;

  otxRc = OtxDrvGetState(g_hBurst, &eDevState); 
  DemoCheckRc(otxRc);

  if (eDevState == OTX_DEVICE_WORKING) {
    DemoLogf("Burst device is already enable\n");
  }
  else {
    // By enabling the Burst device, we start the process of transferring data
    // from the hardware to the driver via DMA.  The data is stored in the Burst
    // device's buffer for later processing by connected logical devices.
    DemoLogf("Enable g_hBurst\n");
    otxRc = OtxDrvEnable(g_hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(otxRc);
  }


  if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_FALSE) {
    DemoLogf("Burst Receiver Hwy[%d]: Receiving data (no file logging).\n", nHwy);
    g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter = 0;
    g_hLogicDevs.mBurstReceiver[nHwy].bActive = OTX_TRUE;
  }
  else {
    DemoLogf("Burst Receiving Hwy[%d] already running.\n", nHwy);
  }


  if(g_nH100Direction == DEMO_H100_TRANSMIT) {
    if (g_hLogicDevs.mBurstSender[nHwy].bActive == OTX_FALSE) {
      DemoLogf("Burst Sender Hwy[%d]. Sending a gamma pattern continuously\n", nHwy);

      g_hLogicDevs.mBurstSender[nHwy].pTxFile = NULL;

      // The starting point for the gamma pattern will be value=timeslot
      g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = 0;
      for (i = 0; i < (OTX_UINT32) DEMO_MAX_BURST_TX; i++) {
        g_hLogicDevs.mBurstSender[nHwy].sqGammaData[i] = (unsigned char) i;
//        g_hLogicDevs.mBurstSender[nHwy].sqGammaData[i] = (unsigned char) nHwy;
      }

      g_hLogicDevs.mBurstSender[nHwy].bActive = OTX_TRUE;

      g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = DEMO_BURST_BUF_SIZE;
    } 
    else {
      DemoLogf("Burst Sender Hwy[%d] already running.\n", nHwy);
    }
  }

  if (!g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode &&
      !g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode) {
    // Gamma mode is not enabled. So enable it.
    int n;
    OTX_UINT8 *pData;

    pData = g_hLogicDevs.mBurstReceiver[nHwy].sqGammaTestData;
    memset(pData, 0, DEMO_MAX_BURST_TX);
    for (n = 0; n < DEMO_MAX_BURST_TX; n++) {
      g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[n] = 0;
    }
  
    g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_TRUE;
    DemoLogf("Burst Receiver Hwy[%d] Gamma Test Mode is now enabled\n", nHwy);

  } else {
    DemoLogf("Burst Receiver Hwy[%d] Test Mode already running\n", nHwy);
  }

}


/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId, OTX_UINT32 nBlockSize)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;


  // validate parameters
  if ((nSourceId < 0) || (nSourceId > g_demoEnv.coBurstHighways)) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstSender[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  if(DEMO_BURST_BUF_SIZE > nBlockSize - g_demoEnv.nBurstHwyTimeslots) {
    // Not enough room to fit the entire buffer
    return;
  }

START:
  if (g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0], 0xFF, sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]));

    if (g_hLogicDevs.mBurstSender[nSourceId].nTsToSend == -1) {
      // All TS
      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = fread(
        &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf,
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );
      
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32]; // On some boards this is bigger than we actually use
      int cbSingleTsBuf;
      int byteCount = 0;
      int cbMaxSingleTsBuf = DEMO_BURST_BUF_SIZE / g_demoEnv.nBurstHwyTimeslots;
      int x;
      
      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        cbMaxSingleTsBuf,
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );

      if (cbSingleTsBuf > 0) {
        if ((cbSingleTsBuf > 0) && (cbSingleTsBuf < cbMaxSingleTsBuf)) {
          // We are at the end of the file, and we don't have a full buffer to send.
          // Pad with 0xFF at the end of the buffer to make sure that we don't underrun the Burst Sender
          DemoLogf("Pad with 0xFF [%d] at the end of the buffer to make sure that we don't underrun the Burst Sender\n", (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]));
          memset(&singleTsBuf[cbSingleTsBuf], 0xFF, (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]) );
          cbSingleTsBuf = cbMaxSingleTsBuf;
        }
        
        // Dispurse the single TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        //DemoLogf("g_demoEnv.coBurstHighways=%d, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend=%d \n", g_demoEnv.nCurrentHwy, g_demoEnv.coLi, g_demoEnv.coBurstHighways, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend);
        for ( x = g_hLogicDevs.mBurstSender[nSourceId].nTsToSend; 
        (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
        x += g_demoEnv.nBurstHwyTimeslots
          ) {
          g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount++];
        }
        
        assert(byteCount == cbSingleTsBuf);
          
      }

      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = cbSingleTsBuf * g_demoEnv.nBurstHwyTimeslots;

    }
    
    if (g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf > 0) {
      ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;

      assert(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf == DEMO_BURST_BUF_SIZE);

      otxRc = OtxBurstSendData(g_hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
      DemoCheckRc(otxRc);
      
      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      if (ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
        DemoLogf("DemoBurstSender[%d]: ui32 (%d) != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf (%d)\n", nSourceId, ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      }
      
    } else {
      
      if (g_hLogicDevs.mBurstSender[nSourceId].bFileLoop) {
        if (g_hLogicDevs.mBurstSender[nSourceId].bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        goto START;
      } 
      else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
        g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;
      }
    }
    
  } 
  else {
    // Send a gamma or fixed pattern. GAMMA means that we increase the 
    // value by one for each frame (continuously until aborted)
    OTX_UINT32 x;
    OTX_UINT32 y;
    OTX_UINT8  *pDataTo;
    OTX_UINT8  *pDataFrom;
    OTX_UINT8  sqDataStart[DEMO_MAX_BURST_TX];
    OTX_BOOL   bGamma = OTX_FALSE;

    pDataTo = &sqDataStart[0];
    pDataFrom = &g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[0];
    for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
      *pDataTo = *pDataFrom;
      pDataTo++;
      pDataFrom++;
    }

    pDataTo = &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0];

    if (g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode) {
      pDataFrom = g_sqSendFixedBuf[nSourceId];
    }
    else {
      if(g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
        bGamma = OTX_TRUE;
      }
      else {
        // Need to send something send fixed
        pDataFrom = g_sqSendFixedBuf[nSourceId];
      }
    }

    for (x = 0; x < DEMO_BURST_BUF_SIZE; x += g_demoEnv.nBurstHwyTimeslots) {
      // Set the frame data pointer, all channel are updated
      if(bGamma) {
        pDataFrom = &g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[0];
      }
      for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
        *pDataTo = *pDataFrom;
        if (bGamma) {
          *pDataFrom = *pDataFrom + DEMO_GAMMA_INCREMENT;
        }
        pDataTo++;
        pDataFrom++;
      }
    }

    ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;
    otxRc = OtxBurstSendData(g_hBurst, 
                             nSourceId, 
                             g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, 
                             &ui32);
    DemoCheckRc(otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    // (since we set the trigger amount to the buffer size, there should
    // have been room to buffer this, so these two should be the same).
    //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
    if(ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
      if((ui32 == 0) && (bGamma)) {
        // sent no data
        OTX_UINT32 y;
        pDataTo = &g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[0];
        pDataFrom = &sqDataStart[0];
        for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
          *pDataTo = *pDataFrom;
          pDataTo++;
          pDataFrom++;
        }
        g_hLogicDevs.mBurstSender[nSourceId].nGammaZeroXferCnt++;
      }
      else {
        DemoLogf("DemoBurstSender[%d]: OtxBurstSendData did not copy entire buffer.\n", nSourceId);
        DemoLogf("   Buffer size sent %d\n", g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
        DemoLogf("   Buffer size returned %d\n", ui32);
        DemoLogf("g_FetchCount %d, g_FetchCountCur %d, nBlockSize %d\n",  
                 g_FetchCount,
                 g_FetchCountCur,
                 nBlockSize); 
      }
    }
  }

  return;
}


/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstCheckDataFrame(OTX_UINT8 *pInputData, OTX_UINT32 nHwy)
{
  OTX_UINT8 *pCheckValue;
  OTX_UINT8 *pData;
  OTX_UINT32 nTs;
  OTX_BOOL   *pDataErr;

  g_nHwyFrameCnt[nHwy]++;

  pDataErr = &g_hLogicDevs.mBurstReceiver[nHwy].bGammaDisplayDataErr;
  pCheckValue = g_sqSendFixedBuf[nHwy];
  pData = pInputData;

  for(nTs=0; nTs < g_demoEnv.nBurstHwyTimeslots; nTs++) {
    if(*pData != *pCheckValue) {
      if((*pDataErr)) {
        DemoLogf("\nErr: Hwy(%d) Expected 0x%X Received 0x%X, Time slot %d\n",  
                 nHwy,
                 *pCheckValue,
                 *pData,
                 nTs);

        DemoLogf("\n");

        g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[0]++;
        if(g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[0] >= 10) {
          *pDataErr = OTX_FALSE;
          g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[0] = 0;
        }
      }
    }

    pData++;
    pCheckValue++;
  }
}


/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId, OTX_UINT32 nBlockSize)
{
  OTX_RESULT    otxRc;
  OTX_UINT32    ui32;
  OTX_UINT32    nBufferSize;
  OTX_UINT32    nLoopErrCnt = 0;
  OTX_UINT32    nDisplayCnt = 0;

  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;
  OTX_BOOL      bDisplayBuffer = OTX_FALSE;
  OTX_BOOL      bGamma = OTX_FALSE;
  OTX_UINT8     *pBuf;

  static OTX_UINT8 HwyBuf[DEMO_MAX_HWY_NO][DEMO_BURST_BUF_SIZE];

  // validate parameters
  if ((nSourceId < 0) || (nSourceId > g_demoEnv.coBurstHighways)) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  g_hLogicDevs.mBurstReceiver[nSourceId].nDataBlockCounter++;
  // Each highway has a buffer get this highway's buffer address
  pBuf = &HwyBuf[nSourceId][0];

  memset(pBuf, 0, DEMO_BURST_BUF_SIZE);

  // we want the frame reference id and time
  ui32 = DEMO_BURST_BUF_SIZE;
  otxRc = OtxBurstReadDataEx(g_hBurst, nSourceId, pBuf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  // (since we set the trigger amount to the buffer size, there should
  // have been room to buffer this, so these two should be the same).
  //assert(ui32 == sizeof(buf));
  if(ui32 != DEMO_BURST_BUF_SIZE) {
    if(ui32 == 0) {
      // Nothing for use to process buffer empty
      return;
    }
    // Process what we received

    /*
    DemoLogf("DemoBurstReceiver[%d] Receive a short buffer size %d should be %d\n", 
             nSourceId, 
             ui32, 
             DEMO_BURST_BUF_SIZE);
    */
  }

  // We may not always get a full buffer
  nBufferSize = ui32;
  // Every frame in the buffer will need data, 
  // check for a partial frame.
  if(nBufferSize % g_demoEnv.nBurstHwyTimeslots) {
    DemoLogf("DemoBurstReceiver[%d] Buffer contains a partial frame\n", 
             nSourceId);
    DemoLogf("DemoBurstReceiver[%d] Buf size %d, Frame size %d\n", 
             nSourceId, 
             nBufferSize,
             g_demoEnv.nBurstHwyTimeslots);
    DemoLogf("g_FetchCount %d, g_FetchCountCur %d, nBlockSize %d\n",  
             g_FetchCount,
             g_FetchCountCur,
             nBlockSize); 

    // The size needs to be a multiple of the timeslots
    nBufferSize = (ui32/g_demoEnv.nBurstHwyTimeslots) * g_demoEnv.nBurstHwyTimeslots;
  }


  if(nSourceId != 0) {
    // demo code only supports a single highway
    return; // do not check data at this time
  }


  // Check the received value if we are in Gamma or fixed test mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode ||
      g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
    OTX_UINT32 nFrame;
    OTX_UINT32 nFrameSize;
    OTX_UINT32 nBufCount;
    OTX_UINT32 nTimeSlot;
    OTX_UINT8  *pBufData;
    OTX_UINT8  *pData;

    nFrameSize = g_demoEnv.nBurstHwyTimeslots;
    nLoopErrCnt = 0;
    nBufCount   = 0;
    pData = g_sqSendFixedBuf[nSourceId];
    pBufData    = pBuf;
    if (g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode) {
      // Fixed data pattern is saved in the buffer
      pData = g_sqSendFixedBuf[nSourceId];
    }
    else {
      if(g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
        bGamma = OTX_TRUE;
        // GAMMA pattern: update for each frame
        pData = &g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[0];
      }
    }
    
    for (nFrame = 0; nFrame < nBufferSize; nFrame += nFrameSize) {
      if (bGamma) {
        // GAMMA pattern: update for each frame
        pData = &g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[0];
#if DEMO_GAMMA_FRAME_CNT
        {
          OTX_UINT32 nCheck;
          for(nCheck = 0; nCheck < nFrameSize; nCheck++) {
            if(pData[0] != pData[nCheck]) {
              // All data should be the same
              // correct the data at this point
              pData[nCheck] = pData[0];
            }
          }
        }
#endif
      }
      if(g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode) {
        if(g_hLogicDevs.mBurstReceiver[nSourceId].bTestReset == OTX_FALSE) {
          OTX_UINT32 nFrameAdd;
          OTX_UINT8  *pFrame;

          // Display first error only
          nFrameAdd = (OTX_UINT32) pBufData + nFrame;
          pFrame = (OTX_UINT8 *) nFrameAdd;
          DemoBurstCheckDataFrame(pFrame, nSourceId);
        }
      }
      else
      for (nTimeSlot = 0; nTimeSlot < g_demoEnv.nBurstHwyTimeslots; nTimeSlot++) {

#if 0
        // The following code re-enables the display of any data that is 
        // not correct. If all is working no errors are displayed. But it 
        // can fill the display with data very fast if errors are detected.
        if(nLoopErrCnt) {
          // enable it again for the next frame of data in this buffer
          g_hLogicDevs.mBurstReceiver[nSourceId].bGammaDisplayDataErr = OTX_TRUE;
        }
#endif

        if(g_hLogicDevs.mBurstReceiver[nSourceId].bTestReset == OTX_TRUE) {
          // User wants to reset the error count and start value
          if (bGamma) {
            *pData = *pBufData;
          }
          g_hLogicDevs.mBurstReceiver[nSourceId].nGammaTestErrors[nTimeSlot] = 0;
        }
        else
        if (*pBufData != *pData) {
          // Data mismatch. Step error count and resync
          if(g_hLogicDevs.mBurstReceiver[nSourceId].bGammaDisplayDataErr) {

            // Found a buffer with bad data
            DemoLogf("\nErr: Hwy(%d) Expected 0x%X Received 0x%X, Block Counter %d, Frame Counter %d, Time slot %d \n",  
                     nSourceId,
                     *pData,
                     *pBufData,
                     g_hLogicDevs.mBurstReceiver[nSourceId].nDataBlockCounter,
                     nFrame,
                     nTimeSlot);
            DemoLogf("g_FetchCount %d, g_FetchCountCur %d,  nGammaZeroXferCnt %d\n",  
                     g_FetchCount,
                     g_FetchCountCur,
                     g_hLogicDevs.mBurstSender[nSourceId].nGammaZeroXferCnt);

            nLoopErrCnt++;

            if(nDisplayCnt == 0) {
              // Display this buffer
              nDisplayCnt = nFrame;
              bDisplayBuffer = OTX_TRUE;
            }

            g_hLogicDevs.mBurstReceiver[nSourceId].bGammaDisplayDataErr = OTX_FALSE;
          }
          // Data was in error, next check will be based on this value
          if (bGamma) {
#if DEMO_GAMMA_FRAME_CNT
            // GAMMA data is normally configured with each timeslot having a 
            // different value. But for debug the frame can be set to the same 
            // value in all timeslots.
            if((*pBufData != g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[0] - 1) &&
               (nTimeSlot != 0)) {
              // Bad data all TS are the same
              *pData = g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[0] - 1;
            }
            else {
              *pData = *pBufData;
            }
#else
              *pData = *pBufData;
#endif
          }
          g_hLogicDevs.mBurstReceiver[nSourceId].nGammaTestErrors[nTimeSlot]++;
        }
        
        if (bGamma) {
          // Step the GammaTestData value, which is the next value that we expect
          *pData = *pData + DEMO_GAMMA_INCREMENT;
        }
        pData++;
        pBufData++;
        nBufCount++;
      } // For time slot
    } // For frame
  }
  if(nLoopErrCnt) {
    g_hLogicDevs.mBurstReceiver[nSourceId].bGammaDisplayDataErr = OTX_FALSE;
  }
  if(g_hLogicDevs.mBurstReceiver[nSourceId].bTestReset == OTX_TRUE) {
    g_hLogicDevs.mBurstReceiver[nSourceId].bTestReset = OTX_FALSE;
    g_hLogicDevs.mBurstReceiver[nSourceId].nDataBlockCounter = 0;
    DemoLogf("Burst Hwy[%d] Reset test\n", nSourceId);
  }

  // Echo if we're in verbose mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose || bDisplayBuffer) {
    OTX_UINT32 x;
    OTX_UINT32 nEndCnt;
    // Display one buffer only, set above if error is detected
    bDisplayBuffer = OTX_FALSE;
    if(nLoopErrCnt) {
      DemoLogf("\nBurst Receiver error count %d\n", nLoopErrCnt);
    }
    DemoPrintTimeStamp(dtFrameRefTime);

    // Note: we only display the least significant 32 bits of the frame reference id
    DemoLogf("Burst Receiver Hwy[%d] (len=%.5d) ID:%08x, Starting Offset 0x%X\n", 
              nSourceId, 
              DEMO_BURST_BUF_SIZE, 
              (OTX_UINT32) (nFrameRefId & 0xffffffff),
              nDisplayCnt);

    // Normally it starts at frame 0, but an error above can set it
    // to the frame that was in error. 
    // Print one frame of data.
    x = nDisplayCnt;
    nEndCnt = x + g_demoEnv.nBurstHwyTimeslots;
    while( (x < nEndCnt) && (x < DEMO_BURST_BUF_SIZE) ) {
      DemoLogf("%.2X ", pBuf[x]);
      x++;
      if(!(x%16)) {
        DemoLogf("\n"); 
      }
    }

    DemoLogf("\n");
  }
  
  // write data to a file if desired
  if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile && 
      g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    int cbWritten;

    if (g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(pBuf, 
                         nBufferSize, 
                         nBufferSize, 
                         g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != (int) DEMO_BURST_BUF_SIZE) {
        if (!g_nBurstExitFlag) {
          DemoLogf("Burst Receiver Hwy[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, DEMO_BURST_BUF_SIZE, cbWritten);
        }
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      OTX_UINT32 x;

      byteCount = 0;
      for (
          x = g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord;
          x < nBufferSize; 
          x += g_demoEnv.nBurstHwyTimeslots
      ) {
        singleTsBuf[byteCount++] = pBuf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != byteCount) {
        if (!g_nBurstExitFlag) {
          DemoLogf("Burst Receiver Hwy[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
        }
      } else {
        //DemoLogf("Burst Receiver Hwy[%d] (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}


void DemoConnectBurstDevicesToH100(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset 
    of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a 
    characterized but its capacity (in 8 kbit/s increments). The user can create 
    a pipe with the driver function <f OtxDrvCreatePipe>. A pipe is connected to 
    physical or logical devices and mapped to specific time-slots with the driver 
    function <f OtxDrvConnectPipe>.

*******************************************************************************/
  OTX_UINT32  nDirection // Direction of H100, if 0 we receive data
)
{
  OTX_RESULT result;
  OTX_HANDLE hPipe;
  OTX_UINT32 nTs;
  OTX_UINT32 nH100Ts;
  OTX_INT32 nH100Highway;
  OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
  OTX_UINT8 sqH100TimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
  OTX_INT32 sqTssBurstHighway[4];
  OTX_INT32 sqTssH100Highway[4];
  OTX_UINT32 nH100MaskSize;
  OTX_UINT8  *pH100Mask;
  
  // Close any pipes that may have been created
  DemoDisconnectBurstDevices();

  sqTssH100Highway[0]  = DEMO_H100_HWY_0; 
  sqTssH100Highway[1]  = DEMO_H100_HWY_1; 
  sqTssH100Highway[2]  = DEMO_H100_HWY_2; 
  sqTssH100Highway[3]  = DEMO_H100_HWY_3; 

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ALVIS_PCI_EXPRESS:
    sqTssBurstHighway[0]  = DEMO_BURST_HWY_0;
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
    break;
  }

  DemoLogf("Creating pipes for Burst %s TSS H100Hwy=%d, BurstHwy=%d, Burst TS size=%d\n", 
           nDirection == 0 ? "Sender" : "Receiver",
           sqTssH100Highway[0], 
           sqTssBurstHighway[0], 
           g_demoEnv.nBurstHwyTimeslots);

  nH100Highway = 0;
  nH100Ts = 0;


  // --------------------------------------------------------------------
  // Create pipe: Burst -> H100, one pipe for each time slot
  // --------------------------------------------------------------------
  for (nTs = 0; nTs < g_demoEnv.nBurstHwyTimeslots; nTs++) {

    result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);

    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
    memset(sqH100TimeSlotMask, 0, sizeof(sqH100TimeSlotMask));
    sqBurstTimeSlotMask[nTs]    = 0xFF;
    sqH100TimeSlotMask[nH100Ts] = 0xFF;

    // The Highway is configured by the driver during initialization
    // and may not support the selected data rate.
    if(g_demoEnv.nH100HwyRate == 2048) {
      pH100Mask = sqH100TimeSlotMask;
      nH100MaskSize = 32;
    }
    else
    if(g_demoEnv.nH100HwyRate == 8192) {
      pH100Mask = sqBurstTimeSlotMask;
      nH100MaskSize = 128;
    }
    else {
      // Not supported at this time
      DemoLogf("Creating pipes for Burst size %d not supported\n", g_demoEnv.nH100HwyRate);
      return;
    }

    if(nDirection == 0) {
      g_nH100Direction = DEMO_H100_TRANSMIT;
      // ---------------------------
      // Connect pipe: Burst -> H100
      // ---------------------------
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_INPUT,  
                                 g_hTss, 
                                 sqTssBurstHighway[0], 
                                 sqBurstTimeSlotMask, 
                                 g_demoEnv.nBurstHwyTimeslots);
      DemoCheckRc(result);
    
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_OUTPUT, 
                                 g_hTss, 
                                 sqTssH100Highway[nH100Highway], 
                                 pH100Mask, 
                                 nH100MaskSize);
      DemoCheckRc(result);

      // Save the pipe handle it could be used later.
      // The pipe will need to be deleted when we are finished.
      g_hPipes.hH100Out[nH100Highway][nTs] = hPipe;
    }
    else 
    if(nDirection == 1) {
      g_nH100Direction = DEMO_H100_RECEIVE;
      // --------------------------- 
      // Create pipe: H100 -> Burst
      // --------------------------- 
      result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
    
      // ---------------------------
      // Connect pipe: H100 -> Burst
      // ---------------------------
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_INPUT,  
                                 g_hTss, 
                                 sqTssH100Highway[nH100Highway], 
                                 pH100Mask, 
                                 nH100MaskSize);
      DemoCheckRc(result);
    
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_OUTPUT, 
                                 g_hTss, 
                                 sqTssBurstHighway[0], 
                                 sqBurstTimeSlotMask, 
                                 g_demoEnv.nBurstHwyTimeslots);
      DemoCheckRc(result);

      // Save the pipe handle it could be used later.
      // The pipe will need to be deleted when we are finished.
      g_hPipes.hH100In[nH100Highway][nTs] = hPipe;
    }
    else {
      DemoLogf("Type (%d) not supported by this demo program\n", nDirection);
      nTs = g_demoEnv.nBurstHwyTimeslots; // exit for loop
    }

    //DemoPrintPipeConnections(hPipe);

    nH100Ts++;
    if((nH100Ts == 32) && (nTs < g_demoEnv.nBurstHwyTimeslots - 1)) {
      nH100Ts = 0;
      nH100Highway++;
      DemoLogf("Creating pipes for Burst %s TssH100Hwy=%d, TssBurstHwy=%d, BurstTs=%d\n", 
               nDirection == 0 ? "Sender" : "Receiver",
               sqTssH100Highway[nH100Highway], 
               sqTssBurstHighway[0],
               nTs + 1);
    }
  }
}


/*******************************************************************************

  @func
    Remove pipe connections

*******************************************************************************/
void DemoDisconnectBurstDevices(void)
{
  OTX_UINT32 nPipe;
  OTX_UINT32 nHwy;
  OTX_RESULT otxRc;

  g_nH100Direction = DEMO_H100_NOT_CONNECTED; // show that it is not connected

  // For each pipe:
  //     Check to make sure there was a pipe created
  //     Disable the pipe connection
  //     Close the pipe connect
  //     Set the global to unused
  for (nPipe = 0; nPipe < DEMO_MAX_BURST_TX; nPipe++) {

    if(g_hPipes.hBurstIn[nPipe] != OTX_INVALID_HANDLE_VALUE) {
      otxRc = OtxDrvDisconnectPipe(g_hPipes.hBurstIn[nPipe], OTX_PIPE_INPUT);
      if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
        DemoCheckRc(otxRc);
      }
      otxRc = OtxDrvDisconnectPipe(g_hPipes.hBurstIn[nPipe], OTX_PIPE_OUTPUT);
      if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
        DemoCheckRc(otxRc);
      }

      // Close the pipe
      otxRc = OtxDrvClose(g_hPipes.hBurstIn[nPipe]); 
      DemoCheckRc(otxRc);

    }
    if(g_hPipes.hBurstOut[nPipe] != OTX_INVALID_HANDLE_VALUE) {
      otxRc = OtxDrvDisconnectPipe(g_hPipes.hBurstOut[nPipe], OTX_PIPE_INPUT);
      if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
        DemoCheckRc(otxRc);
      }
      otxRc = OtxDrvDisconnectPipe(g_hPipes.hBurstOut[nPipe], OTX_PIPE_OUTPUT);
      if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
        DemoCheckRc(otxRc);
      }

      // Close the pipe
      otxRc = OtxDrvClose(g_hPipes.hBurstOut[nPipe]); 
      DemoCheckRc(otxRc);
    }

    // Set handle to unused
    g_hPipes.hBurstIn[nPipe]  = OTX_INVALID_HANDLE_VALUE;
    g_hPipes.hBurstOut[nPipe] = OTX_INVALID_HANDLE_VALUE;
  }

  for (nHwy = 0; nHwy < DEMO_MAX_HWY_NO; nHwy++) {

    for (nPipe = 0; nPipe < DEMO_MAX_BURST_TX; nPipe++) {
      if(g_hPipes.hH100In[nHwy][nPipe] != OTX_INVALID_HANDLE_VALUE) {
        otxRc = OtxDrvDisconnectPipe(g_hPipes.hH100In[nHwy][nPipe], OTX_PIPE_INPUT);
        if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
          DemoCheckRc(otxRc);
        }
        otxRc = OtxDrvDisconnectPipe(g_hPipes.hH100In[nHwy][nPipe], OTX_PIPE_OUTPUT);
        if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
          DemoCheckRc(otxRc);
        }

        // Close the pipe
        otxRc = OtxDrvClose(g_hPipes.hH100In[nHwy][nPipe]); 
        DemoCheckRc(otxRc);

      }

      g_hPipes.hH100In[nHwy][nPipe]  = OTX_INVALID_HANDLE_VALUE;

      if(g_hPipes.hH100Out[nHwy][nPipe] != OTX_INVALID_HANDLE_VALUE) {
        otxRc = OtxDrvDisconnectPipe(g_hPipes.hH100Out[nHwy][nPipe], OTX_PIPE_INPUT);
        if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
          DemoCheckRc(otxRc);
        }
        otxRc = OtxDrvDisconnectPipe(g_hPipes.hH100Out[nHwy][nPipe], OTX_PIPE_OUTPUT);
        if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
          DemoCheckRc(otxRc);
        }

        // Close the pipe
        otxRc = OtxDrvClose(g_hPipes.hH100Out[nHwy][nPipe]); 
        DemoCheckRc(otxRc);
      }

      g_hPipes.hH100Out[nHwy][nPipe] = OTX_INVALID_HANDLE_VALUE;
    }
  }
}



/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32 nCurrentDmp  = 0;
  OTX_RESULT otxRc;

  if(commandChar > 0x20) {
    DemoLogf("\n Command > %c \n", commandChar);
  }
  else {
    DemoLogf("\n");
  }

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("\n");
    DemoLogf("\nCmd   Synopsis, Current DMP#%d\n", nCurrentDmp);
    DemoLogf("===   ========\n");
    DemoLogf(" ?    Display Help text\n");
    DemoLogf(" h    Ditto\n");
    DemoLogf(" 0-7  Select current DMP (used for subsequent commands)\n");
    DemoLogf(" a    Send Debug command to device g_hBoard\n");
    DemoLogf(" A    Stop Burst Loopback test\n");
    DemoLogf(" b    Boot DMP%d \n", nCurrentDmp);
    DemoLogf(" B    Set DMP to Console  \n");
    DemoLogf(" g    Enable (start) Burst Loopback GAMMA test\n");
    DemoLogf(" G    Clear Burst loopback status counters\n");
    DemoLogf(" j    Print Gamma Test Results for current H(condensed form)\n");
    DemoLogf(" J    Read 32 bit register from board device\n");
    DemoLogf(" K    Write 32 bit register to board device\n");
    DemoLogf(" C    Send Constant Byte on TSS Highway\n");
    DemoLogf(" c    Connect H.100 to Burst device\n");
    DemoLogf(" &    Send Constant Byte on TS0 equal to Highway number\n");
    DemoLogf(" +    Disconnect and Close all pipe connection\n");
    DemoLogf(" v    Toggle Burst verbose\n");
    DemoLogf(" d    Display HW burst buffer\n");

    break;


  case '+':
    // Close any pipes that may have been created
    DemoLogf("Closing all burst pipe connection\n");
    DemoDisconnectBurstDevices();
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    // Select current DMP
    {
      nCurrentDmp = commandChar - '0';
      DemoLogf("DMP#%d selected\n", nCurrentDmp);

      if((nCurrentDmp == 0) || (nCurrentDmp == 1)) {
        g_demoEnv.nCurrentHwy = 0;
      }
      else
      if((nCurrentDmp == 2) || (nCurrentDmp == 3)) {
        g_demoEnv.nCurrentHwy = 1;
      }
      else
      if((nCurrentDmp == 4) || (nCurrentDmp == 5)) {
        g_demoEnv.nCurrentHwy = 2;
      }
      else {
        g_demoEnv.nCurrentHwy = 3;
      }
      DemoLogf("Burst#%d selected\n", g_demoEnv.nCurrentHwy);

    }
    break;

  case 'a':
    {
      char strBuf[12];
      static OTX_UINT32 nCmd = 0;

      DemoLogf("Debug command for device g_hBoard 0x%X\n", g_hBoard);

      DemoLogf("Enter command (%d) 0=no command sent\n", nCmd);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nCmd = strtoul(strBuf, NULL, 0);
      }
      if(nCmd) {
        OtxDrvDbgDump(g_hBoard, nCmd);
      }
      DemoLogf("Done\n");
    }
    break;

  case 'A':
    {
      DemoLogf("End Burst Loopback test\n");

      DemoBurstEndTest();
    }
    break;

  case 'g':
    {
      OTX_UINT32 nDmpBurstHwy;
      // The pipe connections should have been setup. Just start
      // sending data.
      for(nDmpBurstHwy = 0; nDmpBurstHwy < g_demoEnv.coBurstHighways; nDmpBurstHwy++) {
        // This test does not need the DMP to be installed.
        // The pipe was connected between the H100 output and
        // the burst input to loop data. 
        DemoBurstTest(nDmpBurstHwy);
      }
    }

    break;

  case 'G':  // Zero Gamma/Fixed Test Error Counters
    {
      OTX_UINT32 nDmpBurstHwy;

      for(nDmpBurstHwy = 0; nDmpBurstHwy < DEMO_MAX_HWY_NO; nDmpBurstHwy++) {

        g_hLogicDevs.mBurstReceiver[nDmpBurstHwy].bTestReset = OTX_TRUE;

        // Will display the next buffer of data that fails, but only one buffer
        g_hLogicDevs.mBurstReceiver[nDmpBurstHwy].bGammaDisplayDataErr = OTX_TRUE;       
        
        DemoLogf("Burst Receiver Hwy[%d] Test Results are now zero\n", 
                 nDmpBurstHwy);
      }
    }
    break;

#if 0 // Used for debug of hardware
  case 'e':
    {
      OTX_UINT32 nDmpBurstHwy = 0;

      // The pipe connections should have been setup. Just start
      // sending data.
      for(nDmpBurstHwy = 0; nDmpBurstHwy < g_demoEnv.coBurstHighways; nDmpBurstHwy++) {
        // This test does not need the DMP to be installed.
        // The pipe was connected between the burst output and
        // the burst input to loop data. This test currently 
        // only verifies the data path for the burst device.
        DemoBurstTestFixedData(nDmpBurstHwy);
      }
    }

    break;

  case 'E':  // Zero Fixed or GAMMA Data Test Error Counters
    {
      OTX_UINT32 nDmpBurstHwy;

      for(nDmpBurstHwy = 0; nDmpBurstHwy < DEMO_MAX_HWY_NO; nDmpBurstHwy++) {

        g_nHwyFrameCnt[nDmpBurstHwy] = 0;

        g_hLogicDevs.mBurstReceiver[nDmpBurstHwy].bTestReset = OTX_TRUE;

        // Will display the next buffer of data that fails, but only one buffer
        g_hLogicDevs.mBurstReceiver[nDmpBurstHwy].bGammaDisplayDataErr = OTX_TRUE;       
        
        DemoLogf("Burst Receiver Hwy[%d] Data Test Results are now zero\n", 
                 nDmpBurstHwy);
      }
    }
    break;
#endif

  case 'j':  // Test Results (condensed form)
    {
      OTX_UINT32 nDmpHwy;

      for(nDmpHwy = 0; nDmpHwy < g_demoEnv.coBurstHighways; nDmpHwy++) {
        if (g_hLogicDevs.mBurstReceiver[nDmpHwy].bActive == OTX_TRUE) {
          if (g_hLogicDevs.mBurstReceiver[nDmpHwy].bGammaTestMode ||
              g_hLogicDevs.mBurstReceiver[nDmpHwy].bFixedTestMode) {
            if (g_hLogicDevs.mBurstReceiver[nDmpHwy].nDataBlockCounter == 0) {
              DemoLogf("Error: No burst data was received, Highway[%d]\n", nDmpHwy);
            }
            else {
              OTX_UINT32 y;
              OTX_INT32 nErrCnt;
              DemoLogf("Burst Receiver Hwy[%d] Gamma Test Results:\n", nDmpHwy);
              for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
                nErrCnt = g_hLogicDevs.mBurstReceiver[nDmpHwy].nGammaTestErrors[y];
                if ((y % 32) == 0) {
                  DemoLogf("TS[%.3d] ", y);
                }
                if ((y % 4) == 0) {
                  DemoLogf(" "); // Group by 4 timeslots to make it easier to read
                }
                if ((y % 8) == 0) {
                  DemoLogf(" "); // Group by 8 timeslots to make it easier to read
                }
                if (nErrCnt > 9) {
                  DemoLogf("*"); // Indicates a number larger than 9
                } else {
                  DemoLogf("%1d", nErrCnt);
                }
                if (((y+1) % 32) == 0) {
                  DemoLogf("\n");
                }
              }
              DemoLogf("\n");
            }
          } 
          else {
            DemoLogf("Test Mode is not enabled\n");
          }
        } 
        else {
          DemoLogf("Burst Receiver Thread Highway[%d] is not running\n", nDmpHwy);
        }
      }
    }
    break;

  case 'b':
    {
      OTX_INT32 c;
      OTX_HANDLE hDmp = g_hDmp[nCurrentDmp];

      DemoLogf("Boot DMP[%d]\n", nCurrentDmp);

      DemoLogf("Select Boot Source:\n");

      DemoLogf("0 = UART (mainly used to reload u-boot)\n");
      DemoLogf("1 = FLASH (normal boot source)\n");
      DemoLogf("2 = DSP Selfboot (use when debugging with Code Composer)\n");
      DemoLogf("3 = I changed my mind. I don't want to boot the DMP.\n");
      
      c = OsaConsoleGetCh();

      if(commandChar > 0x20) {
        DemoLogf("%c pressed\n", c);

        switch (c - '0') {
        case 0:
          DemoLogf("Booting from UART\n");
          otxRc = OtxDmpBoot(hDmp, OTX_DMP_BOOT_UART); DemoCheckRc(otxRc);
          break;
        case 1:
          DemoLogf("Booting from FLASH\n");
          otxRc = OtxDmpBoot(hDmp, OTX_DMP_BOOT_FLASH); DemoCheckRc(otxRc);
          break;
        case 2:
          DemoLogf("DSP Self-boot\n");
          otxRc = OtxDmpBoot(hDmp, OTX_DMP_BOOT_DSP); DemoCheckRc(otxRc);
          break;
        default:
          DemoLogf("Boot operation aborted\n");
          break;
        }
      }
      else {
        DemoLogf("\n");
      }

    
      DemoLogf("Done\n");
    }
    DemoLogf("\n");

    break;

  case 'B':
    {
      OTX_INT32 c;
      OTX_UINT32 nAttribValue;

      otxRc = OtxDrvGetAttributeValueUint32(g_hBoard, 
                                             OTX_ATTR_BRD_SERIAL_PORT_SEL, 
                                             &nAttribValue); 
      if(otxRc == OTX_S_FALSE) {
        DemoLogf("Optional I/O board has been detected.\n"
                 "All console ports are connected to the I/O board.\n");
      }
      else {
        DemoCheckRc(otxRc);
      }

      DemoLogf("DMP%d is mapped to console connector\n", nAttribValue);

      DemoLogf("Select:\n");
      DemoLogf("0-7 for DMP to be mapped to console port\n");
      DemoLogf("8   Abort, keep current setting\n");
      
      c = OsaConsoleGetCh();

      if(commandChar > 0x20) {

        DemoLogf("%c pressed\n", c);

        c = c - '0';

        switch (c) {
        case 0:
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
          otxRc = OtxDrvSetAttributeValueUint32(g_hBoard, 
                                                 OTX_ATTR_BRD_SERIAL_PORT_SEL, 
                                                 c); 

          otxRc = OtxDrvGetAttributeValueUint32(g_hBoard, 
                                                 OTX_ATTR_BRD_SERIAL_PORT_SEL, 
                                                 &nAttribValue); 
          DemoCheckRc(otxRc);

          DemoLogf("DMP%d is now mapped to console connector\n", nAttribValue);
          break;

        case 8:
          DemoLogf("User aborted\n");
          break;
        default:
          DemoLogf("Mapping of console connector aborted unknown entry\n");
          break;
        }
      }
      else {
        DemoLogf("\n");
      }

      DemoLogf("Done\n");
    }
    DemoLogf("\n");

    break;    

  case 'c':
    {
      char szBuf[64];
      static OTX_INT32 nTransferType;
      szBuf[0] = 0;
      
      DemoLogf("H.100 Transmit or Recieve\n");
      DemoLogf("Enter R or T : ");
      OsaConsoleGetStr(szBuf);
      if(szBuf[0]) {
        if((szBuf[0] == 'R') || (szBuf[0] == 'r')) {
          // Burst device will receive data from H.100
          nTransferType = 1; // Recieve data from H.100
          DemoConnectBurstDevicesToH100(nTransferType);
        }
        else
        if((szBuf[0] == 'T') || (szBuf[0] == 't')) {
          // Burst device will transmit data to H.100
          nTransferType = 0; // Send data to H.100
          DemoConnectBurstDevicesToH100(nTransferType);
        }
        else {
          DemoLogf("Abort entry - %c\n", szBuf[0]);
        }
      }
      else {
        DemoLogf("Abort no entry\n");
      }
    }
    break;

  case 'C':
    {
      OTX_UINT8 highWay;
      OTX_UINT8 channel;
      OTX_UINT8 byteVal;
      char wValStr[12];
      
      DemoLogf("Set Constant Byte on Channel\n");
      
      DemoLogf("Highway: ");
      OsaConsoleGetStr(wValStr);
      highWay = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Channel: ");
      OsaConsoleGetStr(wValStr);
      channel = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Constant Byte: ");
      OsaConsoleGetStr(wValStr);
      byteVal = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      otxRc = OtxBrdTssConstByte(g_hTss, highWay, channel, byteVal); 
      DemoCheckRc(otxRc);

      DemoLogf("Done.\n");
    }
    break;

  case '&':
    {
      OTX_UINT32 nDmpHwy;

      // Set Burst highway TS0 to the highway number.
      // Burst Highways are 8-11 on the TSS, DMP use 0-15
      for(nDmpHwy = 8; nDmpHwy < DEMO_MAX_HWY_NO + 8; nDmpHwy++) {
        otxRc = OtxBrdTssConstByte(g_hTss, nDmpHwy, 0, (OTX_UINT8)nDmpHwy); 
        DemoCheckRc(otxRc);
        DemoLogf("Setting Highway(%d) timeslot(0) to %d\n", nDmpHwy, nDmpHwy);
      }
    }
    break;

  case 'J':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from 32 bit register in the g_hBoard device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      otxRc = OtxDrvReadReg32(g_hBoard, nAddress, &readVal); DemoCheckRc(otxRc);

      DemoLogf("hBoard Address 0x%.4X = 0x%.8X\n", nAddress, (OTX_UINT32)readVal);
    }
    break;

  case 'K':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to 32 bit register in the g_hBoard device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      otxRc = OtxDrvWriteReg32(g_hBoard, nAddress, nValue); DemoCheckRc(otxRc);

      DemoLogf("Wrote 0x%.8X to hBoard Address 0x%.4X\n", nValue, nAddress);
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'v':

    g_bBurstVerbose = !g_bBurstVerbose;

    if(g_bBurstVerbose) {
      // Enable only current DMP burst verbose mode
      g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bVerbose = g_bBurstVerbose;
      DemoLogf("Burst DMP Highway[%d] verbose set to %s\n",
               g_demoEnv.nCurrentHwy,
               g_bBurstVerbose ? "Enabled" : "Disabled");
    }
    else {
      // Disable all just to make sure none are active
      OTX_UINT32 nDmpHwy;
      for(nDmpHwy = 0; nDmpHwy < DEMO_MAX_HWY_NO; nDmpHwy++) {
        g_hLogicDevs.mBurstReceiver[nDmpHwy].bVerbose = g_bBurstVerbose;

        DemoLogf("Burst DMP Highway[%d] verbose set to %s\n",
                 nDmpHwy,
                 g_bBurstVerbose ? "Enabled" : "Disabled");
      }
    }
    break;

  case 'd':
    {
      OTX_UINT32 nCnt;
      OTX_INT32 c;

      DemoLogf("Display Burst HW buffer, Tranmit or Receive.\n");

      DemoLogf("Press T for Transmit, R for Receive\n");
      c = OsaConsoleGetCh();

      if (c >= 't' || c <= 'T') {
        for(nCnt = 0; nCnt < 16; nCnt++) {
          DemoDumpDataBuffer(0x20000 + (nCnt * 0x400), 32);
        }
      }
      else 
      if (c >= 'r' || c <= 'R') {
        for(nCnt = 0; nCnt < 16; nCnt++) {
          DemoDumpDataBuffer(0x10000 + (nCnt * 0x400), 32);
        }
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case 0xD: // CR
    // No command just return
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds 
  // is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}

/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_DMP:
    // Event from the DMP device
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DMP%d Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DMP%d Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (pEventData->m_nCode == OTX_DMP_EC_LINK_INITIALIZED) {
      g_bDmpLinkInitialized[pEventData->m_nSourceId] = OTX_TRUE;
    }
    break;
    
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_ETHERNET_SWITCH:
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT otxRc;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      otxRc = OtxDrvGetEventDataEx(g_hEventQueue, 
                                   sqEventData, 
                                   DEMO_EVENT_DATA_BUF_SIZE, 
                                   &coEventsFetched); 

      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        // We have a driver event (one at least)

        if(coEventsFetched > DEMO_EVENT_DATA_BUF_SIZE) {
          // Buffer is not that big
          DemoLogf("ERROR: Event fetched %d events buffer holds %d\n", 
                   coEventsFetched,
                   DEMO_EVENT_DATA_BUF_SIZE);

          coEventsFetched = DEMO_EVENT_DATA_BUF_SIZE;
        }
        i = 0;
        while(i < coEventsFetched) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
          i++;
        } // while()
      }
    } while (OTX_RESULT_CODE(otxRc) == OTX_S_OK);
  }

  return 0;
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
    A seperate funcing for burst so that is is less likely
    another event from a different device will block (sleep)
    which would cause the burst device to over/under flow.
*******************************************************************************/
void  *DemoBurstIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT otxRc;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationBurstEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;
      OTX_BOOL bHighwayProcessedR[DEMO_MAX_HWY_NO];
      OTX_BOOL bHighwayProcessedS[DEMO_MAX_HWY_NO];

      otxRc = OtxDrvGetEventDataEx(g_hBurstEventQueue, 
                                   sqEventData, 
                                   DEMO_EVENT_DATA_BUF_SIZE, 
                                   &coEventsFetched); 

      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        // give each highway a chance before processing it again
        for(i=0; i < (OTX_INT32) DEMO_MAX_HWY_NO; i++) {
          if(i < (OTX_INT32) g_demoEnv.coBurstHighways) {
            bHighwayProcessedR[i] = OTX_TRUE;
            bHighwayProcessedS[i] = OTX_TRUE;
          }
          else {
            bHighwayProcessedR[i] = OTX_FALSE;
            bHighwayProcessedS[i] = OTX_FALSE;
          }
        }

        // We have a driver event (one at least)
        if(coEventsFetched > DEMO_EVENT_DATA_BUF_SIZE) {
          // Buffer is not that big
          DemoLogf("ERROR: Event fetched %d events buffer holds %d\n", 
                   coEventsFetched,
                   DEMO_EVENT_DATA_BUF_SIZE);

          coEventsFetched = DEMO_EVENT_DATA_BUF_SIZE;
        }
        i = 0;
        g_FetchCount = coEventsFetched;
        g_FetchCountCur = i;
        while(i < coEventsFetched) {
          switch (sqEventData[i].m_eDeviceType) {
            case OTX_DEVICE_BURST:
              switch (sqEventData[i].m_nCode) {
              case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
                if(bHighwayProcessedR[sqEventData[i].m_nRequestId]) {
                  DemoBurstReceiver(sqEventData[i].m_nRequestId, // Highway
                                    sqEventData[i].m_nParam); // How much to RX
                  // finished that one
                  bHighwayProcessedR[sqEventData[i].m_nRequestId] = OTX_FALSE;
                }
                break;
              case OTX_BURST_SENDER_EC_BUF_DEPLETED:
                if(bHighwayProcessedS[sqEventData[i].m_nRequestId]) {
                  DemoBurstSender(sqEventData[i].m_nRequestId,  // Highway
                                  sqEventData[i].m_nParam); // How much to TX
                  // finished that one
                  bHighwayProcessedS[sqEventData[i].m_nRequestId] = OTX_FALSE;
                }
                break;
              case OTX_BURST_EC_BUFFER_OVERRUN:
                 DemoLogf("Event Burst user buffer over/under run(%d)\n", 
                          sqEventData[i].m_nParam);
                 break;

              default:
                break;
              }
              break;          
            default:
              // Unhandled event
              DemoLogf("Event from an unknown object 0x%X\n", 
                       sqEventData[i].m_hDevice, 
                       sqEventData[i].m_eDeviceType);
              break;
          }
          i++;
          g_FetchCountCur = i;

        } // while()
      }
    } while (OTX_RESULT_CODE(otxRc) == OTX_S_OK);
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  char commandChar;
  OTX_INT32 exitFlag = 0;
  OTX_RESULT otxRc;
  OTX_UINT32  i;
  OTX_UINT32  x;

  g_nH100Direction = DEMO_H100_NOT_CONNECTED; // show that it is not connected

  for (i = 0; i < DEMO_MAX_DMP_NO; i++) {
    g_hDmp[i]   = OTX_INVALID_HANDLE_VALUE;
    g_hLiOut[i] = OTX_INVALID_HANDLE_VALUE;
    g_bDmpLinkInitialized[i] = OTX_FALSE;
	}
  
  for (i = 0; i < DEMO_MAX_BURST_TX; i++) {
    g_hPipes.hBurstIn[i]  = OTX_INVALID_HANDLE_VALUE;
    g_hPipes.hBurstOut[i] = OTX_INVALID_HANDLE_VALUE;
  }

  for (i = 0; i < DEMO_MAX_HWY_NO; i++) {
    // Set each Burst test buffer to a unique value
    for (x = 0; x < (OTX_UINT32) DEMO_MAX_BURST_TX; x++) {
      g_hLogicDevs.mBurstSender[i].sqGammaData[x] = (unsigned char) i + 1;
    }

    // Clear error counters
    g_hLogicDevs.mBurstSender[i].nGammaZeroXferCnt = 0;

    for (x = 0; x < (OTX_UINT32) DEMO_MAX_BURST_TX; x++) {
      g_hLogicDevs.mBurstReceiver[i].nGammaTestErrors[x] = 0;
    }

    for (x = 0; x < (OTX_UINT32) DEMO_MAX_BURST_TX; x++) {
      g_hPipes.hH100In[i][x]  = OTX_INVALID_HANDLE_VALUE;
      g_hPipes.hH100Out[i][x] = OTX_INVALID_HANDLE_VALUE;
    }
	}

  printf("Demo App for Alvis-PCI-Express, %s, Odin TeleSystems Inc.\n", __DATE__);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen("AlvisBasic.log", "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", "AlvisBasic.log");
    OsaConsoleExit();
		exit(-1);
  }

  otxRc = OtxDrvConnectLib(); 
  DemoCheckRc(otxRc);

  if(g_demoEnv.nBoardType == OTX_DEVICE_ALVIS_PCI_EXPRESS) {
    // Create the event queue to be used by the driver to inform the application
    // about events, such as line status changes, or Ethernet switch events, etc.
    // Create a notification Event
    otxRc = OtxDrvCreateEvent(&g_hNotificationEvent);
    if (OTX_RESULT_CODE(otxRc) == OTX_E_NOTIMPL) {
      g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects  
      otxRc = OTX_S_OK;
    }
    DemoCheckRc(otxRc);

    // Create Event Queue for driver notifications
    otxRc = OtxDrvCreateEventQueue(128, 
                                   g_hNotificationEvent, 
                                   &g_hEventQueue); 
    DemoCheckRc(otxRc);

    // Enable Event Queue
    otxRc = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); 
    DemoCheckRc(otxRc);


    // Create a event queue just for burst 
    otxRc = OtxDrvCreateEvent(&g_hNotificationBurstEvent);
    if (OTX_RESULT_CODE(otxRc) == OTX_E_NOTIMPL) {
      g_hNotificationBurstEvent = (OTX_EVENT) NULL;   
      otxRc = OTX_S_OK;
    }
    DemoCheckRc(otxRc);

    // Create Event Queue for driver notifications
    otxRc = OtxDrvCreateEventQueue(128, 
                                   g_hNotificationBurstEvent, 
                                   &g_hBurstEventQueue); 
    DemoCheckRc(otxRc);

    // Enable Event Queue
    otxRc = OtxDrvEnable(g_hBurstEventQueue, OTX_TASK_SYNC,0); 
    DemoCheckRc(otxRc);


    // Open board
    otxRc = OtxDrvOpenPhysicalDevice(0, 
                                     (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, 
                                     0, 
                                     0, 
                                     g_hEventQueue, 
                                     &g_hBoard); 
    DemoCheckRc(otxRc);

    // Open the Ethernet switch
    otxRc = OtxDrvOpenPhysicalDevice(g_hBoard, 
                                     OTX_DEVICE_ETHERNET_SWITCH, 
                                     0, 
                                     0, 
                                     g_hEventQueue, 
                                     &g_hEthSw);
    DemoCheckRc(otxRc);
    DemoLogf("OTX_DEVICE_ETHERNET_SWITCH = 0x%X\n", g_hEthSw);

    // Get a handle to the Time-Space Switch (TSS) on the board
    DemoLogf("Open OTX_DEVICE_TSS\n");
    otxRc = OtxDrvOpenPhysicalDevice(g_hBoard, 
                                     OTX_DEVICE_TSS, 
                                     0, 
                                     0, 
                                     g_hEventQueue, 
                                     &g_hTss); 
    DemoCheckRc(otxRc);

    // Open DMPs
    for (i = 0; i < g_demoEnv.nDmp; i++) {
		  otxRc = OtxDrvOpenPhysicalDevice(g_hBoard, 
                                       OTX_DEVICE_DMP, 
                                       i, 
                                       i, 
                                       OTX_INVALID_HANDLE_VALUE, 
                                       &g_hDmp[i]);
      DemoCheckRc(otxRc);
	  }

    otxRc = OtxDrvEnable(g_hBoard, OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(otxRc);

    otxRc = OtxDrvEnable(g_hEthSw, OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(otxRc);

    for (i = 0; i < g_demoEnv.nDmp; i++) {
      //otxRc = OtxDrvEnable(g_hDmp[i], OTX_TASK_SYNC, 5000L); 
      //DemoCheckRc(otxRc);
    }

    if(g_demoEnv.eClkSource == OTX_CLOCK_SOURCE_INTERNAL) {
      DemoLogf("Set board clock OTX_CLOCK_MODE_H100_ETCF internal\n");
    }
    else
    if(g_demoEnv.eClkSource == OTX_CLOCK_SOURCE_A_CLOCKS_ETCF) {
      DemoLogf("Set board clock OTX_CLOCK_MODE_H100_ETCF from H.100\n");
    }
    else {
      DemoLogf("Unknown board clock 0x%X\n", g_demoEnv.eClkSource);
      OsaConsoleExit();
		  exit(-1);
    }
    otxRc = OtxBrdSetClocks(g_hBoard, 
                            OTX_CLOCK_MODE_H100_ETCF, 
                            g_demoEnv.eClkSource); 
    DemoCheckRc(g_hBoard);

    //
    // Setup Burst device
    //
    // Get a handle to the physical Burst device.
    otxRc = OtxDrvOpenPhysicalDevice(g_hBoard, 
                                     OTX_DEVICE_BURST, 
                                     0, 
                                     0, 
                                     g_hBurstEventQueue, 
                                     &g_hBurst); 
    DemoCheckRc(otxRc);

    DemoLogf("Burst RX FIFO trigger is 0x%x (%d)\n", DEMO_BURST_BUF_SIZE, DEMO_BURST_BUF_SIZE);

    // set device buffer event trigger levels
    otxRc = OtxDrvSetAttributeValueUint32(g_hBurst, 
                                          OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  
                                          DEMO_BURST_BUF_SIZE); 
    DemoCheckRc(otxRc);

    DemoLogf("Burst TX FIFO trigger is 0x%x (%d)\n", DEMO_BURST_BUF_SIZE, DEMO_BURST_BUF_SIZE);

    otxRc = OtxDrvSetAttributeValueUint32(g_hBurst, 
                                          OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    
                                          DEMO_BURST_BUF_SIZE); 
    DemoCheckRc(otxRc);

    // Optionally (mainly for testing, or for improving system performance for single 
    // direction burst applications, disable one direction of the burst device
    if (g_demoEnv.bDisableBurstRx) {
      DemoLogf("Burst RX interrupts will be turned off when the Burst device is enabled\n");
      otxRc = OtxDrvSetAttributeValueBool(g_hBurst, 
                                          OTX_ATTR_BURST_RECV_DISABLED,  
                                          OTX_TRUE); 
      DemoCheckRc(otxRc);
    }
    if (g_demoEnv.bDisableBurstTx) {
      DemoLogf("Burst TX interrupts will be turned off when the Burst device is enabled\n");
      otxRc = OtxDrvSetAttributeValueBool(g_hBurst, 
                                          OTX_ATTR_BURST_SEND_DISABLED,  
                                          OTX_TRUE); 
      DemoCheckRc(otxRc);
    }

    {
      OTX_UINT32 nDrvTxBufSize;
      OTX_UINT32  ui32;
      char *pInitDataBuf;


      //OTX_ATTR_BURST_SENDER_FIFO_SIZE, Host buffer
      otxRc = OtxDrvGetAttributeValueUint32(g_hBurst, 
                                            OTX_ATTR_BURST_SENDER_FIFO_SIZE,
                                            &nDrvTxBufSize); 
      DemoCheckRc(otxRc);

      /*
      ** Prime burst transmit buffer in OTX driver with our data.
      ** This data will be transmitted then the data from the 
      ** input file will be transmitted.
      */
      DemoLogf("Load Burst transmit buffer with 0x12, size 0x%X\n", nDrvTxBufSize);

      pInitDataBuf = (char *) malloc(nDrvTxBufSize);

      memset(pInitDataBuf, 
             0x12, 
             nDrvTxBufSize);

      ui32 = nDrvTxBufSize;
      otxRc = OtxBurstInitSendBuffer(g_hBurst, 
                                     0, 
                                     pInitDataBuf, 
                                     &ui32);
      // Finished with this buffer. 
      // Return it here while we have the chance.
      free(pInitDataBuf);
      DemoCheckRc(otxRc);
    }

    // Set the H.100 Rate (default is 2048 kbps)    
    DemoLogf("Setting rate to %d for all H.100 highways.\n", g_demoEnv.nHbusRate);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_A, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_B, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_C, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_D, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_E, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_F, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_G, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);
    otxRc = OtxDrvSetAttributeValueUint32(g_hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_H, g_demoEnv.nHbusRate); DemoCheckRc(otxRc);

    // Connect to the Physical CT Buses
    DemoLogf("Open OTX_DEVICE_CT_BUS \n");
    otxRc = OtxDrvOpenPhysicalDevice(0, 
                                     OTX_DEVICE_CT_BUS, 
                                     0, 
                                     0, 
                                     g_hEventQueue, 
                                     &g_hCtBus); 
    DemoCheckRc(otxRc);

    // NOTE: need to increase the following attributes when necessary
    // Limit the number of highways that OTX can use on this CT Bus 
    // (only to demonstrate this functionality)
    otxRc = OtxDrvSetAttributeValueUint32(g_hCtBus, OTX_ATTR_CTBUS_OTX_HWYS, 4); 
    DemoCheckRc(otxRc);
    // Limit the number of timeslots that OTX can use on this CT Bus 
    // (only to demonstrate this functionality)
    otxRc = OtxDrvSetAttributeValueUint32(g_hCtBus, OTX_ATTR_CTBUS_OTX_TIMESLOTS_PER_HWY, 32); 
    DemoCheckRc(otxRc);

    otxRc = OtxDrvConnectBoard(g_hCtBus, g_hBoard); 
    DemoCheckRc(otxRc);

    // If we do not close the device the other cards will not be able to 
    // connect to the bus
    OtxDrvClose(g_hCtBus);

    DemoLogf("Enable OTX_DEVICE_TSS\n");
    otxRc = OtxDrvEnable(g_hTss, OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(otxRc);

  }
  else {
    DemoLogf("Error: this demo only supports OTX_DEVICE_ALVIS_PCI_EXPRESS\n");
    DemoHelpPanel();
    OsaConsoleExit();
		exit(1);
  }

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    otxRc = OtxDrvGetAttributeType(g_hBoard, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(otxRc);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    otxRc = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(otxRc);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    otxRc = OtxDrvGetAttributeType(g_hBoard, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(otxRc);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    otxRc = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(otxRc);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    otxRc = OtxDrvGetAttributeType(g_hBoard, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(otxRc);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    otxRc = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(otxRc);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  OsaThreadStart(DemoIdleFunction);
  OsaThreadStart(DemoBurstIdleFunction);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }


  printf("Complete\n");
  DemoBurstEndTest();
  // Close any pipes that may have been created
  DemoDisconnectBurstDevices();

  OtxDrvReleaseLib();

  return 0;
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nNote this demo only supports OTX_DEVICE_ALVIS_PCI_EXPRESS boards\n");
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /N<boardNo>    : Number of DMP devices installed 1-8 (Default %d)\n", DEMO_DEFAULT_DMP_NO);
  DemoLogf(" /Q             : Exit without prompting on errors\n");
  DemoLogf(" /D[T|R]        : Disable one direction of the Burst device\n");
  DemoLogf("                   T = Disable Burst TX\n");
  DemoLogf("                   R = Disable Burst RX\n");
  DemoLogf(" /O[x]        : Clock Source, where\n");
  DemoLogf("                  x = I Internal clock source (default) or\n");
  DemoLogf("                  x = H to recover from H.100 A-clock (ETCF mode)\n");
  DemoLogf(" /R[2048|4096|8192] : Use H100 Rate (default 2048)\n");
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->nBoardType         = OTX_DEVICE_ALVIS_PCI_EXPRESS;
  envP->pszInputStr        = NULL;
  envP->bExitOnError       = OTX_FALSE;
  envP->boardNo            = DEMO_DEFAULT_BOARD_NO;
  envP->nDmp               = DEMO_DEFAULT_DMP_NO;
  envP->eClkSource         = OTX_CLOCK_SOURCE_INTERNAL;

  envP->bDisableBurstTx    = OTX_FALSE; // We want this enable 
  envP->bDisableBurstRx    = OTX_FALSE; // We want this enable 
  envP->nBurstHwyTimeslots = DEMO_MAX_BURST_TX;
  envP->coBurstHighways    = 1;   // H.100 demo currently supports 1 highway DEMO_MAX_HWY_NO;
  envP->nCurrentHwy        = 0;

  // H.100 bus rate must the same for all cards on the chain
  envP->nHbusRate          = 2048;
  envP->nH100HwyRate       = 2048;

  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);

  for (i = 1; i < argc; i++) {   // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'N') {   // DMP number (1-8)
          envP->nDmp = (OTX_INT32)strtol(argv[i]+2, NULL, 0);
          if((envP->nDmp == 0) || (envP->nDmp > DEMO_MAX_DMP_NO)) {
            DemoLogf("Error: DMP value (%d)out of range \n",envP->nDmp);
            DemoHelpPanel();
            OsaConsoleExit();
				    exit(1);
          }
        } else if (*(argv[i]+1) == 'D') {   // Disable one direction of burst
          if (*(argv[i]+2) == 'T') {   // TX
            envP->bDisableBurstTx = OTX_TRUE;
          } else if (*(argv[i]+2) == 'R') {   // T1
            envP->bDisableBurstRx = OTX_TRUE;
          } else {
            DemoLogf("Error: Invalid Burst direction-disabling command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'O') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if ( (*(argv[i]+2) == 'H') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS_ETCF;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'R') {   // Hbus Rate
          envP->nHbusRate = strtol(argv[i]+2, NULL, 0);
          switch (envP->nHbusRate) {
          case 2048:
          case 4096:
          case 8192:
            break; // Acceptable value
          
          default:
            DemoLogf("Error: Invalid Hbus Rate %d. Valid rates are 2048, 4096 and 8192 kbps.\n", envP->nHbusRate);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}

