/* $Workfile: OtxJtagD.h $   $Revision: 7 $    $Modtime: 4/28/08 8:20p $  */
#ifndef OTX_JTAG_DEF_H
#define OTX_JTAG_DEF_H
/*******************************************************************************

  @doc PublicHwApi OtxJtagD

  Copyright (C) 1998 - 2008 Odin TeleSystems Inc., All Rights Reserved

  OTX Generic Hardware Driver API

  @module
  OtxJtagD | Constants, Macros, and Types for the JTAG Functions (Odin internal)

  @contents2
  OtxJtagD.h |

  @index const enum type struct | OtxJtagD

  @contents2
  OTX JTAG Device Macros, Constants, and Types | The JTAG device function are only used for Odin internal functions

  @head3 Enumerated Types  |

  @index enum | OtxJtagD

  @head3 Struct Types |

  @index struct | OtxJtagD

  @head3 Constants |

  @index const | OtxJtagD

  @head3 Other Types |

  @index type | OtxJtagD

*******************************************************************************/

#include "OtxType.h"
#include "OtxErr.h"

#ifdef __cplusplus
  extern "C" {
#endif

#if defined(__linux__)
#pragma pack(1)
#elif defined(__BORLANDC__)
#pragma option push -a1
#else
#pragma pack(push,1)
#endif


/******************************************************************************
  @enum
    Control commands to the TBC device.
    
  @xref
    <f OtxJtagControl>

******************************************************************************/
enum OtxJtagControlCmdE {
  OTX_JTAG_CLK_PULSE      = 0x203, // @emem  Issue a positive pulse on TCLK (Low-High-Low)
  OTX_JTAG_SET_TCLK       = 0x204, // @emem  Set TCLK high
  OTX_JTAG_SET_TMS        = 0x205, // @emem  Set TMS0 high; alias for OTX_JTAG_SET_TMS0
  OTX_JTAG_SET_TDI        = 0x206, // @emem  Set TDI high
  OTX_JTAG_UNSET_TCLK     = 0x207, // @emem  Set TCLK low
  OTX_JTAG_UNSET_TMS      = 0x208, // @emem  Set TMS0 low; alias for OTX_JTAG_UNSET_TMS0
  OTX_JTAG_UNSET_TDI      = 0x209, // @emem  Set TDI low.
  OTX_JTAG_READ_TDO       = 0x20A, // @emem  Read bit value of TDO
  OTX_JTAG_SET_TMS0       = 0x20C, // @emem  Set TMS0 high
  OTX_JTAG_SET_TMS1       = 0x20D, // @emem  Set TMS1 high
  OTX_JTAG_SET_TMS2       = 0x20E, // @emem  Set TMS2 high
  OTX_JTAG_SET_TMS3       = 0x20F, // @emem  Set TMS3 high
  OTX_JTAG_SET_TMS4       = 0x210, // @emem  Set TMS4 high
  OTX_JTAG_SET_TMS5       = 0x211, // @emem  Set TMS5 high
  OTX_JTAG_SET_TMS6       = 0x212, // @emem  Set TMS6 high
  OTX_JTAG_SET_TRST       = 0x213, // @emem  Set TRST high
  OTX_JTAG_UNSET_TMS0     = 0x214, // @emem  Set TMS0 low
  OTX_JTAG_UNSET_TMS1     = 0x215, // @emem  Set TMS1 low
  OTX_JTAG_UNSET_TMS2     = 0x216, // @emem  Set TMS2 low
  OTX_JTAG_UNSET_TMS3     = 0x217, // @emem  Set TMS3 low
  OTX_JTAG_UNSET_TMS4     = 0x218, // @emem  Set TMS4 low
  OTX_JTAG_UNSET_TMS5     = 0x219, // @emem  Set TMS5 low
  OTX_JTAG_UNSET_TMS6     = 0x21A, // @emem  Set TMS6 low
  OTX_JTAG_UNSET_TRST     = 0x21B, // @emem  Set TRST low

  // Special commands (which can be used to reduce the frequency of calls into the driver)
  OTX_JTAG_TCLK_L_H                    = 0x21C, // @emem  First set TCLK low, then set TCLK high
  OTX_JTAG_TDI_H_TCLK_L_H              = 0x21D, // @emem  First set TDI high, then set TCLK low, then set TLCK high
  OTX_JTAG_TDI_L_TCLK_L_H              = 0x21E, // @emem  First set TDI low, then set TCLK low, then set TLCK high
  OTX_JTAG_TDI_H_TCLK_L_RD_TDO_TCLK_H  = 0x21F, // @emem  First set TDI high, then set TCLK low, then read TDO, then set TLCK high
  OTX_JTAG_TDI_L_TCLK_L_RD_TDO_TCLK_H  = 0x220, // @emem  First set TDI low, then set TCLK low, then read TDO, then set TLCK high
  OTX_JTAG_TMS_H_TCLK_L_H              = 0x221, // @emem  First set TMS high (the TMS number that was previously used in a standard set TMS call), then set TCLK low, then set TCLK high
  OTX_JTAG_TMS_L_TCLK_L_H              = 0x222, // @emem  First set TMS low (the TMS number that was previously used in a standard set TMS call), then set TCLK low, then set TCLK high

}; 

typedef enum OtxJtagControlCmdE OtxJtagControlCmdE;


/******************************************************************************
  @enum 
    JTAG Attributes
******************************************************************************/
enum OtxJtagAttributesE {
  OTX_ATTR_JTAG_HM2X20_HIGHZ              = OTX_ATTR_END +  0,  // @emem (OTX_ATTRIB_BOOL) Read/Write. 
                                                                // <nl>If set to OTX_TRUE the Hermod-2x20 board is tri-stated from the Hermod-PCI board.

  OTX_ATTR_JTAG_USE_DELAYS                = OTX_ATTR_END +  1,  // @emem (OTX_ATTRIB_BOOL) Read/Write. Default is OTX_TRUE.
                                                                // <nl>If set to OTX_TRUE the JTAG module will use a 1us delay before each signal transition
                                                                // <nl>If set to OTX_FALSE the JTAG module will issue each transition as quickly as possible
  OTX_ATTR_JTAG_END                       = OTX_ATTR_END +  2
};

typedef enum OtxJtagAttributesE OtxJtagAttributesE;


/******************************************************************************
  @enum
    JTAG Event Codes
******************************************************************************/
enum OtxJtagEventCodesE {
  // None

  OTX_JTAG_END_OF_EVENT_CODES
};

typedef enum OtxJtagEventCodesE OtxJtagEventCodesE;


/******************************************************************************
  @enum
    JTAG Error Result Codes
******************************************************************************/
enum OtxJtagResultCodesE {
  // None 

  OTX_JTAG_RES_END = OTX_RES_END + 0
};

typedef enum OtxJtagResultCodesE OtxJtagResultCodesE;


#if defined(__linux__)
#pragma pack()
#elif defined(__BORLANDC__)
#pragma option pop
#else
#pragma pack(pop)
#endif

#ifdef __cplusplus
  } // extern "C"
#endif


/*******************************************************************************

  Copyright (C) Odin TeleSystems Inc., 1998 - 2008


  All rights reserved. No part of this computer program may be reproduced,
  stored in any retrieval system, or transmitted, in any form or by any
  means, electronic, mechanical, photocopying, recording, or otherwise,
  without the prior written permission of Odin TeleSystems Inc., or without
  a permission stipulated in the agreement/contract under which this computer
  program was supplied.

*******************************************************************************/

#endif // OTX_HDLC_DEF_H
