/* $Workfile: Arni6x6Demo.c $       $Revision: 42 $      $Modtime: 9/17/08 12:36p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    Arni6x6Demo.c

    A basic example console application demonstrating the use of OTX Driver
    with the Arni-6x6-PCI or the Arni-6x6-PCI board.

    The sample application configures 6 FXO POTS Line Interfaces.
    The application provides a simple console interface to manage the POTS 
    Line Interface (LI) devices.
   
    The user can use the following simple keystrokes as commands to go off-hook
    and on-hook, to dial numbers, etc.

    In this basic demo the FXO POTS LIs are supposed to be connected to a telephone 
    line or a telephone line simulator.
    
    The main command keys are listed in the following:
     - Press 'h' or '?' for help. 'q' to quit.
     

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxCst.h"
#include "OtxDsp.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"


/*******************************************************************************
  Macros, Types
*******************************************************************************/

#define DEMO_DSP_MAX 6
#define DEMO_LI_MAX  6

#define DEMO_TASK_REF           1L                      // Task Reference for a non-blocking call
#define DEMO_DEFAULT_LOG_FILE   "Arni6x6Demo.log"       // Name of the LOG file
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_DEFAULT_FXO_NO     0

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k

#define DEMO_U_LAW_SILENCE   0xFF    // Sending this byte produces silence in u-law mode
#define DEMO_A_LAW_SILENCE   0xD5    // Sending this byte produces silence in a-law mode

#define DEMO_HWY_DSP0   0
#define DEMO_HWY_DSP1   1
#define DEMO_HWY_DSP2   2
#define DEMO_HWY_DSP3   3
#define DEMO_HWY_DSP4   4
#define DEMO_HWY_DSP5   5
#define DEMO_HWY_BURST  8

// Demo timeslot (TS) definition.
// Each device connection is assigned two TS to allow for 16 bit data paths.
// the following table is used to assign each connection the base TS.
#define DEMO_TS_DSP_POTS_LI_TX         2   // Input
#define DEMO_TS_DSP_CPT               10   // Input
#define DEMO_TS_DSP_DTMF_DETECTOR     12   // Input
#define DEMO_TS_DSP_VAD               14   // Input
#define DEMO_TS_DSP_CID               16   // Input
#define DEMO_TS_DSP_VOICE_MIX_INPUT0  20   // Input  
#define DEMO_TS_DSP_VOICE_MIX_INPUT1  22   // Input  
#define DEMO_TS_DSP_POTS_LI_RX         0   // Output
#define DEMO_TS_DSP_VOICE_MIX_OUTPUT   2   // Output
#define DEMO_TS_DSP_DTMF_DIALER       20   // Output  

#define DEMO_TS_BURST_REC_DAA          0   // Input to Burst Device
#define DEMO_TS_BURST_RECORD_PB        2   // Input to Burst Device
#define DEMO_TS_BURST_DTMF_DIALER      4   // Input to Burst Device
#define DEMO_TS_BURST_VOICE_MIX        6   // Input to Burst Device
#define DEMO_TS_BURST_PLAYBACK         0   // Output from Burst Device

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char           *logFileName;
  char           *outFileName;
  int             boardNo;
  int             fxoNo;
  OtxPhysicalDeviceTypeE  nBoardType;
  OtxLogicalDeviceTypeE  eDetectorType;
  OtxDataFormatE  law;
  OtxCstPotsCfgSetE  nSetId;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeBurstRecord(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static char DemoDtmfParamToChar(int nParam);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
static int DemoExit(int nExitCode);
static void DemoPrintPresentationMessage(OtxCstCidPresentationMessageS *pCidMsg);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue        = OTX_INVALID_HANDLE_VALUE;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
FILE       *g_fStreamFile = NULL;
OTX_BOOL    g_bWatchDogThreadActive = OTX_FALSE;
OTX_BOOL    g_bKickTheWatchDog = OTX_FALSE;
OTX_BOOL    g_bLineSensingThreadActive = OTX_FALSE;
OTX_BOOL    g_bLineSensePending = OTX_FALSE;
OTX_BOOL    g_bLineSenseVerbose = OTX_FALSE;
OTX_INT32   g_nLineSenseLastValue = -1;
OTX_UINT32  g_nDmaEventCnt = 0;
OTX_BOOL    g_bLibConnected = OTX_FALSE;
OTX_BOOL    g_bConsoleUsed = OTX_FALSE;

static OsaMutexS g_demoLogfCritSec;

struct {  // Handles to Physical Devices
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hDsp[DEMO_LI_MAX];
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
} g_hPhysDevs;


// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
  OTX_BOOL    bUseAlternateSend;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  
};

// Pots Li device
struct sPotsLi {
  OTX_BOOL    bOffHook;
};


struct {  // Handles to Logical Devices
  OTX_HANDLE hPotsLi[DEMO_LI_MAX];
  OTX_HANDLE hVaDetector[DEMO_LI_MAX];
  OTX_HANDLE hCidDetector[DEMO_LI_MAX];
  OTX_HANDLE hCptDetector[DEMO_LI_MAX];
  OTX_HANDLE hDtmfDetector[DEMO_LI_MAX];
  OTX_HANDLE hDtmfDialer[DEMO_LI_MAX];
  OTX_HANDLE hVoiceMixer[DEMO_LI_MAX];
  OTX_HANDLE hUserApp[DEMO_LI_MAX];     // only used for debugging of the DSP program
} g_hLogicDevs;

struct {
  struct sBurstSender mBurstSender;        
  struct sBurstReceiver mBurstReceiver;    
} g_sBurstDevs;

struct {
  struct sPotsLi mPotsLi;        
} g_sPotsDevs[DEMO_LI_MAX];


struct {  // Handles to Pipes
  OTX_HANDLE hPotsRx[DEMO_LI_MAX]; // 128kbps capacity
  OTX_HANDLE hPotsTx[DEMO_LI_MAX]; // 128kbps capacity
  OTX_HANDLE hMixer0[DEMO_LI_MAX]; // 128kbps capacity
  OTX_HANDLE hMixer1[DEMO_LI_MAX]; // 128kbps capacity
  OTX_HANDLE hMixer1a[DEMO_LI_MAX]; // 64kbps capacity
  OTX_HANDLE hMixer1b[DEMO_LI_MAX]; // 64kbps capacity

  // Dummy Pipes used for Logical Device Setup performed from command console
  OTX_HANDLE hDsp2Ts10In[DEMO_LI_MAX];
  OTX_HANDLE hDsp2Ts12In[DEMO_LI_MAX];
  OTX_HANDLE hDsp2Ts14In[DEMO_LI_MAX];
  OTX_HANDLE hDsp2Ts16In[DEMO_LI_MAX];
} g_hPipes;

char *g_sqNumber0 = "201";

/*******************************************************************************

   ==== Global Data setup Universal Multifrequency Tone Detector (UMTD) ====
   These structure defines the creation parameters for the UMTD objects used 
   to build the communication packet for the DSP CPT initialization. These
   structures are used by the OtxCstCptDetectorInitEx() initialization.
   See "Universal Multifrequency Tone Detector (UMTD) Algorithm User's Guide"
   TI document SPRU638, for additional information on UMTD setup.

*******************************************************************************/
#define DEMO_TONE_TSADAPTPAUSE OTX_CST_CPT_DETECTOR_IUMTD_TSADAPTPAUSE
#define DEMO_TONE_TSGUARDTIME  OTX_CST_CPT_DETECTOR_IUMTD_TSGUARDTIME
#define DEMO_TONE_TSPOROSITY   OTX_CST_CPT_DETECTOR_IUMTD_TSPOROSITY
#define DEMO_TONE_TSSIGNALEND  OTX_CST_CPT_DETECTOR_IUMTD_TSSIGNALEND
#define DEMO_TONE_TSREPEAT     OTX_CST_CPT_DETECTOR_IUMTD_TSREPEAT
#define DEMO_TONE_TSQ35BUSY    OTX_CST_CPT_DETECTOR_IUMTD_TSQ35BUSY
#define DEMO_TONE_TSMINTIME    OTX_CST_CPT_DETECTOR_IUMTD_TSMINTIME
#define DEMO_TONE_TSTONESTART  OTX_CST_CPT_DETECTOR_IUMTD_TSTONESTART


#define DEMO_TONE_END_FLAGS DEMO_TONE_TSGUARDTIME | DEMO_TONE_TSSIGNALEND | DEMO_TONE_TSREPEAT

// Frequency planner. Frequencies are in Hz.
OTX_INT16 g_CPTD_frequencies[] =
{
  305, 360, 400, 440, 480, 536
};

#define DEMO_BUSY_T0 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_T1 DEMO_TONE_TSPOROSITY   | DEMO_TONE_TSQ35BUSY   | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_T2 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_T3 DEMO_TONE_TSPOROSITY   | DEMO_TONE_TSQ35BUSY   | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_T4 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_T5 DEMO_TONE_TSPOROSITY   | DEMO_TONE_TSQ35BUSY   | DEMO_TONE_END_FLAGS

// parameters for Q35 busy tone
OTX_UINT16 g_BusyTone_Q35[] =
{
  DEMO_BUSY_T0, 75,  700, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.070 sec - 0.700 sec
  DEMO_BUSY_T1, 10,  800, 0x1111, 0x0011, // pause,           0.010 sec - 0.800 sec
  DEMO_BUSY_T2, 75,  700, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.075 sec - 0.700 sec
  DEMO_BUSY_T3, 10,  800, 0x1111, 0x0011, // pause,           0.010 sec - 0.800 sec
  DEMO_BUSY_T4, 75,  700, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.075 sec - 0.700 sec
  DEMO_BUSY_T5, 10,  800, 0x1111, 0x0011  // pause,           0.010 sec - 0.800 sec
};

#define DEMO_DIAL_T0 DEMO_TONE_TSMINTIME  | DEMO_TONE_TSSIGNALEND | DEMO_TONE_TSREPEAT

// parameters for Q35 Dial tone
OTX_UINT16 g_DialTone_Q35[] =
{
  DEMO_DIAL_T0, 2600, 0x7fff, 0x4443, 0x0034 // 340 Hz - 500 Hz, > 2.6 sec, indicate each time
};

//DEMO_TONE_TSTONESTART report successful 1st frame detection
#define DEMO_RINGBACK_T0 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME | DEMO_TONE_TSTONESTART
#define DEMO_RINGBACK_T1 DEMO_TONE_TSPOROSITY   | DEMO_TONE_END_FLAGS

// parameters for Q35 RingBack tone
OTX_UINT16 g_RingBackTone_Q35[] =
{
  DEMO_RINGBACK_T0,  620,   2550, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.67 sec - 2.5 sec
  DEMO_RINGBACK_T1, 2950,   6050, 0x1111, 0x0011  // pause, 3 sec - 6 sec, indicate once
};

#define DEMO_LONGTONE_T0 DEMO_TONE_TSMINTIME | DEMO_TONE_TSSIGNALEND | DEMO_TONE_TSREPEAT 

// parameters for Long tone
OTX_UINT16 g_LongTone[] = 
{
  DEMO_LONGTONE_T0, 1000, 0x7fff, 0x4443, 0x0034 // 340 Hz - 500 Hz, > 1.0 sec, indicate each time
};

#define DEMO_DIAL_ITALY_T0 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_DIAL_ITALY_T1 DEMO_TONE_TSGUARDTIME
#define DEMO_DIAL_ITALY_T2 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_DIAL_ITALY_T3 DEMO_TONE_END_FLAGS

// parameters for Italian Dial tone
OTX_UINT16 g_DialTone_Italy[] =
{
  DEMO_DIAL_ITALY_T0, 500,  700, 0x4443, 0x0034, // 425 Hz, 0.6 sec
  DEMO_DIAL_ITALY_T1, 900, 1100, 0x1111, 0x0011, // pause,  1.0 sec
  DEMO_DIAL_ITALY_T2, 100,  300, 0x4443, 0x0034, // 425 Hz, 0.2 sec
  DEMO_DIAL_ITALY_T3, 100,  300, 0x1111, 0x0011  // pause,  0.2 sec
};

#define DEMO_RINGBACK_ROBUST_T0 DEMO_TONE_TSADAPTPAUSE
#define DEMO_RINGBACK_ROBUST_T1 DEMO_TONE_TSPOROSITY | DEMO_TONE_END_FLAGS

// parameters for Q35 (robust) RingBack tone
OTX_UINT16 g_RingBackTone_Q35_robust[] =
{
  DEMO_RINGBACK_ROBUST_T0,  620, 2550, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.67 sec - 2.5 sec
  DEMO_RINGBACK_ROBUST_T1, 1950, 6050, 0x1111, 0x0011  // pause, 2 sec - 6 sec, indicate once
}; // pause limits decreased to 2 sec for Japan

#define DEMO_RINGBACK_2PULSE_T0 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_RINGBACK_2PULSE_T1 DEMO_TONE_TSGUARDTIME
#define DEMO_RINGBACK_2PULSE_T2 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_RINGBACK_2PULSE_T3 DEMO_TONE_END_FLAGS

// parameters for RingBack tone (2 pulses)
OTX_UINT16 g_RingBackTone_2pulse[] =
{
  DEMO_RINGBACK_2PULSE_T0,  350,  550, 0x4443, 0x0034, // 350 Hz - 500 Hz, 0.4 sec
  DEMO_RINGBACK_2PULSE_T1,  150,  300, 0x1111, 0x0011, // pause,           0.2 sec
  DEMO_RINGBACK_2PULSE_T2,  350,  550, 0x4443, 0x0034, // 350 Hz - 500 Hz, 0.4 sec
  DEMO_RINGBACK_2PULSE_T3, 1950, 6050, 0x1111, 0x0011  // pause,           2.0 sec
}; // for Great Britain and Singapore

#define DEMO_BUSY_SINGAPORE_T0 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_SINGAPORE_T1 DEMO_TONE_TSPOROSITY   | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_SINGAPORE_T2 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_SINGAPORE_T3 DEMO_TONE_TSPOROSITY   | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_SINGAPORE_T4 DEMO_TONE_TSADAPTPAUSE | DEMO_TONE_TSGUARDTIME
#define DEMO_BUSY_SINGAPORE_T5 DEMO_TONE_TSPOROSITY   | DEMO_TONE_END_FLAGS

// parameters for Singapore Busy tone
OTX_UINT16 g_BusyTone_Singapore[] =
{
  DEMO_BUSY_SINGAPORE_T0, 700, 800, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.750 sec
  DEMO_BUSY_SINGAPORE_T1, 700, 800, 0x1111, 0x0011, // pause,           0.750 sec
  DEMO_BUSY_SINGAPORE_T2, 700, 800, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.750 sec
  DEMO_BUSY_SINGAPORE_T3, 700, 800, 0x1111, 0x0011, // pause,           0.750 sec
  DEMO_BUSY_SINGAPORE_T4, 700, 800, 0x4443, 0x0034, // 340 Hz - 500 Hz, 0.750 sec
  DEMO_BUSY_SINGAPORE_T5, 700, 800, 0x1111, 0x0011  // pause,           0.750 sec
}; // pause and tone duartions are increased (for Singapore)



// Recognized CPTD signals: busy, ringback, dial tone
OtxCstCptIUMTD_SignalS g_CPTD_Q35[] =
{
  OTX_CST_CPT_ADD_UMTD_SIGNAL(OTX_CST_CPT_SIGNAL_ICPTDDET_BUSY,     6, g_BusyTone_Q35),
  OTX_CST_CPT_ADD_UMTD_SIGNAL(OTX_CST_CPT_SIGNAL_ICPTDDET_DIAL,     1, g_DialTone_Q35),
  OTX_CST_CPT_ADD_UMTD_SIGNAL(OTX_CST_CPT_SIGNAL_ICPTDDET_RING,     2, g_RingBackTone_Q35),
  OTX_CST_CPT_ADD_UMTD_SIGNAL(OTX_CST_CPT_SIGNAL_ICPTDDET_LONGTONE, 1, g_LongTone)
};


OtxCstCptIUMTD_SeriesParamsS g_CPTD_series =
{
  g_CPTD_frequencies,                      // pFrequencyList
  50,                                    // Bandwidth, Hz
  {
    (OTX_INT16)(0x8000*0.015),           // (-36.5 dBm) signalThreshold
    (OTX_INT16)(0x8000*0.008),           // (-41.9 dBm) pauseThreshold
    (OTX_INT16)(0x8000*0.250),           // (-12   dBc) PauseAdapt
    (OTX_INT16)(0x8000*0.700),           //             SignalQuality
    (OTX_INT16)(0x8000*0.707),           // (-3    dBc) SpuriousLevel
    0,                                   // spuriousCol
    0,                                   // spuriousRow
    0,                                   // spuriousCol2
    0,                                   // spuriousRow2
    0,                                   // dtmfTwist
    0,                                   // dtmf2harm
    sizeof(g_CPTD_frequencies)/sizeof(OTX_INT16),      // nFrequencies
    sizeof(g_CPTD_Q35)/sizeof(OtxCstCptIUMTD_SignalS), // nSignals
    (OtxCstCptIUMTD_SignalS*)g_CPTD_Q35,               // pSignalList
    NULL
  }
};

// CPT detector configuration
OtxCstCptIUMTD_ProfileS g_IUMTD_CPTDParams =
{
  sizeof(OtxCstCptIUMTD_ProfileS), // size
  1,                               // nDataFormats
  5000,                            // dBm0
  0x2000,                          // MaxLevel
  &g_CPTD_series,
  1,
  10                               // min frame size to process
};
  

/*
**  The "Voice Activity Detector (VAD) Algorithm Users Guide" SPRU635 date 
**  March 2003 was also used to develop the following code definitions.
**  
**  ======== IVAD_Params ========
**  This structure defines the creation parameters for all VAD objects
*/

// Average speech amplitude;  
// signal with higher amplitude will be recognized as speech.
#define DEMO_IVAD_DEFHIGHAMP       16000  // limits (0...32767)  

// Minimum noise amplitude;  
// signal with lower amplitude will be recognized as noise.  
#define DEMO_IVAD_DEFLOWAMP         80  // limits (0...32767)  

// Used for VAD energetic thresholds adjustment. 
// The lower this number is, the higher the probability that noise 
// will be recognized as speech, and vice versa.
#define DEMO_IVAD_DEFQUALITY        -19662 // limits (-32768...32767)  

// Degree of closeness of background noise to broadband white noise; 
// higher value corresponds to non-broadband noise.
#define DEMO_IVAD_DEFSPFLAT         3000  // limits (0...32767)

// Sets the degree of noise;  
// decreasing the filed value corresponds to more stationary noise.
#define DEMO_IVAD_DEFUPDATECOEF     -19662 // limits (-32768...32767)

// Minimum energy update rate; 
// it is used for checking noise adaptation correctness.
#define DEMO_IVAD_DEFUPDATERATE     0x2000 // limits (0...32767)

// Number of frames after speech frames will be marked as  
// IVAD_VOICE_HANGOVER even if they contains noise 
#define DEMO_IVAD_DEFNOISESMOOTH    0  // limits (0...32767)

// Count of voice frames before VAD result is switched to IVAD_VOICE; 
// before this VAD result will be IVAD_NOISE_HANGOVER
#define DEMO_IVAD_DEFSPEECHSMOOTH   0  // limits (0...32767)

// Power of two of frames, used for noise estimation and adaptation; 
// increasing of this value corresponds to decreasing adaptation speed.
#define DEMO_IVAD_DEFINITFRAMES     4   // limits (0...6)

// Modes of VAD by appropriate flags;
#define DEMO_IVAD_DEFFLAGS          -1   // limits (0x0000...0xfff0)

// Number of Linear Prediction Coefficients (LPC);
// accuracy of spectrum envelope to be reproduced by Comfort Noise Generator.
#define DEMO_IVAD_DEFCNGORDER       10  // limits  (0...10)

// Eenergy smoothing for Comfort Noise Generator (CNG).
#define DEMO_IVAD_DEFCNGENERGYCOEF  0x4000 // limits (0...32767)

OtxCstIVAD_Params g_InitVadParms = {
  sizeof(OtxCstIVAD_Params)/sizeof(OTX_UINT16),  // size of structure 
  DEMO_IVAD_DEFHIGHAMP,       // highAmp
  DEMO_IVAD_DEFLOWAMP,        // lowAmp
  DEMO_IVAD_DEFQUALITY,       // quality
  DEMO_IVAD_DEFSPFLAT,        // spFlat
  DEMO_IVAD_DEFUPDATECOEF,    // updateCoef
  DEMO_IVAD_DEFUPDATERATE,    // updateRate
  DEMO_IVAD_DEFNOISESMOOTH,   // noiseSmooth
  DEMO_IVAD_DEFSPEECHSMOOTH,  // speechSmooth
  DEMO_IVAD_DEFINITFRAMES,    // initFrames
  DEMO_IVAD_DEFFLAGS,         // flags
  DEMO_IVAD_DEFCNGORDER,      // cgnOrder
  DEMO_IVAD_DEFCNGENERGYCOEF, // cngEnergyCoef
};



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Opens the Physical Devices on the Arni-16-PCI board

  @comm
    Calls function <f OtxDrvOpenPhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Get Handle to the Physical ARNI Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXO_PCI_PLUS:
    DemoLogf("Opening Arni-6-FXO-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X3_PCI_PLUS:
    DemoLogf("Opening Arni-6x3-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXS_PCI_PLUS:
    DemoLogf("Opening Arni-6-FXS-PCI-Plus board\n");
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", 
             g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  
  // Check if the watchdog timer has triggered before we continue
  {
    OTX_UINT32 readVal;
    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0, 0x04, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (readVal == 0) {
      DemoLogf("The watchdog timer has triggered. A reboot of the PC is necessary.\n");
      DemoExitQuestion();
    }
  }

  // Disable the Watchdog Timer, and enable it when we have completed the configuration
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_FALSE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Get a handle to the Time-Space Switch (TSS) on the board (just for debugging purposes)
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hBoardController0, result);


  for (i = g_demoEnv.fxoNo; i < DEMO_DSP_MAX; i++) {
    DemoLogf("Open OTX_DEVICE_DSP #%d\n", i);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  for (i = g_demoEnv.fxoNo; i < DEMO_DSP_MAX; i++) {
    if (g_demoEnv.outFileName) {
      DemoLogf("Loading program file %s to DSP#%d \n", g_demoEnv.outFileName, i);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], g_demoEnv.outFileName, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    } else {
      DemoLogf("Loading OTX_DSP_SPM_OTXCST2 to DSP#%d \n", i);
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXCST2); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnablePhysicalDevices\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = g_demoEnv.fxoNo; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
    
}


void DemoCreateLogicalDevices(void)
/*******************************************************************************
  @func
    Create the needed Logical Devices

  @comm
    The logical devices are different from the phyiscal devices in that the
    the number of logical devices is not fixed. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  nDsp;
  OTX_INT32  nLi;

  for (nLi = g_demoEnv.fxoNo; nLi < DEMO_LI_MAX; nLi++) {
    
    nDsp = nLi;
    
    DemoLogf("Creating Logical Devices on Dsp%d\n", nDsp);
    
    if (g_demoEnv.outFileName) {
      // For debug prints
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nLi], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, nLi, &g_hLogicDevs.hUserApp[nLi]); 
      //DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);
      DemoLogf("Created User App device [%d], Handle = 0x%X (for debug printfs)\n", nLi, g_hLogicDevs.hUserApp[nLi]);
    }
    
    // Create Pots Li Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_LI_POTS #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_LI_POTS, g_hEventQueue, nLi, &g_hLogicDevs.hPotsLi[nLi]); DemoCheckRc(g_hLogicDevs.hPotsLi[nLi], result);

    // Create Voice Activity Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hVaDetector[nLi]); DemoCheckRc(g_hLogicDevs.hVaDetector[nLi], result);

    // Creatr CID Detector devices
    //DemoLogf("Creating CID Detector device [%d]\n", i);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_CALLER_ID_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hCidDetector[nLi]); DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);

    // Create Cpt Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_CPT_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_CPT_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hCptDetector[nLi]); DemoCheckRc(g_hLogicDevs.hCptDetector[nLi], result);

    // Create Dtmf Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_DTMF_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_DTMF_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hDtmfDetector[nLi]); DemoCheckRc(g_hLogicDevs.hDtmfDetector[nLi], result);

    // Create Voice Mixer Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_VOICE_MIXER #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_VOICE_MIXER, g_hEventQueue, nLi, &g_hLogicDevs.hVoiceMixer[nLi]); DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);

    
    // Create DTMF Dialers Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_DTMF_DIALER #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_DTMF_DIALER, g_hEventQueue, nLi, &g_hLogicDevs.hDtmfDialer[nLi]); DemoCheckRc(g_hLogicDevs.hDtmfDialer[nLi], result);
    
  } 
  
}


void DemoConnectDevices(void)
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset
    of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a 
    characterized but its capacity (in 8 kbit/s increments). The user can create 
    a pipe with the driver function <f OtxDrvCreatePipe>. A pipe is connected to 
    physical or logical devices and mapped to specific time-slots with the 
    driver function <f OtxDrvConnectPipe>.
*******************************************************************************/
{
  OTX_INT32 nDsp;
  OTX_INT32 nTs;
  OTX_INT32 nHwy;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];


  // Create the Pipes (one for each direction) between the Burst Device and the DSPs
  for (nDsp = g_demoEnv.fxoNo; nDsp < DEMO_LI_MAX; nDsp++) {

    // Create pipes (one for each direction)
    DemoLogf("Creating and connection pipes\n");

    // -------------------------
    // Connect pipe: Dsp -> Dsp (DTMF dialer)
    // -------------------------
    result = OtxDrvCreatePipe(128, &g_hPipes.hPotsTx[nDsp]);
    DemoCheckRc(g_hPipes.hPotsTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_POTS_LI_TX;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_POTS_LI_TX + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_POTS_LI_TX;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_POTS_LI_TX + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsTx[nDsp], result);

    // -------------------------
    // Connect pipe: (Daa ->) Dsp --> Dsp (POTS Li)
    // -------------------------
    result = OtxDrvCreatePipe(128, &g_hPipes.hPotsRx[nDsp]);
    DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_POTS_LI_RX;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_POTS_LI_RX + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_POTS_LI_RX;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_POTS_LI_RX + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[nDsp], OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);


    // -------------------------
    // Connect pipe: Dsp --> Dsp (Voice Mixer 0 input)
    // -------------------------
    result = OtxDrvCreatePipe(128, &g_hPipes.hMixer0[nDsp]);
    DemoCheckRc(g_hPipes.hMixer0[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT0;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT0 + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer0[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer0[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT0;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT0 + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer0[nDsp], OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer0[nDsp], result);


    // -------------------------
    // Connect pipe: Dsp --> Dsp (Voice Mixer 1 input)
    // -------------------------
    result = OtxDrvCreatePipe(128, &g_hPipes.hMixer1[nDsp]);
    DemoCheckRc(g_hPipes.hMixer1[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1 + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1;
    sqTimeSlotMask[nTs] = 0xFF;
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1 + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1[nDsp], OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1[nDsp], result);


    // -------------------------
    // Connect pipe: Burst -> Dsp (Voice Mixer 1 input to burst device)
    // -------------------------
    result = OtxDrvCreatePipe(64, &g_hPipes.hMixer1a[nDsp]);
    DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_PLAYBACK;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1;
    sqTimeSlotMask[nTs] = 0xFF;
    nHwy = nDsp;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

    result = OtxDrvCreatePipe(64, &g_hPipes.hMixer1b[nDsp]);
    DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_PLAYBACK + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1 + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

  }

  for (nDsp = g_demoEnv.fxoNo; nDsp < DEMO_LI_MAX; nDsp++) {

    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------


    // Create the following Connections in receive direction:
    //   Pots Li  --> Pipe --->  DtmfDetector 
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hPotsLi[nDsp], -1, g_hPipes.hPotsRx[nDsp], -1); DemoCheckRc(g_hLogicDevs.hPotsLi[nDsp], result);
    switch (g_demoEnv.eDetectorType) {
    case OTX_LDEVICE_CST_CPT_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hCptDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    case OTX_LDEVICE_CST_DTMF_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hDtmfDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    case OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hVaDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    case OTX_LDEVICE_CST_CALLER_ID_DETECTOR:
      result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsRx[nDsp], -1, g_hLogicDevs.hCidDetector[nDsp], -1); DemoCheckRc(g_hPipes.hPotsRx[nDsp], result);
      break;
    default:
      DemoLogf("Error: Unknown detector type configured\n");
      break;
    }

    // Make the following connections in transmit direction:
    //   DtmfDialer  --> Pipe --> Pots Li
    result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsTx[nDsp], -1, g_hLogicDevs.hPotsLi[nDsp], -1); DemoCheckRc(g_hPipes.hPotsTx[nDsp], result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hDtmfDialer[nDsp], -1, g_hPipes.hMixer0[nDsp], -1); DemoCheckRc(g_hLogicDevs.hDtmfDialer[nDsp], result);

    // Logically connect the one output from the Voice Mixer (to the DAA tx pipe)
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hVoiceMixer[nDsp], -1, g_hPipes.hPotsTx[nDsp], -1); DemoCheckRc(g_hLogicDevs.hDtmfDialer[nDsp], result);

    // Logically connect the two inputs to the Voice Mixer
    result = OtxDrvConnectLogicalDevice(g_hPipes.hMixer0[nDsp], -1, g_hLogicDevs.hVoiceMixer[nDsp], 0); DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes.hMixer1[nDsp], -1, g_hLogicDevs.hVoiceMixer[nDsp], 1); DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);




    // Create connection using Dummy Pipe connections for logical
    // devices that will be reconfigured by user from console command.
    // 
    // Setup CPT
    if(g_demoEnv.eDetectorType !=  OTX_LDEVICE_CST_CPT_DETECTOR) {
      result = OtxDrvCreatePipe(64, &g_hPipes.hDsp2Ts10In[nDsp]);
      DemoCheckRc(g_hPipes.hDsp2Ts10In[nDsp], result);

      memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
      nTs = DEMO_TS_DSP_CPT;
      sqTimeSlotMask[nTs] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hDsp2Ts10In[nDsp], 
                                 OTX_PIPE_OUTPUT, 
                                 g_hPhysDevs.hDsp[nDsp],   
                                 -1,         
                                 sqTimeSlotMask, 
                                 32);
      DemoCheckRc(g_hPipes.hDsp2Ts10In[nDsp], result);


      result = OtxDrvConnectLogicalDevice(g_hPipes.hDsp2Ts10In[nDsp], 
                                          -1, 
                                          g_hLogicDevs.hCptDetector[nDsp], 
                                          -1); 
      DemoCheckRc(g_hPipes.hDsp2Ts10In[nDsp], result); 
    }
    else {
      // Not used at this time.
      g_hPipes.hDsp2Ts10In[nDsp] = OTX_INVALID_HANDLE_VALUE;
    }

    // Setup DTMF Detector
    if(g_demoEnv.eDetectorType !=  OTX_LDEVICE_CST_DTMF_DETECTOR) {
      result = OtxDrvCreatePipe(64, &g_hPipes.hDsp2Ts12In[nDsp]);
      DemoCheckRc(g_hPipes.hDsp2Ts12In[nDsp], result);

      memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
      nTs = DEMO_TS_DSP_DTMF_DETECTOR;
      sqTimeSlotMask[nTs] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hDsp2Ts12In[nDsp], 
                                 OTX_PIPE_OUTPUT, 
                                 g_hPhysDevs.hDsp[nDsp],   
                                 -1,         
                                 sqTimeSlotMask, 
                                 32);
      DemoCheckRc(g_hPipes.hDsp2Ts12In[nDsp], result);


      result = OtxDrvConnectLogicalDevice(g_hPipes.hDsp2Ts12In[nDsp], 
                                          -1, 
                                          g_hLogicDevs.hDtmfDetector[nDsp], 
                                          -1); 
      DemoCheckRc(g_hPipes.hDsp2Ts12In[nDsp], result); 
    }
    else {
      // Not used at this time.
      g_hPipes.hDsp2Ts12In[nDsp] = OTX_INVALID_HANDLE_VALUE;
    }

    // Setup VAD
    if(g_demoEnv.eDetectorType !=  OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR) {
      result = OtxDrvCreatePipe(64, &g_hPipes.hDsp2Ts14In[nDsp]);
      DemoCheckRc(g_hPipes.hDsp2Ts14In[nDsp], result);

      memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
      nTs = DEMO_TS_DSP_VAD;
      sqTimeSlotMask[nTs] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hDsp2Ts14In[nDsp], 
                                 OTX_PIPE_OUTPUT, 
                                 g_hPhysDevs.hDsp[nDsp],   
                                 -1,         
                                 sqTimeSlotMask, 
                                 32);
      DemoCheckRc(g_hPipes.hDsp2Ts14In[nDsp], result);


      result = OtxDrvConnectLogicalDevice(g_hPipes.hDsp2Ts14In[nDsp], 
                                          -1, 
                                          g_hLogicDevs.hVaDetector[nDsp], 
                                          -1); 
      DemoCheckRc(g_hPipes.hDsp2Ts14In[nDsp], result); 
    }
    else {
      // Not used at this time.
      g_hPipes.hDsp2Ts14In[nDsp] = OTX_INVALID_HANDLE_VALUE;
    }
    
    // Setup CID device
    if(g_demoEnv.eDetectorType !=  OTX_LDEVICE_CST_CALLER_ID_DETECTOR) {
      result = OtxDrvCreatePipe(64, &g_hPipes.hDsp2Ts16In[nDsp]);
      DemoCheckRc(g_hPipes.hDsp2Ts16In[nDsp], result);

      memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
      nTs = DEMO_TS_DSP_CID;
      sqTimeSlotMask[nTs] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hDsp2Ts16In[nDsp], 
                                 OTX_PIPE_OUTPUT, 
                                 g_hPhysDevs.hDsp[nDsp],   
                                 -1,         
                                 sqTimeSlotMask, 
                                 32);
      DemoCheckRc(g_hPipes.hDsp2Ts16In[nDsp], result);


      result = OtxDrvConnectLogicalDevice(g_hPipes.hDsp2Ts16In[nDsp], 
                                          -1, 
                                          g_hLogicDevs.hCidDetector[nDsp], 
                                          -1); 
      DemoCheckRc(g_hPipes.hDsp2Ts16In[nDsp], result); 
    }
    else {
      // Not used at this time.
      g_hPipes.hDsp2Ts16In[nDsp] = OTX_INVALID_HANDLE_VALUE;
    }
  }

  // Used to verify data input
  DemoInitializeBurstRecord();
}


void DemoInitializeBurstRecord(void)
/*******************************************************************************
  @func
    Initialize the pipes needed to connect the burst device for data record
    of the input signals.
*******************************************************************************/
{
  OTX_INT32 nHwy;
  OTX_INT32 nTs;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];


  static OTX_HANDLE hPipeRecDaa0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeRecDaa1 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDbgVoiceMixer0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDbgVoiceMixer1 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDbgDialer0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDbgDialer1 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDbgRecordPb0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDbgRecordPb1 = OTX_INVALID_HANDLE_VALUE;

  DemoLogf("Active FXO device %d\n", g_demoEnv.fxoNo);

  // Check to see if pipes are already being used.
  if(hPipeRecDaa0 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeRecDaa0, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeRecDaa0, result);

    result = OtxDrvDisconnectPipe(hPipeRecDaa0, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeRecDaa0, result);

    hPipeRecDaa0 = OTX_INVALID_HANDLE_VALUE;
  }

  if(hPipeRecDaa1 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeRecDaa1, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeRecDaa1, result);

    result = OtxDrvDisconnectPipe(hPipeRecDaa1, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeRecDaa1, result);

    hPipeRecDaa1 = OTX_INVALID_HANDLE_VALUE;
  }

  // Check to see if pipes are already being used.
  if(hPipeDbgVoiceMixer0 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeDbgVoiceMixer0, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeDbgVoiceMixer0, result);

    result = OtxDrvDisconnectPipe(hPipeDbgVoiceMixer0, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeDbgVoiceMixer0, result);

    hPipeDbgVoiceMixer0 = OTX_INVALID_HANDLE_VALUE;
  }

  if(hPipeDbgVoiceMixer1 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeDbgVoiceMixer1, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeDbgVoiceMixer1, result);

    result = OtxDrvDisconnectPipe(hPipeDbgVoiceMixer1, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeDbgVoiceMixer1, result);

    hPipeDbgVoiceMixer1 = OTX_INVALID_HANDLE_VALUE;
  }

  // Check to see if pipes are already being used.
  if(hPipeDbgDialer0 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeDbgDialer0, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeDbgDialer0, result);

    result = OtxDrvDisconnectPipe(hPipeDbgDialer0, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeDbgDialer0, result);

    hPipeDbgDialer0 = OTX_INVALID_HANDLE_VALUE;
  }

  if(hPipeDbgDialer1 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeDbgDialer1, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeDbgDialer1, result);

    result = OtxDrvDisconnectPipe(hPipeDbgDialer1, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeDbgDialer1, result);

    hPipeDbgDialer1 = OTX_INVALID_HANDLE_VALUE;
  }

  // Check to see if pipes are already being used.
  if(hPipeDbgRecordPb0 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeDbgRecordPb0, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeDbgRecordPb0, result);

    result = OtxDrvDisconnectPipe(hPipeDbgRecordPb0, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeDbgRecordPb0, result);

    hPipeDbgRecordPb0 = OTX_INVALID_HANDLE_VALUE;
  }

  if(hPipeDbgRecordPb1 != OTX_INVALID_HANDLE_VALUE) {
    // Need to close this connection befor setting up new pipe connection
    result = OtxDrvDisconnectPipe(hPipeDbgRecordPb1, OTX_PIPE_INPUT); 
    DemoCheckRc(hPipeDbgRecordPb1, result);

    result = OtxDrvDisconnectPipe(hPipeDbgRecordPb1, OTX_PIPE_OUTPUT); 
    DemoCheckRc(hPipeDbgRecordPb1, result);

    hPipeDbgRecordPb1 = OTX_INVALID_HANDLE_VALUE;
  }



#if 1 // For additional debugging
  {
    
    // Also connect the DAA samples to the burst (TS0) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeRecDaa0);
    DemoCheckRc(hPipeRecDaa0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = g_demoEnv.fxoNo;
    nTs = DEMO_TS_DSP_POTS_LI_RX;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa0, OTX_PIPE_INPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = DEMO_HWY_BURST;
    nTs = DEMO_TS_BURST_REC_DAA;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa0, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa0, result);

    result = OtxDrvCreatePipe(64, &hPipeRecDaa1);
    DemoCheckRc(hPipeRecDaa1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = g_demoEnv.fxoNo;
    nTs = DEMO_TS_DSP_POTS_LI_RX + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa1, OTX_PIPE_INPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = DEMO_HWY_BURST;
    nTs = DEMO_TS_BURST_REC_DAA + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa1, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa1, result);  
  }

  {    
    // Also connect the output from the Voice Mixer samples to the burst (TS6) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeDbgVoiceMixer0);
    DemoCheckRc(hPipeDbgVoiceMixer0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = g_demoEnv.fxoNo;
    nTs = DEMO_TS_DSP_VOICE_MIX_OUTPUT;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgVoiceMixer0, OTX_PIPE_INPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgVoiceMixer0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = DEMO_HWY_BURST;
    nTs = DEMO_TS_BURST_VOICE_MIX;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgVoiceMixer0, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgVoiceMixer0, result);

    result = OtxDrvCreatePipe(64, &hPipeDbgVoiceMixer1);
    DemoCheckRc(hPipeDbgVoiceMixer1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = g_demoEnv.fxoNo;
    nTs = DEMO_TS_DSP_VOICE_MIX_OUTPUT + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgVoiceMixer1, OTX_PIPE_INPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgVoiceMixer1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nHwy = DEMO_HWY_BURST;
    nTs = DEMO_TS_BURST_VOICE_MIX + 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgVoiceMixer1, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgVoiceMixer1, result);  
  }

  {    
    // Also connect the output from the DTMF Dialer samples to the burst (TS4) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeDbgDialer0);
    DemoCheckRc(hPipeDbgDialer0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_DTMF_DIALER;
    nHwy = g_demoEnv.fxoNo;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgDialer0, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgDialer0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_DTMF_DIALER;
    nHwy = DEMO_HWY_BURST;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgDialer0, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgDialer0, result);

    result = OtxDrvCreatePipe(64, &hPipeDbgDialer1);
    DemoCheckRc(hPipeDbgDialer1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_DSP_DTMF_DIALER + 1;
    nHwy = g_demoEnv.fxoNo;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgDialer1, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgDialer1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_DTMF_DIALER + 1;
    nHwy = DEMO_HWY_BURST;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgDialer1, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgDialer1, result);
  
  }

  {    
    // Also connect the output from the playback file back to the burst (TS2) via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeDbgRecordPb0);
    DemoCheckRc(hPipeDbgRecordPb0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_PLAYBACK;
    nHwy = DEMO_HWY_BURST;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgRecordPb0, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgRecordPb0, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_RECORD_PB;
    nHwy = DEMO_HWY_BURST;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgRecordPb0, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgRecordPb0, result);

    result = OtxDrvCreatePipe(64, &hPipeDbgRecordPb1);
    DemoCheckRc(hPipeDbgRecordPb1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_PLAYBACK + 1;
    nHwy = DEMO_HWY_BURST;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgRecordPb1, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgRecordPb1, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = DEMO_TS_BURST_RECORD_PB + 1;
    nHwy = DEMO_HWY_BURST;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeDbgRecordPb1, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeDbgRecordPb1, result);  
  }
#endif


}


void DemoInitializeLogicalDevices(void)
/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoInitializeLogicalDevices\n");


  DemoLogf("Configuring Lis for country set=");
  switch (g_demoEnv.nSetId) {
  case OTX_CST_POTS_CFG_SET_USA:
    DemoLogf("USA");
    break;
  case OTX_CST_POTS_CFG_SET_GERMANY: 
    DemoLogf("Germany");
    break;
  case OTX_CST_POTS_CFG_SET_UK:
    DemoLogf("United Kingdom");
    break;
  case OTX_CST_POTS_CFG_SET_FRANCE:
    DemoLogf("France");
    break;
  case OTX_CST_POTS_CFG_SET_AUSTRALIA:
    DemoLogf("Australia");
    break;
  case OTX_CST_POTS_CFG_SET_JAPAN:
    DemoLogf("Japan");
    break;
  default:
    DemoLogf("UNKNOWN");
    break;
  }
  DemoLogf("\n");  

  for (i = g_demoEnv.fxoNo; i < DEMO_LI_MAX; i++) {

    // Initialize the POTS LI logical devices
    result = OtxCstPotsInit(
      g_hLogicDevs.hPotsLi[i],
      g_demoEnv.law,
      g_demoEnv.nSetId
      );
    DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);


#if 0 
    // We can use either a simple Init function for the  Voice Activity Detector 
    // (OtxCstVoiceActivityDetectorInit()), where we do not specify the input 
    // parameters or...

    // Initialize the Voice Activity Detector
    result = OtxCstVoiceActivityDetectorInit(
      g_hLogicDevs.hVaDetector[i], 
      g_demoEnv.law
      ); 
    DemoCheckRc(g_hLogicDevs.hVaDetector[i], result);
#else
    // (OtxCstVoiceActivityDetectorInitEx()), where the input parameters are 
    // defined by the user.

    result = OtxCstVoiceActivityDetectorInitEx(
      g_hLogicDevs.hVaDetector[i],
      &g_InitVadParms,
      g_demoEnv.law
      ); 
    DemoCheckRc(g_hLogicDevs.hVaDetector[i], result);

#endif

    // Initialize the CID Detector
    result = OtxCstCidDetectorInit(
      g_hLogicDevs.hCidDetector[i], 
      g_demoEnv.law, 
      OTX_CST_CID_DETECTOR_STD_CFG
      ); 
    DemoCheckRc(g_hLogicDevs.hCidDetector[i], result);

#if 0 
    // We can use either a simple Init function for the CPT Detector 
    // (OtxCstCptDetectorInit()), where we do not specify the input 
    // parameters or...

    // Initialize the CPT Detector
    result = OtxCstCptDetectorInit(
      g_hLogicDevs.hCptDetector[i], 
      g_demoEnv.law, 
      OTX_TRUE,
      OTX_CST_CPT_DETECTOR_Q35_CFG
      ); 
    DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

#else
    // (OtxCstCptDetectorInitEx()), where the input parameters are 
    // defined by the user.

    // Initialize the CPT Detector
    result = OtxCstCptDetectorInitEx(
      g_hLogicDevs.hCptDetector[i], 
      g_demoEnv.law, 
      OTX_TRUE,
      &g_IUMTD_CPTDParams,
      OTX_CST_CPT_DETECTOR_Q35_CFG
      ); 
    DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

#endif
    // Initialize the DTMF Detector
    result = OtxCstDtmfDetectorInit(
      g_hLogicDevs.hDtmfDetector[i], 
      g_demoEnv.law, 
      OTX_TRUE); 
    DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);

    // Initialize the DTMF Dialers
    result = OtxCstDtmfDialerInit(
      g_hLogicDevs.hDtmfDialer[i], 
      g_demoEnv.law 
      ); 
    DemoCheckRc(g_hLogicDevs.hDtmfDialer[i], result);

    // Initialize the Voice Mixer
#if 0 // We can use either a simple Init function for the Voice Mixer (OtxCstVoiceMixerInit()),
      // where we do not specify the gains settings for each input or...
    result = OtxCstVoiceMixerInit(
      g_hLogicDevs.hVoiceMixer[i], 
      g_demoEnv.law, 
      2); 
#else // ... we can use a more advanced Init function for the Voice Mixer where we provide the 
      // gain setting for each input and the output.
    {
      OtxDataFormatE sqInDataFormats[2] = { g_demoEnv.law, g_demoEnv.law }; // Data Format of inputs 
      OtxDataFormatE eOutDataFormat     = g_demoEnv.law; // Data Format of the output
      OTX_INT16 sqInDigGains[2]         = { 0 /* DTMF */, 0 /* Burst */ }; // Digital gain of the inputs (-1=-3dB, 0=0dB, 1=6dB, etc)
      OTX_INT16 nOutDigGain             = 0; // Digital gain of the output (-1=-3dB, 0=0dB, 1=6dB, etc).
      
      result = OtxCstVoiceMixerInitEx(
        g_hLogicDevs.hVoiceMixer[i], 
        sqInDataFormats,
        eOutDataFormat,
        sqInDigGains,
        nOutDigGain,
        2); 
    }
#endif
    DemoCheckRc(g_hLogicDevs.hVoiceMixer[i], result);
  }
}


void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = g_demoEnv.fxoNo; i < DEMO_LI_MAX; i++) {
    DemoLogf("In DemoEnableLogicalDevices DSP%d\n", i);

    result = OtxDrvEnable(g_hLogicDevs.hPotsLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hDtmfDetector[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hCidDetector[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hCidDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hCptDetector[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hVaDetector[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hVaDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hDtmfDialer[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialer[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hVoiceMixer[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hVoiceMixer[i], result);
  }

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX [%d.%d])\n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result));
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


void DemoExitQuestion(void)
/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    if(g_bConsoleUsed == OTX_FALSE) {
      DemoLogf(" \n");
      DemoLogf("Do you want to exit (Y/N) ");
      c = OsaConsoleGetCh();
    }
    else {
      // Some other task is using the console
      c = 'n'; // Do not exit at this time. Let user select quit option.
      DemoLogf("\nError found in result, select q to exit\n");
    }
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    DemoExit(1);
  }
  else {
    DemoLogf(" \n");
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Function that keeps reseting the Watchdog timer on the board
*******************************************************************************/
void  *DemoWatchDogFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_BOOL bEnabled;

  g_bWatchDogThreadActive = OTX_TRUE;
  g_bKickTheWatchDog = OTX_TRUE;

  DemoLogf("DemoWatchDogFunction Thread started\n");

  while (g_bWatchDogThreadActive) {

    result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  &bEnabled); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (!bEnabled) {
      DemoLogf("*********** Watchdog Timer timed out. The PC must be rebooted *********\n");
      OsaThreadSleep(3000); // in milliseconds
      continue;
    }

    if (g_bKickTheWatchDog) {
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_RESET,  OTX_TRUE /* Any value will work */); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      OsaThreadSleep(100); // in milliseconds
    }
  }

  DemoLogf("DemoWatchDogFunction Thread terminating\n");
  return 0;

}

/*******************************************************************************
  @func
    Function that keeps monitoring the Loop current for Far-End-Hangup conditions
*******************************************************************************/
void  *DemoLineSensingFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32  nCurrentLi = g_demoEnv.fxoNo;

  g_bLineSensingThreadActive = OTX_TRUE;

  DemoLogf("DemoLineSensingFunction Thread started\n");
  while (g_bLineSensingThreadActive) {

    nCurrentLi = g_demoEnv.fxoNo;

    if (!g_bLineSensePending) {
      result = OtxCstPotsSenseLine(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1006, 300); 
      if (OTX_SUCCESS(result)) {
        // Waiting for line sense status from DSP
        g_bLineSensePending = OTX_TRUE;
      }

    }

    OsaThreadSleep(100); // in milliseconds
  }

  DemoLogf("DemoLineSensingFunction Thread terminating\n");
  return 0;

}



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}


void DemoPrintCSTEvent(OtxEventDataS *pEventData) 
/*******************************************************************************
  @func
    Only used by the USER_APPLICATION device for debugging of DSP program code.
*******************************************************************************/
{
  OTX_RESULT otxRc;
  char msgStr[256];
  
  if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
    OTX_UINT8  buf[8192];
    OTX_UINT32 nBytesReturned;
    OTX_UINT32 nIoControlCode;      
    
    otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8192, &nBytesReturned, &nIoControlCode);
    if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      switch(nIoControlCode) {
      case 0x2ff:
        {
          char strBuf[1024];
          int k;
          sprintf(strBuf, "Len=%d :",nBytesReturned);
          for(k=0;k<(int)nBytesReturned;k+=2){
            sprintf(&strBuf[0] + strlen(strBuf), "%c",buf[k]);
          }
          sprintf(&strBuf[0] + strlen(strBuf), "\n");
          DemoLogf(strBuf);
        }
        break;
        
      case 0x0200 /*DEMO1_NOTIFY_DEBUG_DATA*/:
        {
          DemoLogf("DEMO1_NOTIFY_DEBUG_DATA (nRequestId=%d, nBytes=%d)\n", pEventData->m_nRequestId, nBytesReturned);
        }
        break;

      case 0x02FE:
        {
          static int counter = 0;

          if (counter++ % 5000 == 0) {
            DemoLogf("Got data\n");
          }

          if (g_fStreamFile) {
            fwrite((OTX_UINT8 *)buf, nBytesReturned, sizeof(OTX_UINT8), g_fStreamFile);
          }

        }
        break;
      default:
        DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
        break;
      }
    } else {
      OTX_RESULT errorCode = otxRc;
      otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
      }
      
    }
    
  } else {
    
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
  }
}




void DemoBurstSender(OTX_UINT32 nSourceId)
/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
{
  OTX_RESULT  otxRc;
  OTX_UINT32  txBufSize;
  OTX_UINT32  txItemSize;
  OTX_UINT32  ui32;
  char nSilence = 0;


  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstSender.bActive) {
    // we're not active, so just return
    return;
  }

  if (g_sBurstDevs.mBurstSender.pTxFile) {

    txItemSize = sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]);
    txBufSize = sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / txItemSize;

    if (g_sBurstDevs.mBurstSender.nTsToSend == -1) {
      // All TS
      g_sBurstDevs.mBurstSender.cbTxBuf = fread(g_sBurstDevs.mBurstSender.sqTxBuf,
                                                txItemSize,
                                                txBufSize,
                                                g_sBurstDevs.mBurstSender.pTxFile);

    } 
    else {
      /*
      ** Prime the transmit buffer with silence. 
      ** Needed for single TS transmission.
      */
      switch (g_demoEnv.law) {
        case OTX_DATA_FORMAT_COMPAND8_A_LAW:
          nSilence = (char) DEMO_A_LAW_SILENCE; // silence in a-law mode
          break;                         
        case OTX_DATA_FORMAT_COMPAND8_U_LAW:
          nSilence = (char) DEMO_U_LAW_SILENCE; // silence in u-law mode
          break;
        case OTX_DATA_FORMAT_UINT16:
          nSilence = 0; // silence in Linear data mode
          break;
        default:
          nSilence = (char) DEMO_U_LAW_SILENCE; // silence in u-law mode
          DemoLogf("Warning: g_demoEnv.law not configured using default\n");
            break;
      }
      memset(&g_sBurstDevs.mBurstSender.sqTxBuf[0], 
             nSilence, 
             txBufSize);

      if(g_demoEnv.law == OTX_DATA_FORMAT_UINT16) {
        // Two TS are needed for OTX_DATA_FORMAT_UINT16
        OTX_UINT8 doubleTsBuf[DEMO_BURST_BUF_SIZE / 16];
        int cbDoubleTsBuf;
        int byteCount = 0;
        int readCount;
        int x;
      
        readCount = sizeof(doubleTsBuf) / sizeof(doubleTsBuf[0]);

        cbDoubleTsBuf = fread(doubleTsBuf,
                              sizeof(doubleTsBuf[0]),
                              readCount,
                              g_sBurstDevs.mBurstSender.pTxFile);
      
        // Dispurse the TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        for ( x = g_sBurstDevs.mBurstSender.nTsToSend; 
              (byteCount < cbDoubleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
              x += 32   // Burst device has 32 TS
            ) {
          // Fill two TS with data for 16 bit device. The following byte order
          // is correct for files created with the goldwave tool
          g_sBurstDevs.mBurstSender.sqTxBuf[x + 1] = doubleTsBuf[byteCount++];
          g_sBurstDevs.mBurstSender.sqTxBuf[x] = doubleTsBuf[byteCount++];
        }
      
        assert(byteCount == cbDoubleTsBuf);
        g_sBurstDevs.mBurstSender.cbTxBuf = cbDoubleTsBuf * 16;

        if (!feof(g_sBurstDevs.mBurstSender.pTxFile)) {
          assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);
        } 
        else {
          DemoLogf("DemoBurstSender[%d]: Reached EOF\n", nSourceId);
        }
      }
      else {
        // Single TS
        OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32];
        int cbSingleTsBuf;
        int byteCount = 0;
        int x;
      
        cbSingleTsBuf = fread(&singleTsBuf,
                              sizeof(singleTsBuf[0]),
                              sizeof(singleTsBuf) / sizeof(singleTsBuf[0]),
                              g_sBurstDevs.mBurstSender.pTxFile
                              );

      
        // Dispurse the single TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        for ( x = g_sBurstDevs.mBurstSender.nTsToSend; 
              (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
              x += 32
            ) {
          g_sBurstDevs.mBurstSender.sqTxBuf[x] = singleTsBuf[byteCount++];
        }
      
        assert(byteCount == cbSingleTsBuf);
        g_sBurstDevs.mBurstSender.cbTxBuf = cbSingleTsBuf * 32;

        if (!feof(g_sBurstDevs.mBurstSender.pTxFile)) {
          assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);
        } 
        else {
          DemoLogf("DemoBurstSender[%d]: Reached EOF\n", nSourceId);
        }
      }
    }
        
    if (g_sBurstDevs.mBurstSender.cbTxBuf > 0) {

      ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, 
                               nSourceId, 
                               g_sBurstDevs.mBurstSender.sqTxBuf, 
                               &ui32);
      DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

      // check the amount buffered w/ how much we requested to be buffered
      // (since we set the trigger amount to the buffer size, there should
      // have been room to buffer this, so these two should be the same).
      if(ui32 != g_sBurstDevs.mBurstSender.cbTxBuf) {
         DemoLogf("DemoBurstSender[%d]: Burst buffer transmit size error\n", 
                  nSourceId);
         DemoLogf("    Expected transfer size %d\n", 
                  g_sBurstDevs.mBurstSender.cbTxBuf);
         DemoLogf("    Actual transfer size %d\n", ui32);
      }    
    } else {

      if (g_sBurstDevs.mBurstSender.bFileLoop) {
        if (g_sBurstDevs.mBurstSender.bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete."
                   " Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_sBurstDevs.mBurstSender.pTxFile);
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n",
                 nSourceId);
        fclose(g_sBurstDevs.mBurstSender.pTxFile);
        g_sBurstDevs.mBurstSender.pTxFile = NULL;
        g_sBurstDevs.mBurstSender.bActive = OTX_FALSE;
      }
    }

  } else {
    // Send a text string (continously until aborted)

    // we MUST round amount to send to be bounded on the number of TS the highway hosts.
    //  (the buffer is representative of a burst DMA buffer for all TS on the highway hosted, TS0-TSn)

    g_sBurstDevs.mBurstSender.cbTxBuf = g_sBurstDevs.mBurstSender.cbTxBuf;
    assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);

    ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    assert(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf);
  }

  return;
}


void DemoBurstReceiver(OTX_UINT32 nSourceId)
/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstReceiver.bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // echo if we're in verbose mode
  if (g_sBurstDevs.mBurstReceiver.bVerbose) {
    int x;

    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d] (len=%.5d) ID:%08x ", nSourceId, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      if ((x % 8) == 0) DemoLogf("\n");
      DemoLogf("%.2X ", buf[x]);
    }
    DemoLogf("\n");
    DemoLogf("\n");
  }

  // write data to a file if desired
  if (g_sBurstDevs.mBurstReceiver.pRxFile) {
    int cbWritten;

    if (g_sBurstDevs.mBurstReceiver.nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      if(g_demoEnv.law != OTX_DATA_FORMAT_UINT16) {
        // Single TS
        OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
        int byteCount = 0;
        int x;

        byteCount = 0;
        for (x = g_sBurstDevs.mBurstReceiver.nTsToRecord; x < (OTX_INT32) sizeof(buf); x += 32) {
          singleTsBuf[byteCount++] = buf[x];
        }
        cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile);
        if (cbWritten != byteCount) {
          DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
        }
      }
      else {
        // 16 bit of data two TS
        OTX_UINT8 doubleTsBuf[DEMO_BURST_BUF_SIZE/16];
        int byteCount = 0;
        int x;

        byteCount = 0;
        for (x = g_sBurstDevs.mBurstReceiver.nTsToRecord; x < (OTX_INT32) sizeof(buf); x += 32) {
          doubleTsBuf[byteCount++] = buf[x+1];
          doubleTsBuf[byteCount++] = buf[x];
        }
        cbWritten = fwrite(doubleTsBuf, sizeof(doubleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile);
        if (cbWritten != byteCount) {
          DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
        }
      }
    }
  }
}


static char *DemoCptParamToStr(int nParam) 
/*******************************************************************************

  @func

  Convert a CPT (Call Progress Tone) Detection parameter to a string.

  @comm

*******************************************************************************/
{
  static char strBuf[128];

  switch (nParam) {

  case OTX_CST_CPT_DETECTOR_NO_TONE:        
    return "Silence, No tone";
  case OTX_CST_CPT_DETECTOR_DIAL_TONE:      
    return "Dial Tone";
  case OTX_CST_CPT_DETECTOR_RING_TONE:      
    return "Ring Tone";
  case OTX_CST_CPT_DETECTOR_BUSY_TONE:      
    return "Busy Tone";
  case OTX_CST_CPT_DETECTOR_FASTBUSY_TONE:  
    return "FastBusy Tone";
  case OTX_CST_CPT_DETECTOR_LONG_TONE:      
    return "Long lasting tone for 1 second";
  case OTX_CST_CPT_DETECTOR_END_OF_TONE:    
    return "End of current tone";
  default: // Error
    // DemoLogf("Unrecognized CPT tone parameter (%d)\n", nParam);
    sprintf(strBuf, "Unrecognized CPT Parameter (nParam=0x%X)", nParam);
    return strBuf;
  }
}



void DemoPrintEvent(OtxEventDataS *pEventData)
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/

{
  OTX_RESULT otxRc;
  char msgStr[256];

  //DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    
  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the DspDemo1.out logical devices
    DemoPrintCSTEvent(pEventData);
    break;

  case OTX_DEVICE_DSP:
    {
      // Non-logical device related event from the DSPs
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_LDEVICE_CST_DTMF_DIALER:
    {
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DIALER_EC_COMPLETE:
        DemoLogf("DtmfDialer[%d] Completed DTMF dialing of %d digits\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      default:
        DemoLogf("DtmfDialer[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
      }
    }
    break;


  case OTX_LDEVICE_CST_CPT_DETECTOR:
    {
      // Event from Cpt Detector
      switch (pEventData->m_nCode) {
      case OTX_CST_CPT_DETECTOR_EC_MTONESTART:
        {
          DemoLogf("CptDetector[%d] MTONESTART, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MTONE:
        {
          DemoLogf("CptDetector[%d] MTONE, Tone confirmed, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MSIGNALEND:
        {
          DemoLogf("CptDetector[%d] MSIGNALEND, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MFAIL:
        {
          DemoLogf("CptDetector[%d] MFAIL, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MPATTERNEND:
        {
          DemoLogf("CptDetector[%d] MPATTERNEND, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      default:
        DemoLogf("CptDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
        
      }
    }
    break;

  case OTX_LDEVICE_CST_DTMF_DETECTOR:
    {
      // Event from Dtmf Detector
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DETECTOR_EC_MTONESTART:
        {
          DemoLogf("DtmfDetector[%d] MTONESTART, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MTONE:
        {
          DemoLogf("DtmfDetector[%d] MTONE, DTMF tone confirmed\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MSIGNALEND:
        {
          DemoLogf("DtmfDetector[%d] MSIGNALEND, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MFAIL:
        {
          DemoLogf("DtmfDetector[%d] MFAIL\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MPATTERNEND:
        {
          DemoLogf("DtmfDetector[%d] MPATTERNEND\n", pEventData->m_nSourceId);
        }
        break;
        
      default:
        DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
        
      }
    }
    break;

  case OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR:
    {
      switch (pEventData->m_nCode) {
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_NOISE:
        DemoLogf("VoiceActivityDetector[%d] Noise present\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_NOISE_HANGOVER:
        DemoLogf("VoiceActivityDetector[%d] Noise Hangover. Possibly turning into voice.\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_VOICE:
        DemoLogf("VoiceActivityDetector[%d] Voice present\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_VOICE_HANGOVER:
        DemoLogf("VoiceActivityDetector[%d] Voice Hangover. Had voice, but possibly turning into noice.\n", pEventData->m_nSourceId);
        break;
      default:
        DemoLogf("VoiceActivityDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
      }
    }
    break;


  case OTX_LDEVICE_CST_LI_POTS:
    {
      // Event from Pots Li
      switch (pEventData->m_nCode) {

      case OTX_CST_POTS_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] A ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_AUTO_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal due to offhook has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_TIP_RING_REVERSAL_DETECTED:
        DemoLogf("PotsLi[%d] Tip/Ring reversal detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_CALLER_ID_DETECTED:
        {
          char szCallerId[40];
          OtxCstCidPresentationMessageS sCidMsg;
          OTX_UINT32 cbAttribValue;
          
          if (pEventData->m_nParam == OTX_CST_CALLER_ID_EC_NOERROR) {
            otxRc = OtxDrvGetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, szCallerId, 40); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("PotsLi[%d] Caller Id = \"%s\"\n", pEventData->m_nSourceId, szCallerId);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("PotsLi[%d] Caller Id Overrun. Caller Id string lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, NULL); 
            } else {
              DemoLogf("PotsLi[%d] Unexpected return code from OtxDrvGetAttributeValueSz (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }
          } else {
            DemoLogf("PotsLi[%d] Caller Id detection failure. nCode=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
          }
          
          otxRc = OtxDrvGetAttributeValue(pEventData->m_hDevice,OTX_ATTR_CST_POTS_CALLER_ID_PRESENTATION_MESSAGE, &sCidMsg, sizeof(sCidMsg), &cbAttribValue); 
          if (OTX_SUCCESS(otxRc)) {
            DemoLogf("PotsLi[%d] Caller Id Presentation Message:\n", pEventData->m_nSourceId);
            DemoPrintPresentationMessage(&sCidMsg);
          } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
            DemoLogf("PotsLi[%d] Caller Id Overrun. Caller Msg lost. Reseting the condition\n", pEventData->m_nSourceId);
            OtxDrvSetAttributeValue(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID_PRESENTATION_MESSAGE, NULL, 0); 
          } else {
            DemoLogf("PotsLi[%d] Unexpected return code from OtxDrvGetAttributeValue (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
          }
          
          
        }
        break;
      case OTX_CST_POTS_OFF_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_TRUE;
        DemoLogf("PotsLi[%d] Off-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_LINE_MON_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook Line Monitor mode command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_SENSE_LINE_COMPLETE:
        {
          double nVal = 2.75 * pEventData->m_nParam;
        
          assert(pEventData->m_nSourceId < DEMO_LI_MAX);
          if (g_bLineSenseVerbose) {
            //DemoPrintTimeStamp(pEventData->m_nTimeStamp);
          }
          if (g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook) {
            // We are in the OffHook state so line sensing will return loop current (3mA per bit)
            if (g_bLineSenseVerbose) {
              DemoLogf("PotsLi[%d] Loop current = %d mA (%d) above minimum\n", pEventData->m_nSourceId, 3 * pEventData->m_nParam, pEventData->m_nParam);
            }

            // Look for Far-End-Hangup. Two consecutive loop current values of zero is an indication of a Far-End-Hangup.
            if ((g_nLineSenseLastValue == 0) && (pEventData->m_nParam == 0)) {
              OTX_RESULT result;
              DemoLogf("PotsLi[%d] Detected Far-End-Hangup. Going OnHook.\n", pEventData->m_nSourceId);
            
              result = OtxCstPotsOnHook(pEventData->m_hDevice, 0x3001, 500); 
              if (!OTX_SUCCESS(result)) {
                DemoLogf("PotsLi[%d] On-Hook command failed\n", pEventData->m_nSourceId);
              }

            }
          } else {
            // We are in the OnHook state so line sensing will return line voltage (2.75V per bit)
            if (g_bLineSenseVerbose) {
              DemoLogf("PotsLi[%d] Line voltage = %2.1f V  (%d)\n", pEventData->m_nSourceId, (double) nVal, pEventData->m_nParam);
            }
          }
          g_nLineSenseLastValue = pEventData->m_nParam;
        }
        break;
      case OTX_CST_POTS_PULSE_GEN_COMPLETE:
        DemoLogf("PotsLi[%d] Pulse Generator command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_READ_REG_COMPLETE:
        DemoLogf("PotsLi[%d] ReadReg command complete Val=0x%.2X\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      case OTX_CST_POTS_WRITE_REG_COMPLETE:
        DemoLogf("PotsLi[%d] WriteReg command complete\n", pEventData->m_nSourceId);
        break;
      default:
        // Unexpected event from this device
        DemoLogf("PotsLi[%d] Unexpected Event (%d)\n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_LDEVICE_CST_CALLER_ID_DETECTOR:
    {
      // Event from Caller Id detector
      switch (pEventData->m_nCode) {

      case OTX_CST_CID_DETECTOR_DETECTED_CID:
        {
          char szCallerId[40];
          OtxCstCidPresentationMessageS sCidMsg;
          OTX_UINT32 cbAttribValue;

          if (pEventData->m_nParam == OTX_CST_CALLER_ID_EC_NOERROR) {
            otxRc = OtxDrvGetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_CID_STRING, szCallerId, 40); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("Cid[%d] Caller Id = \"%s\"\n", pEventData->m_nSourceId, szCallerId);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("Cid[%d] Caller Id Overrun. Caller Id string lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_CID_STRING, NULL); 
            } else {
              DemoLogf("Cid[%d] Unexpected return code from OtxDrvGetAttributeValueSz (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }


            otxRc = OtxDrvGetAttributeValue(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_PRESENTATION_MESSAGE, &sCidMsg, sizeof(sCidMsg), &cbAttribValue); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("Cid[%d] Caller Id Presentation Message:\n", pEventData->m_nSourceId);
              DemoPrintPresentationMessage(&sCidMsg);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("Cid[%d] Caller Id Overrun. Caller Msg lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValue(pEventData->m_hDevice, OTX_ATTR_CST_CID_DETECTOR_PRESENTATION_MESSAGE, NULL, 0); 
            } else {
              DemoLogf("Cid[%d] Unexpected return code from OtxDrvGetAttributeValue (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }


          } else {
            DemoLogf("Cid[%d] Caller Id detection failure. nCode=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
          }

        }
        break;
      default:
        // Unexpected event from this device
        DemoLogf("PotsLi[%d] Unexpected Event (%d)\n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;

  case OTX_DEVICE_BURST:
    {
      switch (pEventData->m_nCode) {
      case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
        DemoBurstReceiver(pEventData->m_nRequestId);
        break;
      case OTX_BURST_SENDER_EC_BUF_DEPLETED:
        if(g_sBurstDevs.mBurstSender.bUseAlternateSend == OTX_FALSE) {
          DemoBurstSender(pEventData->m_nRequestId);
        }
        else {
        }

        if(g_sBurstDevs.mBurstSender.bVerbose) {
          DemoLogf("\nBurstSend[%d] ", pEventData->m_nRequestId);
          DemoPrintTimeStamp(pEventData->m_nTimeStamp);
          DemoLogf("\n");
        }

        break;

      case OTX_BURST_RECEIVER_DMA_COMPLETE:
        DemoLogf("\nBurstReceiver[%d] DMA Transfer Count %d ", 
                 pEventData->m_nRequestId, 
                 pEventData->m_nParam);
        DemoPrintTimeStamp(pEventData->m_nTimeStamp);
        DemoLogf("\n");

        g_nDmaEventCnt++;
        if(g_nDmaEventCnt > 30) {
          OTX_RESULT result;

          // Turn off event so user can have access to command line interface
          result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_DMA_RECEIVE_SIGNAL,  
                                               OTX_FALSE); 
          DemoCheckRc(g_hPhysDevs.hBurst, result);
          result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_DMA_SEND_SIGNAL,  
                                               OTX_FALSE); 
          DemoCheckRc(g_hPhysDevs.hBurst, result);

          g_nDmaEventCnt = 0;
        }

        break;
      case OTX_BURST_SENDER_DMA_COMPLETE:
        DemoLogf("\nBurstSend[%d] DMA Transfer Count %d  ", 
                 pEventData->m_nRequestId, 
                 pEventData->m_nParam);
        DemoPrintTimeStamp(pEventData->m_nTimeStamp);
        DemoLogf("\n");

        g_nDmaEventCnt++;
        if(g_nDmaEventCnt > 30) {
          OTX_RESULT result;
          // Turn off event so user can have access to command line interface
          result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_DMA_RECEIVE_SIGNAL,  
                                               OTX_FALSE); 
          DemoCheckRc(g_hPhysDevs.hBurst, result);
          result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_DMA_SEND_SIGNAL,  
                                               OTX_FALSE); 
          DemoCheckRc(g_hPhysDevs.hBurst, result);

          g_nDmaEventCnt = 0;
        }

        break;

      default:
        otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        }
        break;
      }
      
    }
    break;

  default:
    {
      // Unhandled event
      DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    }
    break;
    
  }
  
}


static void DemoPrintPresentationMessage(OtxCstCidPresentationMessageS *pCidMsg) 
/*******************************************************************************

  @func

  Print a Caller ID Presentation Message

  @comm

*******************************************************************************/
{
  OTX_INT32 i;

  if (pCidMsg->validFields & OTX_CID_VALIDDATE) {
    DemoLogf("date: %c%c/%c%c\n", pCidMsg->date[0], pCidMsg->date[1], pCidMsg->date[2], pCidMsg->date[3]);
    DemoLogf("time: %c%c:%c%c\n", pCidMsg->date[4], pCidMsg->date[5], pCidMsg->date[6], pCidMsg->date[7]);
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLINGLINEIDENTITY) {
    DemoLogf("callingLineIdentity: ");
    for (i = 0; pCidMsg->callingLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->callingLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLEDLINEIDENTITY) {
    DemoLogf("calledLineIdentity: ");
    for (i = 0; pCidMsg->calledLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->calledLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLINGPARTYNAME) {
    DemoLogf("callingPartyName: ");
    for (i = 0; pCidMsg->callingPartyName[i] && (i < 51); i++) {
      DemoLogf("%c", pCidMsg->callingPartyName[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDFIRSTLINEIDENTITY) {
    DemoLogf("firstLineIdentity: ");
    for (i = 0; pCidMsg->firstLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->firstLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDREDIRECTEDNUMBER) {
    DemoLogf("redirectedNumber: ");
    for (i = 0; pCidMsg->redirectedNumber[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->redirectedNumber[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCOMPLEMENTARYCALLINGLINEIDENTITY) {
    DemoLogf("complementaryCallingLineIdentity: ");
    for (i = 0; pCidMsg->complementaryCallingLineIdentity[i] && (i < 21); i++) {
      DemoLogf("%c", pCidMsg->complementaryCallingLineIdentity[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDREASONFORABSENCE) {
    DemoLogf("reasonForAbsence: ");
    switch (pCidMsg->reasonForAbsence) {
    case 'O':
      DemoLogf("'O' - Unavailable (probably out of range)");
      break;
    case 'P':
      DemoLogf("'P' - Private (CLIR involved) call");
      break;
    default:
      DemoLogf("Unknown %c (0x%.4X)", pCidMsg->reasonForAbsence, pCidMsg->reasonForAbsence);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDREASONFORABSENCEOFPARTY) {
    DemoLogf("reasonForAbsenceOfParty: ");
    switch (pCidMsg->reasonForAbsenceOfParty) {
    case 'O':
      DemoLogf("'O' - Unavailable (probably out of range)");
      break;
    case 'P':
      DemoLogf("'P' - Private (CLIR involved) call");
      break;
    default:
      DemoLogf("Unknown %c (0x%.4X)", pCidMsg->reasonForAbsenceOfParty, pCidMsg->reasonForAbsenceOfParty);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDCALLTYPE) {
    DemoLogf("callType: ");
    switch (pCidMsg->callType) {
    case 0x01:
      DemoLogf("Voice Call");
      break;
    case 0x02:
      DemoLogf("CLI Ring Back in case of free call");
      break;
    case 0x03:
      DemoLogf("Calling Name Delivery");
      break;
    case 0x07:
      DemoLogf("Set Clock");
      break;
    case 0x81:
      DemoLogf("Message Waiting Call");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->callType);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDVISUALINDICATOR) {
    DemoLogf("visualIndicator: ");
    DemoLogf("%.4X ", pCidMsg->visualIndicator);
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDNETWORKMESSAGESYSTEMSTATUS) {
    DemoLogf("networkMessageSystemStatus: ");
    DemoLogf("%.4X ", pCidMsg->networkMessageSystemStatus);
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDEXTENSION) {
    DemoLogf("extension: ");
    for (i = 0; pCidMsg->extension[i] && (i < 10); i++) {
      DemoLogf("%c", pCidMsg->extension[i]);
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDTYPEOFFORWARDEDCALL) {
    DemoLogf("typeOfForwardedCall: ");
    switch (pCidMsg->typeOfForwardedCall) {
    case 0x01:
      DemoLogf("Call forwarded on busy");
      break;
    case 0x02:
      DemoLogf("Call forwarded on no reply");
      break;
    case 0x03:
      DemoLogf("Unconditional forwarded call");
      break;
    case 0x04:
      DemoLogf("Deflected call (after alerting)");
      break;
    case 0x05:
      DemoLogf("Deflected call (immediate)");
      break;
    case 0x06:
      DemoLogf("Call forwarded on inability to reach mobile subscriber");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->typeOfForwardedCall);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDTYPEOFCALLINGUSER) {
    DemoLogf("typeOfCallingUser: ");
    switch (pCidMsg->typeOfCallingUser) {
    case 0x00:
      DemoLogf("Origination unknown or unavailable");
      break;
    case 0x03:
      DemoLogf("Virtual Private Network");
      break;
    case 0x04:
      DemoLogf("Mobile phone");
      break;
    case 0x05:
      DemoLogf("Mobile phone + Virtual Private Network");
      break;
    case 0x0A:
      DemoLogf("Ordinary calling subscriber");
      break;
    case 0x0B:
      DemoLogf("Calling subscriber with priority");
      break;
    case 0x0C:
      DemoLogf("Data Call");
      break;
    case 0x0D:
      DemoLogf("Test Call");
      break;
    case 0x0F:
      DemoLogf("Payphone");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->typeOfCallingUser);
      break;
    }
    DemoLogf("\n");
  }
  
  if (pCidMsg->validFields & OTX_CID_VALIDLASTERROR) {
    DemoLogf("lastError: ");
    DemoLogf("%.4X ", pCidMsg->lastError);
    switch (pCidMsg->lastError) {
    case 0x00:
      DemoLogf("no error");
      break;
    case 0x10:
      DemoLogf("checksum error");
      break;
    case 0x11:
      DemoLogf("frame length error (unrecoverable)");
      break;
    case 0x12:
      DemoLogf("unrecognized message type (unrecoverable)");
      break;
    default:
      DemoLogf("Unknown (0x%.4X)", pCidMsg->lastError);
      break;
    }
    DemoLogf("\n");
  }
  
}

static char DemoDtmfParamToChar(int nParam) 
/*******************************************************************************

  @func

  Convert a DTMF parameter to a digit (char).

  @comm

*******************************************************************************/
{
  switch (nParam) {
  case 0:  // 0  (941Hz/1336Hz)
  case 1:  // 1  (697Hz/1209Hz)
  case 2:  // 2  (697Hz/1336Hz)
  case 3:  // 3  (697Hz/1477Hz)
  case 4:  // 4  (770Hz/1209Hz)
  case 5:  // 5  (770Hz/1336Hz)
  case 6:  // 6  (770Hz/1477Hz)
  case 7:  // 7  (852Hz/1209Hz)
  case 8:  // 8  (852Hz/1336Hz)
  case 9:  // 9  (852Hz/1477Hz)
    return '0' + nParam;
  case 10: // A  (697Hz/1633Hz)
  case 11: // B  (770Hz/1633Hz)
  case 12: // C  (852Hz/1633Hz)
  case 13: // D  (941Hz/1633Hz)
    return 'A' - 10 + nParam;
  case 14: // *  (941Hz/1209Hz)
    return '*';
    break;
  case 15: // #  (941Hz/1477Hz)
    return '#';
    break;
  default: // Error
    DemoLogf("Unrecognized DTMF tone parameter (%d)\n", nParam);
    return '-';
  }
}



OTX_BOOL DemoIsBurstDeviceEnabled(void)
/*******************************************************************************

  @func

    Returns the state of the Burst device (enabled or disabled).

*******************************************************************************/
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;

  result = OtxDrvGetState(g_hPhysDevs.hBurst, &eDevState); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);

  if (eDevState == OTX_DEVICE_WORKING) {
    return OTX_TRUE; // The burst device is enabled
  } else {
    return OTX_FALSE; // The burst device is not enabled
  }
}


OTX_BOOL DemoIsVADDeviceEnabled(OTX_INT32 nDsp)
/*******************************************************************************

  @func

    Returns the state of the VAD device (enabled or disabled).

*******************************************************************************/
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;

  result = OtxDrvGetState(g_hLogicDevs.hVaDetector[nDsp], &eDevState); 
  DemoCheckRc(g_hLogicDevs.hVaDetector[nDsp], result);

  if (eDevState == OTX_DEVICE_WORKING) {
    return OTX_TRUE; // The VAD device is enabled
  } else {
    return OTX_FALSE; // The VAD device is not enabled
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  OTX_RESULT        result;
  static OTX_INT32  nCurrentLi = 0;
  static OTX_BOOL   bBurstInBypass = OTX_FALSE;

  // The following is used to configure logical 
  // devices while the application is active.
  static OTX_HANDLE hPipeVAD0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeVAD1 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeCPT0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeCPT1 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDTMF0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeDTMF1 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeNoVoiceMix0 = OTX_INVALID_HANDLE_VALUE;
  static OTX_HANDLE hPipeNoVoiceMix1 = OTX_INVALID_HANDLE_VALUE;

  DemoLogf("'%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd   Synopsis\n");
    DemoLogf("===   ===========================================================\n");
    DemoLogf(" ?    Display Help text\n");
    DemoLogf(" h    Ditto\n");
    DemoLogf(" q    Quit\n");

    DemoLogf(" 0-5  Select active FXO Line Interface\n");
    DemoLogf(" f    Set active Pots Li Off-Hook\n");
    DemoLogf(" n    Set active Pots Li On-Hook\n");
    DemoLogf(" m    Set active Pots Li in On-Hook Line-Monitor mode\n");
    DemoLogf(" c    Toggle verbose line sensing [Reading line voltage or loop current]\n");

    DemoLogf(" d    Call DTMF digits %s on the active Pots Li\n", g_sqNumber0);
    DemoLogf(" D    Dial user specified DTMF digits on the active Pots Li\n");

    DemoLogf(" g    Connect Voice Activity Detector (VAD)\n");
    DemoLogf(" G    Disconnect Voice Activity Detector (VAD)\n");
    DemoLogf(" i    Connect Call Progress Tone (CPT) Detector\n");
    DemoLogf(" I    Disconnect Call Progress Tone (CPT) Detector\n");
    DemoLogf(" j    Connect DTMF Detector\n");
    DemoLogf(" J    Disconnect DTMF Detector\n");
    DemoLogf(" k    Connect DTMF Dialer directly to PotLi (Bypass Voice Mixer)\n");
    DemoLogf(" K    Connect Burst Output directly to PotLi (Bypass Voice Mixer)\n");

    DemoLogf(" l    Flash/hook on the active Pots Li\n");
    DemoLogf(" r    Read a Register from the DAA\n");
    DemoLogf(" w    Write to a Register in the DAA\n");
    DemoLogf(" b    Get Debug data from the DSP\n");
    DemoLogf(" s    Stream receive data to file (toggle). Only for DSP debugging\n");

    DemoLogf(" e    Enable/Disable Burst (toggle)\n");
    DemoLogf(" a    Abort Burst Senders and Receivers\n");
    DemoLogf(" t    Enable/Disable Burst DMA transfer notification (toggle)\n");
    DemoLogf(" u    Get Burst Sender and Receive frame reference data\n");
    DemoLogf(" R    Start Burst Receiver. Prompt for file to recv\n");
    DemoLogf(" S    Start Burst Sender. Prompt for file to send\n");
    DemoLogf(" T    Start Burst Sender and Receive. Prompt for files\n");
    DemoLogf(" L    Send loop. Start sending the file again once EOF is reached.\n");
    DemoLogf(" !    Terminate Watchdog Timer Thread (Crash test)\n");
    DemoLogf(" v(V) Toggle Burst Receiver(Send) Verbose mode\n");

    break;

  case 'v':
    {
      if (g_sBurstDevs.mBurstReceiver.bVerbose) {
        g_sBurstDevs.mBurstReceiver.bVerbose = OTX_FALSE;
      } else {
        g_sBurstDevs.mBurstReceiver.bVerbose = OTX_TRUE;
      }
      DemoLogf("Burst Receiver Verbose is now %s\n", 
               g_sBurstDevs.mBurstReceiver.bVerbose ? "ON" : "OFF");
    }
    break;


  case 'V':
    {
      if (g_sBurstDevs.mBurstSender.bVerbose) {
        g_sBurstDevs.mBurstSender.bVerbose = OTX_FALSE;
      } else {
        g_sBurstDevs.mBurstSender.bVerbose = OTX_TRUE;
      }
      DemoLogf("Burst Sender Verbose is now %s\n", 
               g_sBurstDevs.mBurstSender.bVerbose ? "ON" : "OFF");
    }
    break;


  case '!':
    {
      DemoLogf("Terminating WatchDog Thread (Crash test)\n");
      g_bKickTheWatchDog = OTX_FALSE;
    }
    break;


  case 'e':
    if (DemoIsBurstDeviceEnabled()) {
      // by disabling the Burst device, we stop the process of transferring data
      // from the hardware to the driver via DMA.
      DemoLogf("Disable g_hPhysDevs.hBurst\n");
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckRc(g_hPhysDevs.hBurst, result);
    } else {
      // by enabling the Burst device, we start the process of transferring data
      // from the hardware to the driver via DMA.  The data is stored in the Burst
      // device's buffer for later processing to connected logical devices.
      DemoLogf("Enable g_hPhysDevs.hBurst\n");
      result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckRc(g_hPhysDevs.hBurst, result);
    }
    break;

  case 'a':
    {
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("DemoBurstReceiver: Closing file\n");
          fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
          g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
        }
        g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
        
        if (g_sBurstDevs.mBurstSender.pTxFile) {
          DemoLogf("mBurstSender: Closing file\n");
          fclose(g_sBurstDevs.mBurstSender.pTxFile);
          g_sBurstDevs.mBurstSender.pTxFile = NULL;
        }
        g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      char sqzFile[OTX_MAX_PATH];
      static OTX_INT32 nTsToRecord = -1;

      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE || g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
          DemoLogf("BurstReceiver. Please enter file to receive: ");
          OsaConsoleGetStr(sqzFile);
          if (sqzFile[0] == '\0') {
            DemoLogf("Command aborted.\n");
            break;
          }
          
          if ((g_sBurstDevs.mBurstReceiver.pRxFile = fopen(sqzFile, "wb")) == NULL) {
            DemoLogf("Error: Could not open file %s. Command aborted.\n", sqzFile);
            break;
          }
          
          DemoLogf("BurstReceiver. Please enter single TS[0-31] to receive to file (-1=All) [%d]: ", nTsToRecord);
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0]) {
            nTsToRecord = strtol(sqzBuf, NULL, 0);
          }
          g_sBurstDevs.mBurstReceiver.nTsToRecord = nTsToRecord;
                         
          if (nTsToRecord == -1) {
            DemoLogf("Burst Receiver. Recording all 32 Burst Timeslots into file %s\n", sqzFile);
          } else {
            DemoLogf("Burst Receiver. Recording TS%d of Burst Highway into file %s\n", nTsToRecord, sqzFile);
          }
          
          g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Receiver. Thread already running.\n");
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'S':
    {
      char sqzBuf[OTX_MAX_PATH];
      static OTX_INT32 nTsToSend = 0;
      
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstSender.bActive == OTX_FALSE) {
          DemoLogf("BurstSender. Please enter file to send: ");
          
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            break;
          }
          
          if ((g_sBurstDevs.mBurstSender.pTxFile = fopen(sqzBuf, "rb")) == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          
          if(g_demoEnv.law == OTX_DATA_FORMAT_UINT16) {
            DemoLogf("BurstSender. Sending data from file on TS%d and TS%d of the Burst highway\n", nTsToSend, nTsToSend + 1);
          }
          else {
            DemoLogf("BurstSender. Sending data from file on TS%d of the Burst highway\n", nTsToSend);
          }
          
          g_sBurstDevs.mBurstSender.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Sender already running.\n");
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'u':
    if (DemoIsBurstDeviceEnabled()) {
      OTX_UINT32                ui32;
      OTX_DATETIME      nTimeStamp;
      OtxBurstFrameRefAndTime   FrameRef;             // frame reference data
      static OTX_UINT64         nLastRecvFrame = 0;
      static OTX_UINT64         nLastSendFrame = 0;

      OtxDrvGetCurrentTime(&nTimeStamp);

      DemoLogf("\n");

      DemoLogf("\nCurrent Time( ");
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf(")\n");

      result = OtxDrvGetAttributeValue(g_hPhysDevs.hBurst, OTX_ATTR_BURST_GET_FRAME_REF_DATA, &FrameRef, sizeof(FrameRef), &ui32); DemoCheckRc(g_hPhysDevs.hBurst, result);
      // just skip if we wrapped
      if (FrameRef.m_FrameCounter > nLastRecvFrame) {
        DemoLogf("Burst FRef:: %d fps  Cntr:%08x%08x  Time:%08x%08x \n",
          nLastRecvFrame == 0 ? 0 : (OTX_UINT32) (FrameRef.m_FrameCounter - nLastRecvFrame),
          (OTX_UINT32) (FrameRef.m_FrameCounter >> 32),
          (OTX_UINT32) (FrameRef.m_FrameCounter),
          FrameRef.m_FrameDateTime.sUint32.nHigh,
          FrameRef.m_FrameDateTime.sUint32.nLow
        );
        DemoLogf("\nFrame Time( ");
        DemoPrintTimeStamp(FrameRef.m_FrameDateTime);
        DemoLogf(")\n");
      }
      nLastRecvFrame = FrameRef.m_FrameCounter;

      DemoLogf("\n");

      result = OtxDrvGetAttributeValue(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_GET_FRAME_REF_DATA, &FrameRef, sizeof(FrameRef), &ui32); DemoCheckRc(g_hPhysDevs.hBurst, result);
      // just skip if we wrapped
      if (FrameRef.m_FrameCounter > nLastSendFrame) {
        DemoLogf("Burst FRef:: %d fps  Cntr:%08x%08x  Time:%08x%08x \n",
          nLastSendFrame == 0 ? 0 : (OTX_UINT32) (FrameRef.m_FrameCounter - nLastSendFrame),
          (OTX_UINT32) (FrameRef.m_FrameCounter >> 32),
          (OTX_UINT32) (FrameRef.m_FrameCounter),
          FrameRef.m_FrameDateTime.sUint32.nHigh,
          FrameRef.m_FrameDateTime.sUint32.nLow
        );
        DemoLogf("\nFrame Time( ");
        DemoPrintTimeStamp(FrameRef.m_FrameDateTime);
        DemoLogf(")\n");
      }
      nLastSendFrame = FrameRef.m_FrameCounter;
    }
    else {
      DemoLogf("The Burst device is not enabled. Command ignored\n");
    }
    break;

  case 't':
    {
      OTX_BOOL bDmaTranferMod;  
      // Get current setting
      result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hBurst, 
                                           OTX_ATTR_BURST_DMA_RECEIVE_SIGNAL, 
                                           &bDmaTranferMod); 
      DemoCheckRc(g_hPhysDevs.hBurst, result);

      // toggle setting
      bDmaTranferMod = !bDmaTranferMod;

      // Enable/Disable DMA start signal for first DMA transfer
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                           OTX_ATTR_BURST_DMA_RECEIVE_SIGNAL,  
                                           bDmaTranferMod); 
      DemoCheckRc(g_hPhysDevs.hBurst, result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                           OTX_ATTR_BURST_DMA_SEND_SIGNAL,  
                                           bDmaTranferMod); 
      DemoCheckRc(g_hPhysDevs.hBurst, result);

      DemoLogf("Burst Sender and Receiver DMA signal is %s\n", 
                bDmaTranferMod ? "ON" : "OFF");
      // This counter is used to limit the number of events produced
      g_nDmaEventCnt = 0;
    }
    break;

  case 'T':
    {
      OTX_FP64    nFrmTxTime = (0.000125 * 1000);// rate 8000 frame/sec in mSec
      OTX_UINT32  nTxBufSize;
      OTX_UINT32  nTxItemSize;
      OTX_UINT32  ui32;
      OTX_UINT32  nTxTimeSlots;
      OTX_UINT32  nFrameCount;
      OTX_UINT32  nDrvTxBufSize;
      OTX_UINT32  nHwTxBufSize;
      OTX_UINT32  nBufSendTime;
      OTX_BOOL    bSaveTimeStampMod;
      char *pInitDataBuf;
      char nSilence;
      char sqzBuf[OTX_MAX_PATH];
      char sqzFile[OTX_MAX_PATH];
      static OTX_INT32 nTsToSend = 0;
      static OTX_INT32 nTsToRecord = -1;

      // User has an option to direcly connect DTMF to PotLi
      if( (hPipeNoVoiceMix0 != OTX_INVALID_HANDLE_VALUE) ||
          (hPipeNoVoiceMix1 != OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Warning: Burst device may not be connected to PotLi\n");
      }
      
      // If Voice Mixer is connected to output path total
      // system delay will be 32 frames or 4mSec longer for
      // round trip data.
      if( bBurstInBypass == OTX_FALSE ) {
        DemoLogf("Burst device is connected to Voice Mixer.\n"
                 "Output path delays for system will be 4mSec longer.\n ");
      }

      if(g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE && 
         g_sBurstDevs.mBurstSender.bActive   == OTX_FALSE &&
         g_sBurstDevs.mBurstReceiver.pRxFile == NULL      && 
         g_sBurstDevs.mBurstSender.pTxFile   == NULL      && 
         DemoIsBurstDeviceEnabled()          == OTX_FALSE) {

        DemoLogf("BurstReceiver. Please enter file to receive: ");
        OsaConsoleGetStr(sqzFile);
        if (sqzFile[0] == '\0') {
          DemoLogf("Command aborted.\n");
          break;
        }

        g_sBurstDevs.mBurstReceiver.pRxFile = fopen(sqzFile, "wb");
        if (g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
          DemoLogf("Error: Could not open file %s. Command aborted.\n", 
                   sqzFile);
          break;
        }
        
        DemoLogf("Please enter a single TS[0-31] or (-1=All TS) [%d]: ", 
                 nTsToRecord);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToRecord = strtol(sqzBuf, NULL, 0);
        }
        g_sBurstDevs.mBurstReceiver.nTsToRecord = nTsToRecord;
                       
        DemoLogf("Burst Receiver:\n");
        if (nTsToRecord == -1) {
          DemoLogf("Recording all 32 Burst Timeslots into file %s\n", 
                   sqzFile);
        } else {
          DemoLogf("Recording TS%d of Burst Highway into file %s\n", 
                   nTsToRecord, 
                   sqzFile);
        }
    

        DemoLogf("BurstSender: Please enter file to send: ");
        
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          break;
        }

        g_sBurstDevs.mBurstSender.pTxFile = fopen(sqzBuf, "rb");
        if (g_sBurstDevs.mBurstSender.pTxFile == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }
        
        DemoLogf("BurstSender:\n"); 
        if(g_demoEnv.law == OTX_DATA_FORMAT_UINT16) {
          DemoLogf("Sending data from file on TS%d-%d of the Burst highway\n", 
                   nTsToSend, 
                   nTsToSend + 1);
        }
        else {
          DemoLogf("Sending data from file on TS%d of the Burst highway\n", 
                   nTsToSend);
        }

        /*
        ** Initialize data in Burst buffer
        */
        switch (g_demoEnv.law) {
          case OTX_DATA_FORMAT_COMPAND8_A_LAW:
            nSilence = (char) DEMO_A_LAW_SILENCE; // silence in a-law mode
            break;                         
          case OTX_DATA_FORMAT_COMPAND8_U_LAW:
            nSilence = (char) DEMO_U_LAW_SILENCE; // silence in u-law mode
            break;
          case OTX_DATA_FORMAT_UINT16:
            nSilence = 0; // silence in Linear data mode
            break;
          default:
            nSilence = (char) DEMO_U_LAW_SILENCE; // silence in u-law mode
            DemoLogf("Warning: g_demoEnv.law not configured using default\n");
              break;
        }

        // Prime the transmit buffer with silence
        nTxItemSize = sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]);
        nTxBufSize = sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / nTxItemSize;
        memset(&g_sBurstDevs.mBurstSender.sqTxBuf[0], 
               nSilence, 
               nTxBufSize);

        //OTX_ATTR_BURST_SENDER_FIFO_SIZE, Host buffer
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_SENDER_FIFO_SIZE,
                                               &nDrvTxBufSize); 
        DemoCheckRc(g_hPhysDevs.hBurst, result);

        //OTX_ATTR_BURST_TX_BUFFER_SIZE, DMA HW Buffer
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_TX_BUFFER_SIZE,
                                               &nHwTxBufSize); 
        DemoCheckRc(g_hPhysDevs.hBurst, result);

        //OTX_ATTR_BURST_MAX_TIMESLOTS, number of timeslots
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_MAX_TIMESLOTS,
                                               &nTxTimeSlots); 
        DemoCheckRc(g_hPhysDevs.hBurst, result);


        /*
        ** Prime burst transmit buffer in OTX driver with silence.
        ** This data will be transmitted then the data from the 
        ** input file will be transmitted.
        */
        DemoLogf("Load Burst transmit buffer with silence.\n");

        pInitDataBuf = (char *) malloc(nDrvTxBufSize);

        memset(pInitDataBuf, 
               nSilence, 
               nDrvTxBufSize);
        
        ui32 = nDrvTxBufSize;
        result = OtxBurstInitSendBuffer(g_hPhysDevs.hBurst, 
                                        0, 
                                        pInitDataBuf, 
                                        &ui32);
        // Finished with this buffer. 
        // Return it here while we have the chance.
        free(pInitDataBuf);
        DemoCheckRc(g_hPhysDevs.hBurst, result);

        // Check the amount transferred with how much we 
        // requested to be sent they should be the same.
        if(ui32 != nDrvTxBufSize) {
           DemoLogf("Burst buffer transmit size error.\n");
           DemoLogf("    Expected transfer size %d\n", nDrvTxBufSize);
           DemoLogf("    Actual transfer size %d\n", ui32);
        }
    
        DemoLogf("Buffer timing\n");
        nFrameCount  = nHwTxBufSize/nTxTimeSlots;
        nBufSendTime = (OTX_UINT32)((OTX_FP64)nFrameCount * nFrmTxTime);
        DemoLogf("    HW Buffer size is %d, time %d mSec\n", 
                 nHwTxBufSize, 
                 nBufSendTime);

        nFrameCount  = nDrvTxBufSize/nTxTimeSlots;
        nBufSendTime = (OTX_UINT32) ((OTX_FP64)nFrameCount * nFrmTxTime);
        DemoLogf("    Host Buffer size is %d, time %d mSec\n", 
                 nDrvTxBufSize, 
                 nBufSendTime);

        nFrameCount  = (nDrvTxBufSize + nHwTxBufSize)/nTxTimeSlots;
        nBufSendTime = (OTX_UINT32)((OTX_FP64) nFrameCount * nFrmTxTime);
        DemoLogf("    File should start playing in %d mSec\n", nBufSendTime);
#if 1
        // This will create a lot of events which can cause the system to
        // become slow to respond to user inputs.
        // Enable DMA start signal for first DMA transfer
        result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                             OTX_ATTR_BURST_DMA_RECEIVE_SIGNAL,  
                                             OTX_TRUE); 
        DemoCheckRc(g_hPhysDevs.hBurst, result);
        result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                             OTX_ATTR_BURST_DMA_SEND_SIGNAL,  
                                             OTX_TRUE); 
        DemoCheckRc(g_hPhysDevs.hBurst, result);
        // This counter is used to limit the number of events produced
        g_nDmaEventCnt = 0;
#endif
        g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;
        g_sBurstDevs.mBurstSender.bActive   = OTX_TRUE;

        /*
        ** Resynchornize OTX driver to the system clock 
        */        
        // Get current timestamp type so we do not change it
        result = OtxDrvGetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, 
                                             OTX_ATTR_DRV_TIMESTAMP_FROM_1601, 
                                             &bSaveTimeStampMod); 
        DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
        // Resynchornize time within OTX driver with system time.
        result = OtxDrvSetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, 
                                             OTX_ATTR_DRV_TIMESTAMP_FROM_1601, 
                                             bSaveTimeStampMod); 
        DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);


        // By enabling the Burst device, we start the process of transferring 
        // data from the hardware to the driver via DMA. The data is stored in 
        // the Burst receive buffer. Data is also being transferred from the
        // Burst send buffer to the hardware via a second DMA channel.
        DemoLogf("Enable g_hPhysDevs.hBurst\n");
        result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
        DemoCheckRc(g_hPhysDevs.hBurst, result);
      } 
      else {
        DemoLogf("Burst sender or receiver already running.\n");

        DemoLogf("Do you want to abort current tranfer (Y/N)? ");
        
        OsaConsoleGetStr(sqzBuf);
        if ((sqzBuf[0] == 'y') || (sqzBuf[0] == 'Y')) {
          // Disable DMA start signal for first DMA transfer
          result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_DMA_RECEIVE_SIGNAL,  
                                               OTX_FALSE); 
          DemoCheckRc(g_hPhysDevs.hBurst, result);
          result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, 
                                               OTX_ATTR_BURST_DMA_SEND_SIGNAL,  
                                               OTX_FALSE); 
          DemoCheckRc(g_hPhysDevs.hBurst, result);

          g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
          g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;

          if(DemoIsBurstDeviceEnabled()) {
            // by disabling the Burst device, we stop the process of 
            // transferring data from the hardware to the driver via DMA.
            DemoLogf("Disable g_hPhysDevs.hBurst\n");
            result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
            DemoCheckRc(g_hPhysDevs.hBurst, result);
          } 

          if (g_sBurstDevs.mBurstReceiver.pRxFile) {
            DemoLogf("DemoBurstReceiver: Closing file\n");
            fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
            g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
          }
          g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
    
          if (g_sBurstDevs.mBurstReceiver.pRxFile) {
            DemoLogf("mBurstSender: Closing file\n");
            fclose(g_sBurstDevs.mBurstSender.pTxFile);
            g_sBurstDevs.mBurstSender.pTxFile = NULL;
          }
          g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;

          g_sBurstDevs.mBurstReceiver.bVerbose = OTX_FALSE;
          DemoLogf("Burst Receiver Verbose is now %s\n", 
                   g_sBurstDevs.mBurstReceiver.bVerbose ? "ON" : "OFF");

          g_sBurstDevs.mBurstSender.bVerbose = OTX_FALSE;
          DemoLogf("Burst Sender Verbose is now %s\n", 
                   g_sBurstDevs.mBurstSender.bVerbose ? "ON" : "OFF");
        
        }
      }
    }
    break;


  case 'L':
    {
      g_sBurstDevs.mBurstSender.bFileLoop = !g_sBurstDevs.mBurstSender.bFileLoop;
      DemoLogf("Setting loop file sends %s for Burst Sender[%d] Thread\n", (g_sBurstDevs.mBurstSender.bFileLoop)?"ON":"OFF", 0);
    }
    break;

  case 'b':
    {
      DemoLogf("Debug Dump for Dsp#%d\n", nCurrentLi);
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nCurrentLi], 0);
    }
    break;


  case 'f':
    // Set Active Pots Li Off-Hook ...
    DemoLogf("Pots Li#%d Off-hook\n", nCurrentLi);
    result = OtxCstPotsOffHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1000, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;


  case 'n':
    // Set Active Pots Li On-Hook ...
    DemoLogf("Pots Li#%d On-hook\n", nCurrentLi);
    result = OtxCstPotsOnHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1001, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'm':
    // Set Active Pots Li in On-Hook Line Monitor mode ...
    DemoLogf("Pots Li#%d On-hook Line-Monitor mode\n", nCurrentLi);
    result = OtxCstPotsOnHookLineMon(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1001, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'c':
    // Line Sense
    g_bLineSenseVerbose = !g_bLineSenseVerbose;
    g_bLineSensePending = OTX_FALSE;
    DemoLogf("Pots Li#%d Line Sense is now %s\n", 
             nCurrentLi, 
             g_bLineSenseVerbose ? "Verbose" : "Silent");
    break;

  case 'G':
    // Disconnect VAD
    DemoLogf("Pots Li#%d Disconnect VAD \n", nCurrentLi);
    if(g_demoEnv.eDetectorType == OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR) {
      // This was configured during startup. It was not setup
      // from command console.
      DemoLogf("VAD connected during application startup\n");
      DemoLogf("  No change in connection.\n");
    }
    else {
      if( (hPipeVAD0 == OTX_INVALID_HANDLE_VALUE) ||
          (hPipeVAD1 == OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Pots Li VAD was not connected\n");
      }
      else { 

        // -------------------------
        // Disconnect pipe: Dsp --> Dsp (VAD)
        // -------------------------
        result = OtxDrvDisconnectPipe(hPipeVAD0, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeVAD0, result);
        result = OtxDrvDisconnectPipe(hPipeVAD0, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeVAD0, result);

        result = OtxDrvClose(hPipeVAD0);
        DemoCheckRc(hPipeVAD0, result);

        hPipeVAD0 = OTX_INVALID_HANDLE_VALUE;

        result = OtxDrvDisconnectPipe(hPipeVAD1, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeVAD1, result);
        result = OtxDrvDisconnectPipe(hPipeVAD1, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeVAD1, result);

        result = OtxDrvClose(hPipeVAD1);
        DemoCheckRc(hPipeVAD1, result);

        hPipeVAD1 = OTX_INVALID_HANDLE_VALUE;      

        DemoLogf("Pots Li VAD connection removed\n");
      }
    }
       
    break;

  case 'g':
    // Connect VAD
    DemoLogf("Pots Li#%d connect VAD \n", nCurrentLi);

    if(g_demoEnv.eDetectorType == OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR) {
      // This was configured during startup. It was not setup
      // from command console.
      DemoLogf("VAD connected during application startup\n");
      DemoLogf("  No change in connection.\n");
    }
    else {
      OTX_INT32 nHwy = nCurrentLi;
      OTX_INT32 nTs;
      OTX_UINT8 sqTimeSlotMask[32];

      if( (hPipeVAD0 != OTX_INVALID_HANDLE_VALUE) ||
          (hPipeVAD1 != OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Pots Li VAD was already connected\n");
      }
      else {
        if(!(DemoIsVADDeviceEnabled(nCurrentLi))) {
          result = OtxDrvEnable(g_hLogicDevs.hVaDetector[nCurrentLi], 
                                OTX_TASK_SYNC, 
                                5000L); 
          DemoCheckRc(g_hLogicDevs.hVaDetector[nCurrentLi], result);
        }

        // -------------------------
        // Create and Connect pipe:  Dsp --> Dsp (VAD)
        // -------------------------
        result = OtxDrvCreatePipe(64, &hPipeVAD0);
        DemoCheckRc(hPipeVAD0, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_RX;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeVAD0, 
                                   OTX_PIPE_INPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy,
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeVAD0, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_VAD;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeVAD0, 
                                   OTX_PIPE_OUTPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeVAD0, result); 
        
        // 16 bit data mode (Linear) needs 2nd pipe connection 
        result = OtxDrvCreatePipe(64, &hPipeVAD1);
        DemoCheckRc(hPipeVAD1, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_RX + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeVAD1, 
                                   OTX_PIPE_INPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy,
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeVAD1, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_VAD + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeVAD1, 
                                   OTX_PIPE_OUTPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeVAD1, result); 

        DemoLogf("Pots Li VAD connection setup.\n");      
      }
    }

    break;

  case 'I':
    // Disconnect CPT Detector
    DemoLogf("Pots Li#%d Disconnect CPT Detector \n", nCurrentLi);
    if(g_demoEnv.eDetectorType == OTX_LDEVICE_CST_CPT_DETECTOR) {
      // This was configured during startup. It was not setup
      // from command console.
      DemoLogf("CPT Detector connected during application startup\n");
      DemoLogf("  No change in connection.\n");
    }
    else {
      if( (hPipeCPT0 == OTX_INVALID_HANDLE_VALUE) ||
          (hPipeCPT1 == OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Pots Li CPT Detector was not connected\n");
      }
      else { 

        // -------------------------
        // Disconnect pipe: Dsp --> Dsp (CPT)
        // -------------------------
        result = OtxDrvDisconnectPipe(hPipeCPT0, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeCPT0, result);
        result = OtxDrvDisconnectPipe(hPipeCPT0, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeCPT0, result);

        result = OtxDrvClose(hPipeCPT0);
        DemoCheckRc(hPipeCPT0, result);

        hPipeCPT0 = OTX_INVALID_HANDLE_VALUE;

        result = OtxDrvDisconnectPipe(hPipeCPT1, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeCPT1, result);
        result = OtxDrvDisconnectPipe(hPipeCPT1, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeCPT1, result);

        result = OtxDrvClose(hPipeCPT1);
        DemoCheckRc(hPipeCPT1, result);

        hPipeCPT1 = OTX_INVALID_HANDLE_VALUE;      

        DemoLogf("Pots Li CPT Detector connection removed\n");
      }
    }
       
    break;

  case 'i':
    // Connect CPT
    DemoLogf("Pots Li#%d connect CPT \n", nCurrentLi);

    if(g_demoEnv.eDetectorType == OTX_LDEVICE_CST_CPT_DETECTOR) {
      // This was configured during startup. It was not setup
      // from command console.
      DemoLogf("CPT connected during application startup\n");
      DemoLogf("  No change in connection.\n");
    }
    else {
      OTX_INT32 nHwy = nCurrentLi;
      OTX_INT32 nTs;
      OTX_UINT8 sqTimeSlotMask[32];

      if( (hPipeCPT0 != OTX_INVALID_HANDLE_VALUE) ||
          (hPipeCPT1 != OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Pots Li CPT was already connected\n");
      }
      else {
        // -------------------------
        // Create and Connect pipe:  Dsp --> Dsp (CPT)
        // -------------------------
        result = OtxDrvCreatePipe(64, &hPipeCPT0);
        DemoCheckRc(hPipeCPT0, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_RX;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeCPT0, 
                                   OTX_PIPE_INPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy,
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeCPT0, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_CPT;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeCPT0, 
                                   OTX_PIPE_OUTPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeCPT0, result); 
        
        // 16 bit data mode (Linear) needs 2nd pipe connection 
        result = OtxDrvCreatePipe(64, &hPipeCPT1);
        DemoCheckRc(hPipeCPT1, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_RX + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeCPT1, 
                                   OTX_PIPE_INPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy,
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeCPT1, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_CPT + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeCPT1, 
                                   OTX_PIPE_OUTPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeCPT1, result); 

        DemoLogf("Pots Li CPT Detector connection setup.\n");
      }
    }

    break;

  case 'J':
    // Disconnect DTMF Detector
    DemoLogf("Pots Li#%d Disconnect DTMF Detector \n", nCurrentLi);
    if(g_demoEnv.eDetectorType == OTX_LDEVICE_CST_DTMF_DETECTOR) {
      // This was configured during startup. It was not setup
      // from command console.
      DemoLogf("DTMF Detector connected during application startup\n");
      DemoLogf("  No change in connection.\n");
    }
    else {
      if( (hPipeDTMF0 == OTX_INVALID_HANDLE_VALUE) ||
          (hPipeDTMF1 == OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Pots Li DTMF Detector was not connected\n");
      }
      else { 

        // -------------------------
        // Disconnect pipe: Dsp --> Dsp (DTMF Detector)
        // -------------------------
        result = OtxDrvDisconnectPipe(hPipeDTMF0, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeDTMF0, result);
        result = OtxDrvDisconnectPipe(hPipeDTMF0, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeDTMF0, result);

        result = OtxDrvClose(hPipeDTMF0);
        DemoCheckRc(hPipeDTMF0, result);

        hPipeDTMF0 = OTX_INVALID_HANDLE_VALUE;

        result = OtxDrvDisconnectPipe(hPipeDTMF1, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeDTMF1, result);
        result = OtxDrvDisconnectPipe(hPipeDTMF1, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeDTMF1, result);

        result = OtxDrvClose(hPipeDTMF1);
        DemoCheckRc(hPipeDTMF1, result);

        hPipeDTMF1 = OTX_INVALID_HANDLE_VALUE;      

        DemoLogf("Pots Li DTMF Detector connection removed\n");
      }
    }
       
    break;

  case 'j':
    // Connect DTMF Detector
    DemoLogf("Pots Li#%d connect DTMF Detector \n", nCurrentLi);
    if(g_demoEnv.eDetectorType == OTX_LDEVICE_CST_DTMF_DETECTOR) {
      // This was configured during startup. It was not setup
      // from command console.
      DemoLogf("DTMF Detector connected during application startup\n");
      DemoLogf("  No change in connection.\n");
    }
    else {
      OTX_INT32 nHwy = nCurrentLi;
      OTX_INT32 nTs;
      OTX_UINT8 sqTimeSlotMask[32];

      if( (hPipeDTMF0 != OTX_INVALID_HANDLE_VALUE) ||
          (hPipeDTMF1 != OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Pots Li DTMF Detector was already connected\n");
      }
      else {      
        // -------------------------
        // Create and Connect pipe:  Dsp --> Dsp (DTMF Detector)
        // -------------------------
        result = OtxDrvCreatePipe(64, &hPipeDTMF0);
        DemoCheckRc(hPipeDTMF0, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_RX;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeDTMF0, 
                                   OTX_PIPE_INPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy,
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeDTMF0, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_DTMF_DETECTOR;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeDTMF0, 
                                   OTX_PIPE_OUTPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeDTMF0, result); 
        
        // 16 bit data mode (Linear) needs 2nd pipe connection 
        result = OtxDrvCreatePipe(64, &hPipeDTMF1);
        DemoCheckRc(hPipeDTMF1, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_RX + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeDTMF1, 
                                   OTX_PIPE_INPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy,
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeDTMF1, result);

        memset(sqTimeSlotMask, 
               0, 
               sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_DTMF_DETECTOR + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeDTMF1, 
                                   OTX_PIPE_OUTPUT,  
                                   g_hPhysDevs.hTss,  
                                   nHwy, 
                                   sqTimeSlotMask, 
                                   32);
        DemoCheckRc(hPipeDTMF1, result); 

        DemoLogf("Pots Li DTMF Detector connection setup.\n");
      }
    }

    break;

  case 'k':
    {
      OTX_INT32 nHwy = nCurrentLi;
      OTX_INT32 nDsp = nCurrentLi;
      OTX_INT32 nTs;
      OTX_UINT8 sqTimeSlotMask[32];

      // Check to see if burst sender was connected directly to the PotLi
      if( bBurstInBypass == OTX_TRUE ) {

        bBurstInBypass = OTX_FALSE;

        DemoLogf("Pots Li#%d Removing Burst direct connection to PotsLi\n", 
                 nCurrentLi);      

        // -------------------------
        // Disonnect pipe: Burst -> PotsLi
        // -------------------------
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_INPUT);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_OUTPUT);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_INPUT);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_OUTPUT);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

        // -------------------------
        // Connect pipe: Burst -> Dsp (Voice Mixer1)
        // -------------------------
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_BURST_PLAYBACK;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_BURST_PLAYBACK + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1 + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);
      }

      if( (hPipeNoVoiceMix0 != OTX_INVALID_HANDLE_VALUE) ||
          (hPipeNoVoiceMix1 != OTX_INVALID_HANDLE_VALUE) ) {
        DemoLogf("Pots Li DTMF generator was in bypass configuration\n");

        // -------------------------
        // Disonnect pipe: DTMF -> PotsLi
        // -------------------------
        result = OtxDrvDisconnectPipe(hPipeNoVoiceMix0, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeNoVoiceMix0, result);
        result = OtxDrvDisconnectPipe(hPipeNoVoiceMix1, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeNoVoiceMix1, result);
        
        result = OtxDrvClose(hPipeNoVoiceMix0);
        DemoCheckRc(hPipeNoVoiceMix0, result);

        hPipeNoVoiceMix0 = OTX_INVALID_HANDLE_VALUE;

        result = OtxDrvClose(hPipeNoVoiceMix1);
        DemoCheckRc(hPipeNoVoiceMix1, result);

        hPipeNoVoiceMix1 = OTX_INVALID_HANDLE_VALUE;

        result = OtxDrvEnable(g_hLogicDevs.hVoiceMixer[nDsp], 
                              OTX_TASK_SYNC, 
                              5000L); 
        DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);
        DemoLogf("Pots Li#%d Enabled voice mixer\n", nCurrentLi);

        DemoLogf("Pots Li#%d DTMF generator connected to Voice Mixer Finished\n",
                 nCurrentLi);
      }
      else {      
        // DTMF direct connect to PotsLi
        DemoLogf("Pots Li#%d DTMF generator bypass voice mixer\n", nCurrentLi);

        result = OtxDrvDisable(g_hLogicDevs.hVoiceMixer[nDsp], OTX_TASK_SYNC, 5000L); 
        DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);
        DemoLogf("Pots Li#%d Disabled voice mixer\n", nCurrentLi);
    
        // TSS connection to PotsLi TX
        result = OtxDrvCreatePipe(64, &hPipeNoVoiceMix0);
        DemoCheckRc(hPipeNoVoiceMix0, result);
    
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_TX;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeNoVoiceMix0, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
        DemoCheckRc(hPipeNoVoiceMix0, result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_DTMF_DIALER;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeNoVoiceMix0, OTX_PIPE_INPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
        DemoCheckRc(hPipeNoVoiceMix0, result);

        // Also connect the DAA samples to the burst (TS0) via a TSS connection
        result = OtxDrvCreatePipe(64, &hPipeNoVoiceMix1);
        DemoCheckRc(hPipeNoVoiceMix1, result);
    
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_TX + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeNoVoiceMix1, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
        DemoCheckRc(hPipeNoVoiceMix1, result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_DTMF_DIALER + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(hPipeNoVoiceMix1, OTX_PIPE_INPUT, g_hPhysDevs.hTss, nHwy, sqTimeSlotMask, 32); 
        DemoCheckRc(hPipeNoVoiceMix1, result);

        DemoLogf("Pots Li#%d DTMF Connected to PotLi Finished\n", nCurrentLi);
      }
    }

    break;

  case 'K':
    {
      OTX_INT32 nHwy = nCurrentLi;
      OTX_INT32 nDsp = nCurrentLi;
      OTX_INT32 nTs;
      OTX_UINT8 sqTimeSlotMask[32];

      // Check to see if DTMF was directly connected to PotLi
      if( (hPipeNoVoiceMix0 != OTX_INVALID_HANDLE_VALUE) ||
          (hPipeNoVoiceMix1 != OTX_INVALID_HANDLE_VALUE) ) {
        // DTMF direct connect to PotsLi

        DemoLogf("Pots Li#%d Removing DTMF direct connection to PotsLi\n", 
                 nCurrentLi);      

        // -------------------------
        // Disonnect pipe: DTMF -> PotsLi
        // -------------------------
        result = OtxDrvDisconnectPipe(hPipeNoVoiceMix0, OTX_PIPE_INPUT);
        DemoCheckRc(hPipeNoVoiceMix0, result);
        result = OtxDrvDisconnectPipe(hPipeNoVoiceMix1, OTX_PIPE_OUTPUT);
        DemoCheckRc(hPipeNoVoiceMix1, result);
        
        result = OtxDrvClose(hPipeNoVoiceMix0);
        DemoCheckRc(hPipeNoVoiceMix0, result);

        hPipeNoVoiceMix0 = OTX_INVALID_HANDLE_VALUE;

        result = OtxDrvClose(hPipeNoVoiceMix1);
        DemoCheckRc(hPipeNoVoiceMix1, result);

        hPipeNoVoiceMix1 = OTX_INVALID_HANDLE_VALUE;
      }


      if( bBurstInBypass == OTX_TRUE ) {

        bBurstInBypass = OTX_FALSE;

        DemoLogf("Pots Li#%d Burst Sender was in bypass configuration\n", 
                 nCurrentLi);      

        // -------------------------
        // Disonnect pipe: Burst -> PotsLi
        // -------------------------
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_INPUT);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_OUTPUT);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_INPUT);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_OUTPUT);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

        // -------------------------
        // Connect pipe: Burst -> Dsp (Voice Mixer)
        // -------------------------
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_BURST_PLAYBACK;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_BURST_PLAYBACK + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_VOICE_MIX_INPUT1 + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);


        result = OtxDrvEnable(g_hLogicDevs.hVoiceMixer[nDsp], 
                              OTX_TASK_SYNC, 
                              5000L); 
        DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);
        DemoLogf("Pots Li#%d Enabled voice mixer\n", nCurrentLi);

        DemoLogf("Pots Li#%d Burst Sender connected to Voice Mixer Finished\n", nCurrentLi);
      }
      else {      
        bBurstInBypass = OTX_TRUE;

        // Burst Sender direct connect to PotsLi
        DemoLogf("Pots Li#%d Burst bypass voice mixer\n", 
                 nCurrentLi);
        result = OtxDrvDisable(g_hLogicDevs.hVoiceMixer[nDsp], 
                               OTX_TASK_SYNC, 
                               5000L); 
        DemoCheckRc(g_hLogicDevs.hVoiceMixer[nDsp], result);
        DemoLogf("Pots Li#%d Disabled voice mixer\n", nCurrentLi);

        // -------------------------
        // Disonnect pipe: Burst -> Dsp (Voice Mixer1)
        // -------------------------
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_INPUT);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_OUTPUT);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_INPUT);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);
        result = OtxDrvDisconnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_OUTPUT);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

        // -------------------------
        // Connect pipe: Burst -> Dsp (PotsLi)
        // -------------------------
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_BURST_PLAYBACK;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_TX;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1a[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1a[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_BURST_PLAYBACK + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hTss, DEMO_HWY_BURST, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        nTs = DEMO_TS_DSP_POTS_LI_TX + 1;
        sqTimeSlotMask[nTs] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes.hMixer1b[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32);
        DemoCheckRc(g_hPipes.hMixer1b[nDsp], result);

        DemoLogf("Pots Li#%d Burst Sender connected to PotLi Finished\n", nCurrentLi);
      }
    }

    break;

  case 'l':
    // Flash Hook
    DemoLogf("Pots Li#%d Hook Flash\n", nCurrentLi);
    result = OtxCstPotsHookFlash(g_hLogicDevs.hPotsLi[nCurrentLi], 200, 0x1002, 2000); DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'd':
    {
      // Dial a hard-coded number ...
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialer[nCurrentLi];

      DemoLogf("Sending digits on Pots Li#%d: '%s'", nCurrentLi, g_sqNumber0);

      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxCstDtmfDial(
        hHandle, 
        (unsigned char *)g_sqNumber0, 
        strlen(g_sqNumber0), 
        DEMO_TASK_REF, 
        OTX_TIME_INFINITY); 
      DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'D':
    {
      // Dial a user specified number ...
      char strBuf[32];
      OTX_UINT32 bufferSize;
      static char saveStrBuf[32] = "608";
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialer[nCurrentLi];

      DemoLogf("Last number was [%s] \n", saveStrBuf);
      DemoLogf("Phone number to Dial from Pots Li#%d: ", nCurrentLi);
      OsaConsoleGetStr(strBuf);
      // Did user enter a string
      if(strBuf[0]) {
        bufferSize = OsaStrLen(strBuf) + 1; // Include Null character
        if(bufferSize > 31) {
          DemoLogf("Error: Buffer overflow for string %s:\n", saveStrBuf);
        }
        // Save the new string
        OsaStrNCpy(saveStrBuf, // To
                   strBuf,     // From
                   bufferSize);   
      }
      DemoLogf("Pots Li#%d Dialing number '%s'\n", nCurrentLi, saveStrBuf);

      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxCstDtmfDial(
        hHandle, 
        (unsigned char *)saveStrBuf, 
        OsaStrLen(saveStrBuf), 
        DEMO_TASK_REF, 
        OTX_TIME_INFINITY); 
      DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'p':
    {
      // Pulse dial
      OTX_INT32 c;
      DemoLogf("Pots Li#%d Pulse Dialing. Press digit to dial... [0-9] ", nCurrentLi);
      c = OsaConsoleGetCh();
      DemoLogf("\n");
      result = OtxCstPotsPulseDial(g_hLogicDevs.hPotsLi[nCurrentLi], c, 0x1003, 2000); DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      nCurrentLi = nLi;
      DemoLogf("Li#%d Selected ....\n", nCurrentLi);

      g_demoEnv.fxoNo = nCurrentLi;
      DemoInitializeBurstRecord();
    }
    break;

#if 1 // Only for DSP program debugging
  case 's':
    {
      char sqzBuf[OTX_MAX_PATH];
      
      if (g_fStreamFile) {
        fclose(g_fStreamFile);
        g_fStreamFile = NULL;
        DemoLogf("Closed streaming file\n");
      } else {
        
        DemoLogf("Please enter file to receive: ");
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          break;
        }
        
        if ((g_fStreamFile = fopen(sqzBuf, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }
      }
    }
    break;
#endif

#if 1 // Only for DSP program debugging
  case 'r':
    {
      OTX_UINT32 nCmd = 1; // 1 = Read
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nDummy2 = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Read DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      nCmd = (nDummy3 << 3*8) | (nDummy2 << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;
#endif

#if 1 // Only for DSP program debugging
  case 'w':
    {
      OTX_UINT32 nCmd = 2; // 2 = Write
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nVal = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Write DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      DemoLogf("Please enter value [0x%.2X] ", nVal);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nVal = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      nCmd = (nDummy3 << 3*8) | (nVal << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;
#endif
  


  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;
      
      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);
      
      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
      
    }
    break;
    
  case 'q':
  case 'Q':
    // Quit ...
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) 
/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->outFileName       = NULL;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->fxoNo             = DEMO_DEFAULT_FXO_NO;
  envP->nSetId            = OTX_CST_POTS_CFG_SET_USA;
  envP->nBoardType        = OTX_DEVICE_ARNI_6X6_PCI;
  envP->eDetectorType     = OTX_LDEVICE_CST_DTMF_DETECTOR;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'O') {   // DSP out file
          envP->outFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'F') {   // FXO number
          envP->fxoNo = strtoul(argv[i]+2, NULL, 0);

        } else if (OsaToUpper(*(argv[i]+1)) == 'T') {   // Board Type
          if        (OsaToUpper(*(argv[i] + 2)) == 'A') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'B') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6_FXO_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'D') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X3_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'J') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI;
          } else {            
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else if (*(argv[i]+2) == 'L') {   // 16 bit Raw Data
            envP->law = OTX_DATA_FORMAT_UINT16;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'S') {   // Configuration Set
          if (*(argv[i]+2) == 'U') {          // USA
            envP->nSetId = OTX_CST_POTS_CFG_SET_USA;
          } else if (*(argv[i]+2) == 'G') {   // Germany
            envP->nSetId = OTX_CST_POTS_CFG_SET_GERMANY;
          } else if (*(argv[i]+2) == 'E') {   // UK
            envP->nSetId = OTX_CST_POTS_CFG_SET_UK;
          } else if (*(argv[i]+2) == 'F') {   // France
            envP->nSetId = OTX_CST_POTS_CFG_SET_FRANCE;
          } else if (*(argv[i]+2) == 'A') {   // Australia
            envP->nSetId = OTX_CST_POTS_CFG_SET_AUSTRALIA;
          } else if (*(argv[i]+2) == 'J') {   // Japan
            envP->nSetId = OTX_CST_POTS_CFG_SET_JAPAN;
          } else {
            DemoLogf("Error: Unsupported country specific configuration set %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'D') {   // Detector Type
          if (*(argv[i]+2) == 'C') {          // CPT
            envP->eDetectorType = OTX_LDEVICE_CST_CPT_DETECTOR;
          } else if (*(argv[i]+2) == 'D') {          // DTMF
            envP->eDetectorType = OTX_LDEVICE_CST_DTMF_DETECTOR;
          } else if (*(argv[i]+2) == 'V') {          // Voice Activity
            envP->eDetectorType = OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR;
          } else if (*(argv[i]+2) == 'I') {          // Voice Activity
            envP->eDetectorType = OTX_LDEVICE_CST_CALLER_ID_DETECTOR;
          } else {
            DemoLogf("Error: Unsupported detector type %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


static void DemoHelpPanel(void)
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>    : Log file name\n");
  DemoLogf(" /B<boardNo>     : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /F<fxoNo>       : FXO number to configure (Default is %d)\n", DEMO_DEFAULT_FXO_NO);
  DemoLogf(" /T[A|B|D|J]     : Board Type used for the demo, where\n");
  DemoLogf("                    A  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    B  = Use Arni-6-FXO-PCI-Plus Board\n");
  DemoLogf("                    D  = Use Arni-6x3-PCI-Plus Board\n");
  DemoLogf("                    J  = Use Arni-6-PCI Board\n");
  DemoLogf(" /D[D|C|V|I]     : Detector type, where\n");
  DemoLogf("                    D  = DTMF Detector (default)\n");
  DemoLogf("                    C  = Call Progress Tone Detector\n");
  DemoLogf("                    V  = Voice Activity Detector\n");
  DemoLogf("                    I  = Caller Id Detector\n");
  DemoLogf(" /C[A|U]         : Voice Compression Codec, where\n");
  DemoLogf("                    A  = A-law\n");
  DemoLogf("                    L  = 16 Bit Linear\n");
  DemoLogf("                    U  = u-law (default)\n");
  DemoLogf(" /S[U|G|E|F|A|J] : Country specific setting, where\n");
  DemoLogf("                    U  = Li Configuration for USA (default)\n");
  DemoLogf("                    G  = Li Configuration for Germany\n");
  DemoLogf("                    E  = Li Configuration for United Kingdom\n");
  DemoLogf("                    F  = Li Configuration for France\n");
  DemoLogf("                    A  = Li Configuration for Australia\n");
  DemoLogf("                    J  = Li Configuration for Japan\n");
  DemoLogf(" /O<fileName>    : DSP Out file name (for debugging only)\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  //OTX_INT32  i;
  OTX_INT32  exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char       commandChar;

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    exit(-1);
  }

  DemoLogf("%s, %s, Odin TeleSystems Inc.\n", argv[0], __DATE__);
  DemoLogf("Arni-6x6-PCI, Basic FXO Demo Program.\n");

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("FXO   number %d\n", g_demoEnv.fxoNo);
  if (g_demoEnv.outFileName) {
    DemoLogf("Out file to be loaded is %s\n", g_demoEnv.outFileName);
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    DemoExit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    DemoExit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  g_bLibConnected = OTX_TRUE;

  // Create Event Queues for Driver --> Application Communications
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Arni-16-PCI board
  DemoOpenPhysicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);

    free(pszCpldBuildNumbers);

  }

  // Enable Physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  OsaThreadStart(DemoIdleFunction);

  // Create and get handles to the Logical devices needed by this application
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  // Connect Pipes to Physical and Logical Devices
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable Logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();

#if 0 // Disable watchdog while debugging
  // Enable the Watchdog Timer itself
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Start the Watchdog timer Reset Thread, which will keep the watchdog alive
  OsaThreadStart(DemoWatchDogFunction);
#endif  

  // Start the Line Sensing Thread, which requests period line sensing from the PotsLi to faciliate monitoring of Far-End-Hangup conditions
  OsaThreadStart(DemoLineSensingFunction);

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  DemoLogf("\nPress 'h' for help, 'q' to quit ...\n");

  // use DemoBurstSender() not 
  g_sBurstDevs.mBurstSender.bUseAlternateSend = OTX_FALSE;

  while (!exitFlag) {
    g_bConsoleUsed = OTX_TRUE;
    commandChar = OsaConsoleGetCh();
    g_bConsoleUsed = OTX_FALSE;
		exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Delete allocated resources
  DemoExit(0);

  return 0;
}

static int DemoExit(int nExitCode)
/*******************************************************************************
*******************************************************************************/
{

  DemoLogf("\nExiting...\n");

  // Important to disable the Watchdog Timer since we will not be reseting it after we exit
  // the application. 
  OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_FALSE);

  g_bWatchDogThreadActive = OTX_FALSE;
  g_bLineSensingThreadActive = OTX_FALSE;

  if (DemoIsBurstDeviceEnabled()) {
    if (g_sBurstDevs.mBurstReceiver.pRxFile) {
      DemoLogf("DemoBurstReceiver: Closing file\n");
      fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
      g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
    }
    g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
    
    if (g_sBurstDevs.mBurstReceiver.pRxFile) {
      DemoLogf("mBurstSender: Closing file\n");
      fclose(g_sBurstDevs.mBurstSender.pTxFile);
      g_sBurstDevs.mBurstSender.pTxFile = NULL;
    }
    g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
  }

  OsaThreadSleep(100);   // give threads a chance to exit

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  exit(nExitCode);
}
